# -*- coding: UTF-8 -*-
# pylint: disable=super-init-not-called

"""
Main module: contains main method main/loop
"""

# System imports
import logging
import tornado.ioloop               as ioloop
import tornado.gen                  as gen
import os
import setproctitle
import sys
from django.apps                    import AppConfig

if 'DJANGO_SETTINGS_MODULE' not in os.environ:
    os.environ['DJANGO_SETTINGS_MODULE'] = "symonitoring_alert.settings"

# Project imports
import symonitoring_alert.common    as cmn  # noqa
import symonitoring_alert.no_block  as nb  # noqa

s = cmn.get_settings()

_stop = False


class SymonitoringAlertConfig(AppConfig):
    """AppConfig for symonitoring_alert"""
    # pylint: disable=no-init
    name = 'symonitoring_alert'
    verbose_name = "SyMonitoring Alert"

default_app_config = 'symonitoring_alert.SyMonitoringAlertConfig'


def admin():
    """Django admin for installed package"""
    nb.disabled = True
    from django.core.management import execute_from_command_line
    execute_from_command_line(sys.argv)


def main():
    """symonitoring_alert main routine"""
    # System imports
    import gc
    import django.db                                as db
    import django.conf                              as dc

    # Project imports
    import symonitoring_alert.query                 as q
    # Start sleekxmpp bot
    import symonitoring_alert.communication.xmpp    as xmpp
    import symonitoring_alert.lock                  as lock

    lock.lock()
    xmpp.main_called = True

    cmn.lg.critical("")
    cmn.lg.critical("SETTINGS INFO:")
    cmn.lg.critical("-------------")
    cmn.lg.critical("     Settings module: %s"     % (
        os.environ['DJANGO_SETTINGS_MODULE']
    ))
    cmn.lg.critical("     Debug mode (local): %s"  % s.DEBUG)
    cmn.lg.critical("     Debug mode (Django): %s" % dc.settings.DEBUG)
    cmn.lg.critical("     Dummy mode: %s"          % s.DUMMY_MODE)
    cmn.lg.critical("")
    cmn.lg.critical("     Poll rate: %s"           % s.POLL_RATE)
    cmn.lg.critical("     Hold rate: %s"           % s.HOLD_RATE)
    cmn.lg.critical("     XMPP check rate: %s"     % s.XMPP_CHECK_RATE)
    cmn.lg.critical("     Dependency timeout: %s"  % s.DEPENDENCY_TIMEOUT)
    cmn.lg.critical("     GC rate: %s"             % s.COLLECT_RATE)
    cmn.lg.critical("")
    cmn.lg.critical("     Timing test scale: %s"   % s.TIMING_TEST_SCALE)

    class MainLoop(cmn.ExecutorMixin):
        """Main loop helper class"""
        def __init__(self):
            self.main_active   = False
            self.hold_active   = False
            self.ack_active    = False

        @gen.coroutine
        def initialize(self):
            """Initialize alerts that already exists"""
            self.main_active   = True
            self.hold_active   = True
            self.ack_active    = True
            try:
                yield q.NagiosLog().initialize_alerts()
            finally:
                self.main_active   = False
                self.hold_active   = False
                self.ack_active    = False

        @gen.coroutine
        def main_callback(self):
            """Main periodic callback, will handle the nagioslog and log some
            statistics"""
            # Do not overlap callbacks if we have heavy load
            # Remember this is not threading so it is ok to use flags
            if not self.main_active:
                try:
                    self.main_active = True
                    # Important stuff
                    yield q.NagiosLog().handle()

                    # Logging
                    loop = cmn.io_loop()
                    timeouts  = -1
                    callbacks = -1
                    # pylint: disable=no-member, protected-access
                    try:
                        timeouts = len(loop._timeouts)
                    except AttributeError:  # pragma: no cover
                        pass
                    try:
                        callbacks = len(loop._callbacks)
                    except AttributeError:  # pragma: no cover
                        pass
                    cmn.lg.info(
                        "Polling: %3d/%3d pending timeouts/callbacks",
                        timeouts,
                        callbacks,
                    )
                    if _stop:
                        cmn.io_loop().stop()
                finally:
                    self.main_active = False

        @gen.coroutine
        def hold_callback(self):
            """ Periodic callback for handling hold entries.
            """

            if not self.hold_active:
                try:
                    self.hold_active = True
                    yield q.Hold().handle()
                finally:
                    self.hold_active = False

        @cmn.run_on_executor
        def gc(self):  # pragma: no cover
            """Garbage collect"""
            db.reset_queries()
            res = gc.collect()
            cmn.lg.info(
                "Collected garbage, unreachable: %s",
                res
            )

        @classmethod
        def ipdb(cls):  # pragma: no cover
            """"Run debugger"""
            import ipdb
            ipdb.set_trace()

    ml = MainLoop()
    ioloop.PeriodicCallback(
        cmn.coro_engine(ml.main_callback),
        s.POLL_RATE
    ).start()
    ioloop.PeriodicCallback(
        cmn.coro_engine(ml.hold_callback),
        s.HOLD_RATE
    ).start()
    ioloop.PeriodicCallback(
        cmn.coro_engine(ml.gc),
        s.COLLECT_RATE
    ).start()
    ioloop.PeriodicCallback(
        xmpp.check_start_xmpp,
        s.XMPP_CHECK_RATE
    ).start()
    loop = cmn.io_loop()
    cmn.add_callback(ml.initialize)
    loop.set_blocking_log_threshold(s.LOOP_LOG)
    if s.LOG_SQL:
        logging.getLogger("django.db.backends").setLevel(
            cmn.lg.DEBUG
        )
    loop.start()

setproctitle.setproctitle("syalert")  # pylint: disable=no-member
nb.install()
