# -*- coding: UTF-8 -*-


""" Logic class for alert """

# System imports
import django.db.transaction     as transaction
import weakref

# Project imports
import symonitoring_model.models as mod
import symonitoring_alert.common as cmn


class Alert(cmn.ExecutorMixin):
    """ Logic class for alert """
    __slots__ = (
        '_logdata',
        '_model',
    )

    def __init__(self, logdata):
        super(Alert, self).__init__()
        self._logdata = weakref.ref(logdata)
        self._model   = None

    @property
    def logdata(self):
        """:rtype: symonitoring_alert.logdata.LogData"""
        return self._logdata()

    @property
    def model(self):
        """:rtype: symonitoring_model.models.Alert"""
        return self._model

    @model.setter
    def model(self, value):
        """:param value: symonitoring_model.models.Alert"""
        self._model = value

    def create(self):
        """Create a new alert based on the collected data"""

        # This is kind of circular. I wanted to resolve this, but it gets
        # really ugly if you tries this, because in the end the database is
        # modeled like that. Alert <-> NagiosLog row (logdata)
        logdata = self.logdata
        alert                           = mod.Alert()
        alert.customerid                = logdata.host.customerid
        alert.hostid                    = logdata.host
        alert.serviceinstanceid         = logdata.service_instance
        alert.commandid                 = logdata.command
        alert.datetimeoccured           = logdata.log.datetime
        alert.state                     = logdata.log.state
        alert.output                    = logdata.log.output
        alert.datetimefirstnotification = None
        alert.datetimelastnotification  = None
        alert.slaid                     = logdata.sla
        alert.slacustomerid             = logdata.host.customerid
        alert.problemid                 = logdata.log.problemid
        alert.lastproblemid             = logdata.log.lastproblemid
        alert.acknowledged              = False
        alert.acknowledgedautomatically = False
        alert.acknowledgedprocessed     = False
        alert.save()
        self._model                     = alert
        self.logdata.log.alert          = alert

    @cmn.run_on_executor
    def save_ex(self):  # pragma: no cover
        """ Run save() on executor """
        self.save()

    def save(self):
        """ Save alert model """
        if self._model:
            self._model.save()

    @cmn.run_on_executor
    def sync_ex(self, override_id=None):
        """ Run sync() on executor """
        self.sync(override_id)

    def sync(self, override_id=None):
        """ Get latest alert data """
        if override_id:
            self._model = mod.Alert.objects.get(
                id=override_id
            )
        elif self._model:
            if self._model.id:
                self._model = mod.Alert.objects.get(
                    id=self._model.id
                )

    @cmn.run_on_executor
    def create_update_notification_datetime(self):
        """ Sets the first notification datetime to current datetime. """

        with transaction.atomic():
            self.sync()
            local_now = cmn.now()
            model = self.model
            if model.datetimefirstnotification is None:
                model.datetimefirstnotification = local_now
            model.datetimelastnotification = local_now

            self.save()
