# -*- coding: UTF-8 -*-

""" Syncing (creating/loading) and removing alert state """

# System imports
import datetime
import weakref

# Project imports
import symonitoring_model.models as mod
import symonitoring_alert.common as cmn


class AlertTimeoutReached(Exception):
    """ Defined timout of alert is reached """
    pass


class AlertState(cmn.ExecutorMixin):
    """ Syncing (creating/loading) and removing alert state """
    __slots__ = (
        '_logdata',
        '_timeframe_states',
    )

    def __init__(self, logdata):
        super(AlertState, self).__init__()
        self._logdata = weakref.ref(logdata)
        self._timeframe_states = {}

    @property
    def logdata(self):
        """:rtype: symonitoring_alert.logdata.LogData"""
        return self._logdata()

    @cmn.run_on_executor
    def sync_ex(self):
        """ Run sync() on executor """
        return self.sync()

    def sync(self):
        """Synchronize the database state with logdata state:
        * Loads states for all timeframes
        * Creates states for missing timeframes
        * Deletes state for removed timeframes
        """
        alert = self.logdata.alert.model
        if alert:
            for timeframe in self.logdata.timeframes:
                state = mod.AlertTimeframeState.objects.select_related(
                ).get_or_create(
                    alert     = alert,
                    timeframe = timeframe,
                )[0]
                self._timeframe_states[timeframe] = state
            active_states = set([
                timeframe.id for timeframe in self.logdata.timeframes
            ])
            states = mod.AlertTimeframeState.objects.filter(
                alert_id = alert.id,
            )
            existing_states = set([
                state.timeframe_id for state in states
            ])
            remove_states = existing_states - active_states
            for state in remove_states:
                mod.AlertTimeframeState.objects.get(
                    alert_id=alert.id,
                    timeframe_id=state
                ).delete()

    def get_state(self, timeframe):
        """:rtype: symonitoring_model.models.AlertTimeframeState"""
        return self._timeframe_states[timeframe]

    def increment_rule(self):
        """ Increments all timeframe states """
        for timeframe in self.logdata.timeframes:
            self.increment_rule_timeframe(timeframe)

    def increment_rule_timeframe(self, timeframe):
        """ Increments a timeframe state """
        users = self.logdata.users[timeframe]
        state = self.get_state(timeframe)
        state.rule_index            += 1
        if state.rule_index         >= len(timeframe.ext_rules):
            state.rule_index         = 0
            state.user_index        += 1
            if state.user_index     >= len(users):
                state.user_index     = 0
                state.restart_index += 1

    def get_timeout(self, timeframe):
        """ Calculate the timeout for a timeframe """
        logdata       = self.logdata
        alert         = logdata.alert.model
        elapsed       = cmn.now() - alert.datetimefirstnotification
        users         = logdata.users[timeframe]
        state         = self.get_state(timeframe)
        rule          = timeframe.ext_rules[state.rule_index]
        user          = users[state.user_index]
        rule_max      = max([xrule.delay for xrule in timeframe.ext_rules])
        frame_delay   = rule_max + logdata.sla.delaycycle
        restart_delay = frame_delay * len(users) + logdata.sla.delayrestart
        totaltime     = datetime.timedelta(
            minutes   = rule.delay
        ) + datetime.timedelta(
            minutes   = frame_delay
        ) * state.user_index + datetime.timedelta(
            minutes   = restart_delay
        ) * state.restart_index
        if totaltime > datetime.timedelta(
                minutes = logdata.sla.timeout
        ):
            cmn.lg.debug(
                "AlertTimeout (%s min) of alert %s reached",
                logdata.sla.timeout,
                logdata.alert.model.id,
            )
            raise AlertTimeoutReached()
        timeout       = totaltime - elapsed
        if timeout.total_seconds() < 0:
            timeout   = datetime.timedelta(seconds=0)
        return (rule, user, timeout)

    @cmn.run_on_executor
    def delete_ex(self):  # pragma: no cover
        """ Run delete() on executor """
        return self.delete()

    def delete(self):
        """ Delete all states of closed alerts """
        mod.AlertTimeframeState.objects.filter(
            alert__datetimerecovery__isnull=False
        ).delete()

    @cmn.run_on_executor
    def save_ex(self):
        """ Run save() on executor """
        return self.save()

    def save(self):
        """ Save complete alert state """
        for state in self._timeframe_states.values():
            state.save()
