# -*- coding: UTF-8 -*-

# pylint: disable=missing-docstring,too-many-lines,no-self-use
# pylint: disable=protected-access

# System imports
import datetime
import django.core.exceptions              as dex

# Project imports
import symonitoring_model.models           as mod
import symonitoring_alert.common           as cmn
import symonitoring_alert.handle.nagioslog as nl
import symonitoring_alert.logdata          as ld
import symonitoring_alert.alert.state      as state


notification_host_critical = {
    'address': u'192.24.12.3',
    'attempt': 1,
    'checkcommand': u'check-host-alive',
    'datetime': cmn.local_datetime(datetime.datetime(
        2014, 5, 6, 15, 19, 7
    )),
    'duration': u'0d 0h 0m 0s',
    'hostname': u'localhost',
    'lastproblemid': 47,
    'name': u'localhost',
    'objecttype': u'HOST',
    'output': u'PING CRITICAL - Packet loss = 100%',
    'perfdata': u'rta=5000.000000ms;3000.000000;5000.000000;0.000000 pl=100%;'
                '80;100;0',
    'problemid': 48,
    'processed': u'0',
    'state': u'DOWN',
    'statetype': u'HARD',
    'timestamp': cmn.local_datetime(datetime.datetime(
        2014, 5, 6, 15, 19, 7
    ))
}


class StateTest(cmn.TestCase):
    def __init__(self, *args, **kwargs):
        super(StateTest, self).__init__(*args, **kwargs)
        self._logdata = None

    def prepare_logdata(self, hostname):
        """Creates logdata and alert"""
        handler = nl.NagiosLog()
        log = mod.Nagioslog(
            **notification_host_critical
        )
        log.hostname = hostname
        self.save_log(log)

        @cmn.coroutine
        def coro():
            yield handler.handle(log, no_exec=True)
            raise cmn.Return(handler.logdata)

        return cmn.run_sync(coro)

    def run_load_state(self):
        self._logdata = self.prepare_logdata("timingHost00")
        logdata = self._logdata
        assert isinstance(logdata, ld.LogData)
        alert_state = state.AlertState(logdata)
        with cmn._allow_block_for_test_only():
            alert_state.sync()
        timeframe = logdata.timeframes[0]
        assert isinstance(timeframe, mod.Timeframe)
        timeframe_state = alert_state.get_state(timeframe)
        self.assertEqual(timeframe_state.timeframe.id, timeframe.id)
        # This tests the django orm / caching. If this fails just remove
        # test.
        self.assertEqual(timeframe_state.timeframe, timeframe)
        return alert_state

    def test_create_states(self):
        self.run_load_state()

    def test_save_states(self):
        alert_state = self.run_load_state()
        logdata = alert_state.logdata
        timeframe = logdata.timeframes[0]
        alert_state.increment_rule()
        local_state = alert_state.get_state(timeframe)
        self.assertAlmostEqual(local_state.rule_index, 1)
        with cmn._allow_block_for_test_only():
            db_state = mod.AlertTimeframeState.objects.get(
                timeframe_id=timeframe.id
            )
            self.assertEqual(db_state.rule_index, 0)
            local_state = alert_state.get_state(timeframe)
            self.assertAlmostEqual(local_state.rule_index, 1)
            alert_state.save()
            db_state = mod.AlertTimeframeState.objects.get(
                timeframe_id=timeframe.id
            )
            self.assertEqual(db_state.rule_index, 1)

    def test_delete_states(self):
        alert_state = self.run_load_state()
        logdata = alert_state.logdata
        timeframe = logdata.timeframes[0]
        local_state = alert_state.get_state(timeframe)
        with cmn._allow_block_for_test_only():
            alert_state.save()
            mod.AlertTimeframeState.objects.get(
                id=local_state.id
            )
            logdata._timeframes = []
            alert_state.sync()
            with self.assertRaises(dex.ObjectDoesNotExist):
                mod.AlertTimeframeState.objects.get(
                    id=local_state.id
                )

    def test_add_states(self):
        alert_state = self.run_load_state()
        logdata = alert_state.logdata
        timeframe = logdata.timeframes[0]
        with cmn._allow_block_for_test_only():
            new_frame = mod.Timeframe.objects.get(id=timeframe.id)
            new_frame.id = timeframe.id + 1
            logdata.timeframes.append(new_frame)
            alert_state.sync()
            alert_state.save()
            db_state = mod.AlertTimeframeState.objects.get(
                alert_id=logdata.alert.model.id,
                timeframe_id=new_frame.id
            )
            assert isinstance(db_state, mod.AlertTimeframeState)
