# -*- coding: UTF-8 -*-

"""
Common classes / functions

* ExecutorMixin loads executor according to settings. Used with
tornados tornado.concurrent.run_on_executor
"""

# pylint: disable=bad-whitespace,too-few-public-methods,invalid-name
# pylint: disable=superfluous-parens

# System imports
try:
    import unittest2      as unittest
except ImportError:  # pragma: no cover
    import unittest
import types
import time
import concurrent.futures    as futures
import datetime
import contextlib
import six
import toro
import tornado.gen           as gen
import tornado.ioloop        as ioloop
import tornado.concurrent    as tc
import signal
import logging
import os
from abc import ABCMeta
import django
import django.db                as db

# Project imports
import symonitoring_alert.no_block      as nb

_add_block_installed = False

django.setup()


def raise_handler_needed(e):
    """ Check handler has to raise exceptions """
    return isinstance(e, db.utils.DatabaseError) or \
        TestCase.unittest_active


def run_sync(callback, timeout=None):
    """Starts the `IOLoop`, runs the given function, and stops the loop.

    If the function returns a `.Future`, the `IOLoop` will run
    until the future is resolved. If it raises an exception, the
    `IOLoop` will stop and the exception will be re-raised to the
    caller.

    The keyword-only argument ``timeout`` may be used to set
    a maximum duration for the function. If the timeout expires,
    a `TimeoutError` is raised.

    This method is useful in conjunction with `tornado.gen.coroutine`
    to allow asynchronous calls in a ``main()`` function::

    @gen.coroutine
    def main():
        # do stuff...

        if __name__ == '__main__':
            IOLoop.instance().run_sync(main)
    """
    assert isinstance(callback, (types.FunctionType, types.MethodType)), \
        "run_sync needs a function or method!"
    return io_loop().run_sync(callback, timeout)


def add_callback(callback, *args, **kwargs):
    """Calls the given callback on the next I/O loop iteration.

    It is safe to call this method from any thread at any time,
    except from a signal handler.  Note that this is the **only**
    method in `IOLoop` that makes this thread-safety guarantee; all
    other interaction with the `IOLoop` must be done from that
    `IOLoop`'s thread.  `add_callback()` may be used to transfer
    control from other threads to the `IOLoop`'s thread.

    To add a callback from a signal handler, see
    `add_callback_from_signal`.
    """
    assert isinstance(callback, (types.FunctionType, types.MethodType)), \
        "run_sync needs a function or method!"
    func = coro_engine(callback)
    return io_loop().add_callback(func, *args, **kwargs)


def add_timeout(deadline, callback):
    """Runs the ``callback`` at the time ``deadline`` from the I/O loop.

    Returns an opaque handle that may be passed to
    `remove_timeout` to cancel.

    ``deadline`` may be a number denoting a time (on the same
    scale as `IOLoop.time`, normally `time.time`), or a
    `datetime.timedelta` object for a deadline relative to the
    current time.

    Note that it is not safe to call `add_timeout` from other threads.
    Instead, you must use `add_callback` to transfer control to the
    `IOLoop`'s thread, and then call `add_timeout` from there.
    """
    func = coro_engine(callback)
    return io_loop().add_timeout(deadline, func)


def coro_engine(coro):
    """Wrap coroutine to make it a engine. An engine
    will log exceptions when added as timeout, because
    it does't return a future."""
    # If tornado ever drops support of engine, convert this
    # to a coroutine that catches, logs and reraises any exception
    # because thats the goal of this method.
    assert isinstance(coro, (types.FunctionType, types.MethodType)), \
        "coro_engine needs a function or method!"

    @gen.engine
    def engine_func(*args, **kwargs):
        """Engine wrapper"""
        yield coro(*args, **kwargs)
    return engine_func


def io_loop():
    """ Get the io loop, in way jedi understands.
    :rtype: tornado.ioloop.IOLoop
    """
    return ioloop.IOLoop.instance()


def get_settings():
    """Get the s module"""
    from django.conf import settings
    return settings


def local_datetime(date_time):
    """
    Makes a datetime django compatible.

    Used if we have to enable timezone in future.

    :rtype: datetime.datetime
    """
    return date_time.replace(tzinfo=None)
    # Enable this if we use timezones
    # import pytz.reference
    # return dt.replace(tzinfo=pytz.reference.LocalTimezone())


def now():
    """
    Return now with tzinfo

    :rtype: datetime.datetime
    """
    return local_datetime(
        datetime.datetime.now()
    )


def consts_to_dict(object_):
    """Converts a constants object to a dictionary"""
    new = {}
    for const in dir(object_):
        if not const.startswith("_"):
            new[getattr(object_, const)] = const
    return new


class LogDataLockMixin(object):
    """An object that locks logdata"""
    # pylint: disable=too-few-public-methods
    def __init__(self, *args, **kwargs):
        super(LogDataLockMixin, self).__init__(*args, **kwargs)
        self._logdata = None
        self._lock_obj = None

    @property
    def logdata(self):  # pragma: no cover
        """:rtype: symonitoring_alert.logdata.LogData"""
        return self._logdata

    @logdata.setter
    def logdata(self, val):  # pragma: no cover
        """:type val: symonitoring_alert.logdata.LogData"""
        self._logdata    = val
        self._lock_obj   = val

    @property
    def lock(self):  # pragma: no cover
        """:rtype: toro.Lock"""
        if self._lock_obj:
            return self._lock_obj.lock
        else:
            return self._lock


class JoinableThreadPoolExecutor(futures.ThreadPoolExecutor):
    """A thread pool executor that can be joined"""
    def join(self):
        """Joins the executor queue. When called on the main thread this means
        no tasks are executing"""
        # There is no neat way to do this since the ThreadPoolExecutor doesn't
        # use task_done
        while not self._work_queue.empty():  # pragma: no cover
            time.sleep(0.1)


class ExecutorMixin(object):
    """Add io_loop and executor to a class"""
    # pylint: disable=too-few-public-methods
    executor = None

    @classmethod
    def join(cls):
        """Joins the executor queue. When called on the main thread this means
        no tasks are executing"""
        return ExecutorMixin.executor.join()

    def __init__(self, *args, **kwargs):
        super(ExecutorMixin, self).__init__(*args, **kwargs)
        self.io_loop = io_loop()
        if not ExecutorMixin.executor:
            ExecutorMixin.executor = JoinableThreadPoolExecutor(
                s.WORKERS
            )
        self.executor    = ExecutorMixin.executor
        self._lock       = toro.Lock()
        self._lock_obj   = None

    @property
    def lock(self):
        """:rtype: toro.Lock"""
        if self._lock_obj:
            return self._lock_obj.lock
        else:
            return self._lock


def wait(deadline):
    """Wait for a deadline

    :rtype: tornado.concurrent.Future"""
    fut = tc.Future()

    def timeout():
        """Helper"""
        fut.set_result(None)

    io_loop().add_timeout(
        deadline,
        timeout
    )
    # Pending on tornado version we have to start the future
    try:
        # pylint: disable=no-member
        fut.set_running_or_notify_cancel()
    except AttributeError:
        pass
    return fut


class TaskError(Exception):
    """An error occoured in task"""
    pass


class TaskFuture(tc.Future):
    """Converts a async call with callback to a Future"""
    _error = object()

    @staticmethod
    def check_error(res):
        """Check if result is an error"""
        try:
            if len(res) > 1:
                if res[0] is TaskFuture._error:
                    raise TaskError(res)
        except TypeError:
            pass

    def __init__(self, task, *args, **kwargs):
        super(TaskFuture, self).__init__()
        self.task = task
        self.args = args
        self.kwargs = kwargs
        self.timeout = 10
        run = True
        if "run" in kwargs:
            if not kwargs["run"]:
                run = False
            del kwargs["run"]
        if run:  # pragma: no cover
            self.run_task()

    def timeout_callback(self):  # pragma: no cover
        """Callback for timeout"""
        if self.running():
            self.set_result((TaskFuture._error, "timeout"))

    def error_callback(self, *args):
        """Callback for errors"""
        if self.running():  # pragma: no cover
            if len(args) == 1:
                self.set_result((TaskFuture._error, args[0]))
            else:
                self.set_result((TaskFuture._error, args))

    def result_callback(self, *args):
        """Callback for result"""
        if self.running():  # pragma: no cover
            if len(args) == 1:
                self.set_result(args[0])
            else:
                self.set_result(args)

    def run_task(self):
        """Run the task"""
        assert "callback" not in self.kwargs
        self.kwargs['callback'] = self.result_callback
        if self.timeout:
            io_loop().add_timeout(
                datetime.timedelta(seconds=self.timeout),
                self.timeout_callback
            )
        # Pending on tornado version we have to start the future
        try:
            # pylint: disable=no-member
            self.set_running_or_notify_cancel()
        except AttributeError:
            pass
        self.task(*self.args, **self.kwargs)


if six.PY2:  # pragma: no cover
    # pylint: disable=redefined-builtin
    def bytes(arg):
        """bytes for python 2 is passthrough"""
        return arg
else:  # pragma: no cover
    import builtins

    def bytes(arg):
        """Converting to bytes using UTF-8"""
        return builtins.bytes(arg, encoding="UTF-8")


run_on_executor = tc.run_on_executor
coroutine = gen.coroutine
Return = gen.Return
Task = gen.Task


def get_objs_of_type(type_):
    """Returns objects of a certain type"""
    import gc
    return [x for x in gc.get_objects() if isinstance(x, type_)]


def get_type_count(type_):
    """Returns the count of object of a certain type"""
    return len(get_objs_of_type(type_))


def pympler(type_, browse=False):  # pragma: no cover
    """Display references to objects of a type"""
    from pympler import muppy
    from pympler import summary
    import gc
    all_objects = muppy.get_objects()
    sum1 = summary.summarize(all_objects)
    summary.print_(sum1)

    try:
        objs = muppy.filter(all_objects, Type=type_)
        print("Count: %s" % len(objs))
        fil = objs[0]
    except IndexError:
        return

    if not browse:
        return

    def printobj(string, obj):
        """Helper"""
        print(string % type(obj))

    for level1 in gc.get_referrers(fil):
        printobj("%s", level1)
        for level2 in gc.get_referrers(level1):
            printobj("\t%s", level2)
            for level3 in gc.get_referrers(level2):
                printobj("\t\t%s", level3)
                for level4 in gc.get_referrers(level3):
                    printobj("\t\t\t%s", level4)
                    for level5 in gc.get_referrers(level4):
                        printobj("\t\t\t\t%s", level5)


def builtin(name):  # pragma: no cover
    """Get right builtin depending on python2 or python3"""
    if six.PY2:
        return "__builtin__.%s" % name
    else:
        return "builtins.%s" % name


def memoize(obj):
    """Caching results of method of object"""

    def memoizer(self, *args, **kwargs):
        """Memoize helper"""
        if not hasattr(self, "memoize_cache"):
            self.memoize_cache = {}
        cache = self.memoize_cache
        key = (obj, args, frozenset(kwargs))
        if key not in cache:
            cache[key] = obj(self, *args, **kwargs)
        return cache[key]
    return memoizer


@contextlib.contextmanager
def disable_dummy_mode():
    """Disables dummy mode"""
    old = s.DUMMY_MODE
    s.DUMMY_MODE = False
    yield
    s.DUMMY_MODE = old


@contextlib.contextmanager
def _allow_block_for_test_only():
    """Allow blocking io during a test

    ONLY IS THIS IN UNITTESTS for code that is also covered
    without this!"""

    state = nb.disabled
    nb.disabled = True
    yield
    nb.disabled = state


@contextlib.contextmanager
def _allow_add_handler_test_only():
    """ Allow adding handlers using add_timeout and add_callback.

    ONLY IS THIS IN UNITTESTS ! MAKE SURE THE HANDLERS ARE EXECUTED!

    -> use cmn.wait()"""

    TestCase.remove_add_handler_block()
    yield
    TestCase.install_add_handler_block()


def make_datetime_from_day_time(
        day,
        time
):  # pylint: disable=redefined-outer-name
    """Returns a dateime from a day and a time"""
    return local_datetime(datetime.datetime(
        day.year,
        day.month,
        day.day,
        time.hour,
        time.minute,
        time.second,
    ))


def round_to_two_seconds(delta):
    """ Round timedelta to two seconds """
    return datetime.timedelta(
        seconds=round(delta.total_seconds() / 2.0) * 2
    )


def is_flag_set(flag):
    """ Returns true if the given flag is set."""

    return flag in os.environ


class HandlerInstalledException(Exception):
    """Exception when method to install handles is called while these are
    blocked"""
    pass


@contextlib.contextmanager
def timeout_signal(seconds):
    """ Raise IOError after a timeout """
    def timeout_handler(signum, frame):  # pylint: disable=unused-argument
        """ Helper """
        pass

    original_handler = signal.signal(signal.SIGALRM, timeout_handler)

    try:
        signal.alarm(seconds)
        yield
    finally:
        signal.alarm(0)
        signal.signal(signal.SIGALRM, original_handler)


@six.add_metaclass(ABCMeta)
class TestCase(ExecutorMixin, unittest.TestCase):
    """symonitoring_alert specific unitest class enforces exception on blocking
    code and blocks the call of add_timeout and add_callback"""

    test_data_installed = False
    fake_smtp_server = None
    sqlite_performance_set = False
    unittest_active = False

    def __init__(self, *args, **kwargs):
        import symonitoring_alert.lock as lock
        lock.lock()
        super(TestCase, self).__init__(*args, **kwargs)

    @classmethod
    def set_sqlite_performance(cls):
        """ Make sqlite high performance for tests """
        if not TestCase.sqlite_performance_set:
            TestCase.sqlite_performance_set = True
            from django.db.backends.signals import connection_created

            def activate_performance(
                    sender,
                    connection,
                    **kwargs
            ):  # pylint: disable=unused-argument
                """ Enable integrity constraint with sqlite. """
                if connection.vendor == 'sqlite':
                    cursor = connection.cursor()
                    cursor.execute('PRAGMA synchronous = OFF;')
                    cursor.execute('PRAGMA journal_mode = MEMORY;')

            for connection in db.connections.all():
                activate_performance(cls, connection)

            connection_created.connect(activate_performance)

    @run_on_executor
    def save_log_ex(self, log):
        """ Runs save_log() on executor."""
        self.save_log(log)

    @classmethod
    def save_log(cls, log):
        """ Saving NagiosLog entry """
        with _allow_block_for_test_only():
            log.save()

    @classmethod
    def install_add_handler_block(cls):
        """Installs a block for calling add_*"""
        # pylint: disable=global-variable-undefined,global-statement
        global _add_block_installed
        global _add_callback_orig
        global _add_timeout_orig
        global add_callback
        global add_timeout

        def raise_func(*args, **kwargs):
            """Raises an exception when add_* is called"""
            # pylint: disable=unused-argument
            raise HandlerInstalledException()
        if not _add_block_installed:
            _add_callback_orig   = add_callback
            add_callback         = raise_func
            _add_timeout_orig    = add_timeout
            add_timeout          = raise_func
            _add_block_installed = True

    @classmethod
    def remove_add_handler_block(cls):
        """Installs a block for calling add_*"""
        # pylint: disable=global-statement,global-variable-undefined
        global _add_block_installed
        global add_callback
        global add_timeout
        if _add_block_installed:
            add_callback = _add_callback_orig
            add_timeout = _add_timeout_orig
            _add_block_installed = False

    @classmethod
    def start_fake_smtp_server(cls):
        """Starts a fake debug server"""
        if not TestCase.fake_smtp_server:
            import smtpd
            import asyncore
            import threading

            class FakeSMTPServer(smtpd.SMTPServer):
                """A Fake smtp server"""

                def process_message(
                        self, peer, mailfrom, rcpttos, data
                ):  # pragma: no cover
                    """ Do nothing """
                    inheaders = 1
                    lines = data.split('\n')
                    out = []
                    out.append('---------- MESSAGE FOLLOWS ----------')
                    for line in lines:
                        # headers first
                        if inheaders and not line:
                            out.append('X-Peer: %s' % peer[0])
                            inheaders = 0
                        out.append(line)
                    out.append('------------ END MESSAGE ------------')
                    lg.debug(out)
            TestCase.fake_smtp_server = FakeSMTPServer(
                ('127.0.0.1', s.SMTP_PORT), None
            )
            thread = threading.Thread(
                target=asyncore.loop,
            )
            thread.daemon = True
            thread.start()

    @classmethod
    def insert_test_data(cls):
        """ Read the testdata and insert it into the database """
        if not TestCase.test_data_installed:
            from django.core.management import call_command
            with _allow_block_for_test_only():
                call_command(
                    "loaddata",
                    "test_data.json",
                    verbosity=0
                )
            TestCase.test_data_installed = True

    def setUp(self):
        TestCase.unittest_active = True
        self.set_sqlite_performance()
        self.start_fake_smtp_server()
        self.insert_test_data()
        self.install_add_handler_block()
        self.has_blocked = False

        def blocksignal(signal, frame):  # pylint: disable=redefined-outer-name
            """ Helper to find blocking code sets a flag on the UnitTest
            object"""
            self.has_blocked = True
            io_loop().log_stack(signal, frame)

        @gen.coroutine
        def coro():
            """ Helper to delete online data """
            yield self.delete_online_data_ex()

        io_loop().set_blocking_signal_threshold(
            0.05 * s.TIMING_TEST_SCALE, blocksignal
        )
        run_sync(coro)
        self.sy_set_up()

    def tearDown(self):
        try:
            self.sy_tear_down()

            @gen.coroutine
            def coro():
                """ Helper to delete online data """
                yield self.delete_online_data_ex()
            run_sync(coro)
            self.assertFalse(self.has_blocked)
            self.remove_add_handler_block()
        finally:
            TestCase.unittest_active = False

    def sy_set_up(self):
        """ Set up method of unitests"""
        pass

    def sy_tear_down(self):
        """ Tear down method of unitests"""
        pass

    @run_on_executor
    def delete_online_data_ex(self):
        """ Helper to delete online data """
        return self.delete_online_data()

    @classmethod
    def delete_online_data(cls):
        """ Helper to delete online data """
        import symonitoring_model.models as mod
        mod.Alert.objects.all().delete()
        mod.AlertTimeframeState.objects.all().delete()
        mod.Hold.objects.all().delete()
        mod.HoldInform.objects.all().delete()
        mod.Nagioslog.objects.all().delete()
        mod.Downtime.objects.all().delete()

s = get_settings()
_lg = s.LG
assert isinstance(_lg, logging.Logger)


# Helper for pylint and jedi to recognize logger
def ret_lg():
    """:rtype: logging.Logger"""
    return _lg

lg = _lg
