# -*- coding: UTF-8 -*-

"""
Communication (sending messages)
"""

# System imports
import tornado
import json
import uuid
import socket
import datetime
import smtplib
import os.path
import six
from abc import ABCMeta, abstractmethod
import email.utils                as eu
import email.mime.text            as et
import email.mime.multipart       as em
import tornado.httpclient         as client
import tornado.iostream           as iostream
try:
    import xml.etree.cElementTree as xtr
except ImportError:  # pragma: no cover
    import xml.etree.ElementTree  as xtr


# Project imports
import symonitoring_model.models             as mod
import symonitoring_alert.common             as cmn
import symonitoring_alert.constants          as const
import symonitoring_alert.default_templates  as deftemp
from . import xmpp

s = cmn.get_settings()

_aspsms_envelope = """<?xml version="1.0" encoding="ISO-8859-1"?>
<aspsms>
    <Userkey>{userkey}</Userkey>
    <Password>{password}</Password>
    <Originator>{originator}</Originator>
    <Recipient>
        <PhoneNumber>{recipient}</PhoneNumber>
    </Recipient>
    <MessageData>{text}</MessageData>
    <Action>SendTextSMS</Action>
</aspsms>
"""


@six.add_metaclass(ABCMeta)
class UserCommunicationBase(cmn.ExecutorMixin):
    """Methods to communicate (alert) with a user"""

    __slots__ = (
        '_template',
        '_logdata',
        '_user',
        '_type',
    )

    @property
    def template(self):
        """:rtype: list(string)"""
        return self._template

    @property
    def logdata(self):
        """:rtype: symonitoring_alert.logdata.LogData"""
        return self._logdata

    @property
    def user(self):
        """:rtype: symonitoring_model.models.User"""
        return self._user

    @property
    def type_(self):
        """:rtype: int"""
        return self._type

    def __init__(
            self,
            logdata,
            user,
            template=None,
            type_=const.TemplateType.ALERT,
    ):
        # pylint: disable=too-many-arguments
        super(UserCommunicationBase, self).__init__()
        self._template  = template
        self._logdata   = logdata
        self._user      = user
        self._type      = type_

    @abstractmethod
    def render_template(self, part=None, template=None, is_action=None):
        """Render template"""
        raise NotImplementedError()

    @classmethod
    def report_error(cls, message):
        """Send a plain text email with an error message"""
        try:
            recipient = mod.Setting.objects.filter(
                key='erroremail'
            )[0].value
            if recipient == '':
                raise ValueError("eMail address is empty")
        except (AttributeError, ValueError, IndexError):
            cmn.lg.exception("Error email address not defined in global "
                             "settings")
            return
        coded = et.MIMEText(message.encode('utf-8'), _charset = 'utf-8')
        coded['Subject'] = 'Notification error in symonitoring_alert'
        coded['From']    = s.SENDER
        coded['To']      = recipient
        coded['Date']    = eu.formatdate()
        try:
            mail = smtplib.SMTP(s.SMTP_HOST, s.SMTP_PORT)
            mail.sendmail(s.SENDER, recipient, coded.as_string())
            mail.quit()
        except smtplib.SMTPResponseException:
            cmn.lg.exception("Sending mail to %s failed", recipient)
            return False

    @cmn.run_on_executor
    def report_error_ex(self, message):
        """ Run report_error on executor """
        return self.report_error(message)

    def send_email(self):
        """Send an email according to templates"""
        excluded = self.user.alertnotificationexcludeuser_set.filter(
            slaactionid = const.ActionType.EMAIL,
            customerid  = self.logdata.host.customerid,
        ).count()
        if excluded > 0:
            cmn.lg.debug(
                "Not sending mail to %s (disabled by user)",
                self.user.email,
            )
            return
        cmn.lg.debug("Sending mail to: %s", self.user.email)
        plain   = self.render_template(const.TemplatePart.PART_PLAIN)
        subject = self.render_template(const.TemplatePart.PART_SUBJECT)
        html    = self.render_template(const.TemplatePart.PART_HTML)
        if html:
            html = "<html><head>%s</head><body>%s</body></html>" % (
                subject,
                html
            )
            coded = em.MIMEMultipart('alternative')
            if plain:
                coded.attach(
                    et.MIMEText(
                        plain.encode('utf-8'), "plain", _charset = 'utf-8'
                    )
                )
            coded.attach(
                et.MIMEText(html.encode('utf-8'), "html", _charset = 'utf-8')
            )
        else:  # pragma: no cover
            coded = et.MIMEText(plain.encode('utf-8'), _charset = 'utf-8')
        coded['Subject'] = subject
        coded['From']    = s.SENDER
        coded['To']      = self.user.email
        coded['Date']    = eu.formatdate()
        try:
            mail = smtplib.SMTP(s.SMTP_HOST, s.SMTP_PORT)
            mail.sendmail(s.SENDER, self.user.email, coded.as_string())
            mail.quit()
        except smtplib.SMTPResponseException:
            message = "Sending mail to %s failed" % self.user.email
            self.report_error(message)
            cmn.lg.exception(message)
            return False
        return True

    @cmn.coroutine
    def send_xmpp(self, orig_action=None):
        """Send XMPP message

        :type xmpp: sleekxmpp.ClientXMPP"""
        cmn.lg.debug("Sending xmmp to: %s", self.user.jabber)
        is_action = ""
        if orig_action:
            is_action = "%s: " % const.action_type_dict[orig_action]
        plain   = self.render_template(
            const.TemplatePart.PART_PLAIN, is_action=is_action
        )
        subject = self.render_template(
            const.TemplatePart.PART_SUBJECT, is_action=is_action
        )
        html    = self.render_template(
            const.TemplatePart.PART_HTML, is_action=is_action
        )
        to      = self.user.jabber
        if html:
            res = yield xmpp.send_message(
                mto=to,
                mbody=plain,
                msubject=subject,
                mtype="chat",
                mhtml=html,
            )
        else:
            res = yield xmpp.send_message(
                mto=to,
                mbody=plain,
                msubject=subject,
                mtype="chat",
            )
        raise cmn.Return(res[0] == "ok")

    @cmn.coroutine
    def send_mattermost(self):
        cmn.lg.debug("Sending mattermost message to: %s", self.user.mattermost)

        plain = self.render_template(const.TemplatePart.PART_PLAIN)
        http = tornado.httpclient.AsyncHTTPClient()

        data = {
            'text': plain,
            # for private message channel is @username
            'channel': "@%s" % self.user.mattermost,
            # username message is sent from
            'username': s.WEBHOOK['username'],
        }
        req  = tornado.httpclient.HTTPRequest(
            url     = s.WEBHOOK['url'],
            method  = "POST",
            headers = {'Content-Type': 'application/json'},
            body    = json.dumps(data),
        )
        res  = yield http.fetch(req)
        if res.code is not 200:  # pragma: no cover
            msg = "Sending mattermost message to %s with user %s failed" % (
                s.WEBHOOK['url'], self.user.mattermost, res.code
            )
            self.report_error(msg)
            cmn.lg.exception(msg)
            res = False

        raise cmn.Return(res.code == 200)

    @cmn.coroutine
    def send_sms(self):
        """Send an SMS according to config"""
        if s.ASPSMS:
            res = yield self.send_sms_asp()  # pragma: no cover
        else:
            res = yield self.send_sms_gsm()
        raise cmn.Return(res)

    @cmn.coroutine
    def send_sms_gsm(self):
        """Send SMS via GSM module"""
        path = s.GSM_SPOOL % uuid.uuid4()
        res = yield self.send_sms_gsm_send(path)
        if not res:
            raise cmn.Return(res)
        yield cmn.wait(datetime.timedelta(seconds=10))
        res = yield self.send_sms_gsm_check(path)
        raise cmn.Return(res)

    @cmn.run_on_executor
    def send_sms_gsm_send(self, path):
        """Send SMS via GSM module"""
        cmn.lg.debug("Sending gsm sms to: %s", self.user.sms)
        text = self.render_template()
        envelope = """To: {number}
Flash: 0

{text}"""
        msg = envelope.format(
            number=self.user.sms,
            text=text,
        )
        if six.PY2:
            except_cls = IOError
        else:
            except_cls = OSError
        try:
            fh = open(path, "w")
            fh.write(msg)
            fh.close()
        except except_cls:
            message = "Sending sms to %s failed" % self.user.sms
            self.report_error(message)
            cmn.lg.exception(message)
            raise cmn.Return(False)
        raise cmn.Return(True)

    @cmn.run_on_executor
    def send_sms_gsm_check(self, path):
        """Send SMS via GSM module"""
        return not os.path.exists(path)

    @cmn.coroutine
    def send_sms_asp(self):
        """Send SMS via aspsms"""
        cmn.lg.debug("Sending asp sms to: %s", self.user.sms)
        text = self.render_template()
        http = client.AsyncHTTPClient()
        req  = client.HTTPRequest(
            url    = s.ASPSMS_URL,
            method = "POST",
            body   = _aspsms_envelope.format(
                userkey=s.ASPSMS_USERKEY,
                password=s.ASPSMS_PASSWORD,
                originator=s.ASPSMS_ORIGINATOR,
                recipient=self.user.sms,
                text=text,
            ),
        )
        res  = yield http.fetch(req)
        if res.code is not 200:  # pragma: no cover
            message = "Sending sms to %s failed" % self.user.sms
            yield self.report_error_ex(message)
            raise cmn.Return(False)
        else:
            status = "Nok"
            try:
                x = xtr.fromstring(res.body.strip())
                status = list(x.iter(
                    "ErrorDescription"
                ))[0].text
            except (IndexError, ValueError, xtr.ParseError):
                message = "Sending aspsms to %s failed", self.user.sms
                yield self.report_error_ex(message)
                cmn.lg.exception(message)
            raise cmn.Return(status.lower() == "ok")

    @cmn.coroutine
    def call_phone(self):
        """Call phone number"""
        cmn.lg.debug("Calling: %s", self.user.phone)
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM, 0)
        stream = iostream.IOStream(sock)

        try:
            future = cmn.TaskFuture(
                stream.connect,
                (s.PBX_IP, s.PBX_PORT),
                run=False,
            )
            stream.set_close_callback(future.error_callback)
            future.run_task()
            res = yield future
            cmn.TaskFuture.check_error(res)
            res = yield cmn.TaskFuture(
                stream.read_until_regex,
                cmn.bytes(
                    (
                        "{banner}\r\n"
                    ).format(
                        banner=s.PBX_BANNER
                    )
                )
            )
            cmn.TaskFuture.check_error(res)
            res = yield cmn.TaskFuture(
                stream.write,
                cmn.bytes(
                    (
                        "action: login\r\n"
                        "username: {username}\r\n"
                        "secret: {secret}\r\n"
                        "events: off\r\n\r\n"
                    ).format(
                        username=s.PBX_USER,
                        secret=s.PBX_PW
                    )
                )
            )
            cmn.TaskFuture.check_error(res)
            res = yield cmn.TaskFuture(
                stream.read_until_regex,
                cmn.bytes("Message: Authentication accepted\r\n")
            )
            cmn.TaskFuture.check_error(res)
            res = yield cmn.TaskFuture(
                stream.write,
                cmn.bytes(
                    (
                        "action: originate\r\n"
                        "channel: {channel}/{phone}\r\n"
                        "exten: {dest}\r\n"
                        "context: {context}\r\n"
                        "callerid: {callerid}\r\n"
                        "priority: 1\r\n\r\n"
                    ).format(
                        phone=self.user.phone,
                        dest=s.PBX_DEST,
                        context=s.PBX_CONTEXT,
                        channel=s.PBX_CHANNEL,
                        callerid=s.PBX_CALLERID
                    )
                )
            )
            cmn.TaskFuture.check_error(res)
            res = yield cmn.TaskFuture(
                stream.read_until_regex,
                cmn.bytes("\r\n")
            )
            cmn.TaskFuture.check_error(res)
            res = yield cmn.TaskFuture(
                stream.write,
                cmn.bytes("action: logoff\n\n"),
            )
            stream.close()
        except cmn.TaskError:
            message = "Calling %s failed", self.user.phone
            yield self.report_error_ex(message)
            cmn.lg.exception(message)
            raise cmn.Return(False)
        raise cmn.Return(True)

    @classmethod
    def get_template(cls, user, action_type):
        """Gets the template for a user.

        :rtype: string"""
        ref     = cls.get_template_ref(user, action_type)
        default = deftemp.templates[action_type]
        if ref:  # pragma: no cover
            assert isinstance(ref, mod.AlerttemplateUser)
            return (ref.templateid.content, default)
        return (default, default)

    @classmethod
    def get_template_ref(cls, user, action_type):
        """Gets the template for a user.

        :rtype: symonitoring_model.models.AlerttemplateUser
        """
        assert isinstance(user, mod.User)
        template = None
        try:
            template = mod.AlerttemplateUser.objects.filter(
                userid      = user,
                slaactionid = action_type,
            ).select_related()[0]
        except IndexError:
            pass
        if template:  # pragma: no cover
            return template
        try:
            template = mod.AlerttemplateUser.objects.filter(
                userid__isnull     = True,
                slaactionid = action_type
            ).select_related()[0]
        except IndexError:
            pass
        return template
