# -*- coding: UTF-8 -*-

"""
XMPP Client in a different process
"""

# System imports
import multiprocessing
try:
    import Queue          as queue
except ImportError:
    import queue
import datetime
import tornado.concurrent as tc
import time

# Project import
import symonitoring_alert.common     as cmn

s = cmn.get_settings()
main_called = False

_inq             = multiprocessing.Queue()  # pylint: disable=no-member
_outq            = multiprocessing.Queue()  # pylint: disable=no-member
_active          = {}
_timeout_seconds = 10
_process         = None
_hearthbeat      = cmn.now()


def _response(fd, events):  # pragma: no cover
    """Callback for results from XMPP"""
    # pylint: disable=unused-argument,global-statement
    global _hearthbeat
    try:
        while True:
            resp = _outq.get_nowait()
            if resp[0] == "hb":
                _hearthbeat = cmn.now()
            else:
                idstr = resp[1]
                if idstr in _active:
                    _active[idstr][4].set_result(resp)
                    del _active[idstr]
    except queue.Empty:
        pass

cmn.io_loop().add_handler(
    _outq._reader.fileno(),  # pylint: disable=protected-access
    _response, cmn.ioloop.IOLoop.READ
)


def check_start_xmpp():
    """Checks if the xmpp bot is running, and starts it if not"""
    cmn.lg.debug("Checking xmpp bot")
    global _process     # pylint: disable=global-statement
    global _hearthbeat  # pylint: disable=global-statement
    start = False
    if not _process:
        cmn.lg.debug("sleekxmpp is not yet started")
        start = True
    elif cmn.now() - _hearthbeat > datetime.timedelta(
            seconds=s.XMPP_CHECK_RATE / 1000
    ):
        # Don't restart more often than XMPP_CHECK_RATE
        if not _process.is_alive:
            cmn.lg.debug("sleekxmpp is not alive")
        else:
            cmn.lg.debug("sleekxmpp heartbeat not received")
        start = True

    if start:
        _hearthbeat = cmn.now()
        if _process:
            _process.terminate()
            cmn.lg.critical("sleekxmpp process died!")
        # No threading is allowed during fork, therefore we join the executor
        cmn.ExecutorMixin.executor.join()
        _process = multiprocessing.Process(
            target=_run,
            args=(_inq, _outq)
        )
        _process.daemon = True
        _process.start()
        # Wait for process to start, or we may get a starve-lock
        for _ in range(10):
            if _process.is_alive:
                break
            time.sleep(0.1)


def send_message(*args, **kwargs):
    """Send a command to sleekxmpp"""
    check_start_xmpp()
    idobj = object()
    idstr = str(id(idobj))
    fut = tc.Future()
    # Depending on Tornado version we have to start the future
    try:
        fut.set_running_or_notify_cancel()  # pylint: disable=no-member
    except AttributeError:
        pass

    def _timeout():  # pylint: disable=redefined-outer-name
        """Timeout for results from XMPP"""
        # A timeout means the message is delivered!
        # Thats how the XMPP protocol works:
        # You get an error -> a problem
        # You get nothing -> probably everything ok
        # Don try to fix this, I tried long enough
        try:
            try:
                if fut.running():
                    if idstr in _active:
                        res = _active[idstr]
                        if res[0] == "command":
                            fut.set_result(("ok",))
                    else:
                        fut.set_result(("request not found",))
            finally:
                del _active[idstr]
        except KeyError:
            pass

    cmd = ("command", idstr, args, kwargs)
    _active[idstr] = cmd + (fut, )
    _inq.put(cmd)
    cmn.io_loop().add_timeout(
        datetime.timedelta(seconds=_timeout_seconds),
        _timeout
    )
    return fut


def _run(inq, outq):  # pragma: no cover
    """Runner for the sleekxmpp process

    :type inq: multiprocessing.Queue
    :type outq: multiprocessing.Queue"""
    # We actually test this but coverage doesn't work with multiprocessing

    import sleekxmpp
    import setproctitle

    setproctitle.setproctitle("syalert-xmpp")  # pylint: disable=no-member

    class MessageBot(sleekxmpp.ClientXMPP):
        """Message bot sending messages read from a queue"""

        def __init__(self, jid, password):
            """Added main handlers"""
            super(MessageBot, self).__init__(jid, password)

            self.add_event_handler("session_start", self.session_start)
            self.add_event_handler("message", self.message)

        def session_start(self, _):
            """Complete login"""
            self.send_presence()
            self.get_roster()

        def message(self, msg):
            """Message handler"""
            cmn.lg.debug("Received message: %s", str(dict(msg)))
            if "id" in msg.keys() and "error" in msg.keys():
                outq.put(("error", msg["id"]))

    bot = MessageBot(s.XMPP_USER, s.XMPP_PASSWORD)
    if hasattr(s, "XMPP_ADDRESS"):
        bot.connect((s.XMPP_ADDRESS, 5222))
    else:
        bot.connect()
    bot.process(block=False)
    while True:
        try:
            cmd     = inq.get(timeout=s.XMPP_CHECK_RATE / 1000 / 2)
            idstr   = cmd[1]
            args    = cmd[2]
            kwargs  = cmd[3]
            m       = bot.make_message(*args, **kwargs)
            m["id"] = idstr
            m.send()
        except queue.Empty:
            pass
        except Exception as e:  # pylint: disable=broad-except
            cmn.lg.exception("Sending message failed id: %s", idstr)
            outq.put(("error", idstr, e))
        outq.put(("hb",))

if __name__ == "__main__":
    _run(_inq, _outq)
