# -*- coding: UTF-8 -*-

"""
Testing communication
"""

# System imports
import mock
import datetime
import os

# Project imports
import symonitoring_alert.common               as cmn
import symonitoring_alert.communication.alert  as sca
import symonitoring_alert.handle_test          as htest
import symonitoring_alert.constants            as const
import symonitoring_alert.communication.xmpp   as xmpp

# pylint: disable=missing-docstring

asp_fail = """
<?xml version="1.0"?>
<aspsms>
<ErrorCode>26</ErrorCode>
<ErrorDescription>IP-Address not Authorized for this Operation
</ErrorDescription>
</aspsms>
"""

asp_ok = """
<?xml version="1.0"?>
<aspsms>
<ErrorCode>1</ErrorCode>
<ErrorDescription>Ok</ErrorDescription>
</aspsms>
"""


def _skip_xmpp():
    return not (
        os.path.exists("/etc/ejabberd/ejabberd.cfg") and
        os.path.exists("/vagrant/alert/setup.py")
    )


class CommunicationTest(cmn.TestCase):
    """Testing communication modules: gsm, phone, jabber, asp"""
    # pylint: disable=no-self-use

    def get_comm(self, atype=const.ActionType.SMS):
        """Getting templates from the DB and creating UserCommunication
        object"""
        with cmn._allow_block_for_test_only():
            handler = htest.HandleTest.get_handler_with_logdata()
            self.assertIsNotNone(handler.logdata)
            timeframe = handler.logdata.timeframes[0]
            user      = handler.logdata.users[timeframe][0]
            template  = sca.UserCommunicationAlert.get_template(
                user,
                atype,
            )
            return sca.UserCommunicationAlert(
                handler.logdata,
                user,
                timeframe,
                template,
            )

    def test_sms_via_gsm(self):
        comm = self.get_comm()
        nwait = cmn.wait(datetime.timedelta(seconds=0))
        with cmn.disable_dummy_mode():
            with mock.patch(cmn.builtin("open")) as op:
                with mock.patch("symonitoring_alert.common.wait") as wait:
                    wait.return_value = nwait
                    res = cmn.run_sync(comm.send_sms_gsm)
                    self.assertTrue(op.called)
                    self.assertEqual(op.call_args[0][1], "w")
                    self.assertEqual(
                        op.return_value.write.call_args[0][0],
                        """To: 0790000000
Flash: 0

[CRITICAL] testhost22: ssh (tcp):
Connection refused"""
                    )
                    self.assertTrue(res)

    def test_phone(self):
        comm = self.get_comm()

        def noop(*args, **kwargs):
            # pylint: disable=unused-argument
            pass
        chkerr = cmn.TaskFuture.check_error
        fut = cmn.TaskFuture(noop, run=False)
        fut.set_result(None)
        with mock.patch("symonitoring_alert.common.TaskFuture") as tfut:
            tfut.return_value = fut
            tfut.check_error = chkerr
            res = cmn.run_sync(comm.call_phone)
            self.assertTrue(res)

    def test_phone_err(self):
        comm = self.get_comm()

        def noop(*args, **kwargs):
            # pylint: disable=unused-argument
            pass
        chkerr = cmn.TaskFuture.check_error
        fut = cmn.TaskFuture(noop, run=False)
        err = cmn.TaskFuture._error  # pylint: disable=protected-access
        fut.set_result((err, "timeout"))
        with mock.patch("symonitoring_alert.common.TaskFuture") as tfut:
            tfut.return_value = fut
            tfut.check_error = chkerr
            tfut._error = err  # pylint: disable=protected-access
            res = cmn.run_sync(comm.call_phone)
            self.assertFalse(res)

    @cmn.unittest.skipIf(_skip_xmpp(), "no ejabberd installed")
    def test_xmpp(self):
        xmpp.check_start_xmpp()
        comm = self.get_comm(const.ActionType.XMPP)
        timeout = xmpp._timeout_seconds  # pylint: disable=protected-access
        xmpp._timeout_seconds = 0.6  # pylint: disable=protected-access
        try:
            with cmn.disable_dummy_mode():
                res = cmn.run_sync(comm.send_xmpp)
            self.assertTrue(res)
        finally:
            xmpp._timeout = timeout  # pylint: disable=protected-access

    @cmn.unittest.skipIf(_skip_xmpp(), "no ejabberd installed")
    @cmn.unittest.skip("sleekxmpp stopped sending errors TODO FIX THIS!!!!")
    def test_xmpp_fail(self):
        xmpp.check_start_xmpp()
        comm = self.get_comm(const.ActionType.XMPP)
        comm.user.jabber = "asdlkfjalk@localhost"
        timeout = xmpp._timeout_seconds  # pylint: disable=protected-access
        xmpp._timeout_seconds = 0.6  # pylint: disable=protected-access
        try:
            with cmn.disable_dummy_mode():
                res = cmn.run_sync(comm.send_xmpp)
            self.assertFalse(res)
        finally:
            xmpp._timeout = timeout  # pylint: disable=protected-access

    def test_mattermost(self):
        comm = self.get_comm(const.ActionType.MATTERMOST)

        class Res(object):
            def __init__(self):
                self.code = 200

        @cmn.coroutine
        def coro():
            raise cmn.Return(Res())

        with mock.patch("tornado.httpclient.AsyncHTTPClient") as op:
            op.return_value.fetch.return_value = coro()
            res = cmn.run_sync(comm.send_mattermost)
            self.assertTrue(op.return_value.fetch.called)
            self.assertTrue(res)

    def test_sms_via_asp(self):
        comm = self.get_comm()

        class Res(object):
            def __init__(self):
                self.code = 200
                self.body = asp_ok

        @cmn.coroutine
        def coro():
            raise cmn.Return(Res())
        with mock.patch("tornado.httpclient.AsyncHTTPClient") as op:
            # pylint: disable=assignment-from-no-return
            op.return_value.fetch.return_value = coro()
            res = cmn.run_sync(comm.send_sms_asp)
            self.assertTrue(op.return_value.fetch.called)
            self.assertTrue(res)

    def test_sms_via_asp_fail1(self):
        comm = self.get_comm()

        class Res(object):
            def __init__(self):
                self.code = 200
                self.body = """<DeliveryNotification>
<TransRefNumber>1234</TransRefNumber>
<DeliveryStatus>2</DeliveryStatus>
<SubmissionDate>13062001182055</SubmissionDate>
<NotificationDate>13062001182058</NotificationDate>
<ReasonCode>0</ReasonCode>
</DeliveryNotification>"""

        @cmn.coroutine
        def coro():
            raise cmn.Return(Res())
        with mock.patch("tornado.httpclient.AsyncHTTPClient") as op:
            # pylint: disable=assignment-from-no-return
            op.return_value.fetch.return_value = coro()
            res = cmn.run_sync(comm.send_sms_asp)
            self.assertTrue(op.return_value.fetch.called)
            self.assertFalse(res)

    def test_sms_via_asp_fail2(self):
        comm = self.get_comm()

        class Res(object):
            def __init__(self):
                self.code = 200
                self.body = """"""

        @cmn.coroutine
        def coro():
            raise cmn.Return(Res())
        with mock.patch("tornado.httpclient.AsyncHTTPClient") as op:
            # pylint: disable=assignment-from-no-return
            op.return_value.fetch.return_value = coro()
            res = cmn.run_sync(comm.send_sms_asp)
            self.assertTrue(op.return_value.fetch.called)
            self.assertFalse(res)

    def test_sms_via_asp_fail3(self):
        comm = self.get_comm()

        class Res(object):
            def __init__(self):
                self.code = 200
                self.body = asp_fail

        @cmn.coroutine
        def coro():
            raise cmn.Return(Res())
        with mock.patch("tornado.httpclient.AsyncHTTPClient") as op:
            # pylint: disable=assignment-from-no-return
            op.return_value.fetch.return_value = coro()
            res = cmn.run_sync(comm.send_sms_asp)
            self.assertTrue(op.return_value.fetch.called)
            self.assertFalse(res)
