# -*- coding: UTF-8 -*-

"""Test the executor mixin and threading functionality"""

# System imports
import tornado.gen            as gen
import datetime               as dt
import time
try:
    import unittest2          as unittest
except ImportError:  # pragma: no cover
    import unittest

# Project imports
import symonitoring_alert.common         as cmn
import symonitoring_model.models         as mod

# pylint: disable=missing-docstring


class DjangoTornadoHelper(cmn.ExecutorMixin):
    """Helper to test executor"""
    @cmn.run_on_executor
    def get_first_customer(self):
        """Runs on executor"""
        # pylint: disable=no-self-use
        ob = mod.Customer.objects.select_related()[0]
        return ob


class DjangoTornadoBlockHelper(cmn.ExecutorMixin):
    """Helper to test executor"""
    def __init__(self):
        super(DjangoTornadoBlockHelper, self).__init__()
        self.val = None

    @cmn.run_on_executor
    def get_first_customer(self):
        """Runs on executor"""
        time.sleep(0.3 * cmn.s.TIMING_TEST_SCALE)
        self.val = 2
        ob = mod.Customer.objects.select_related()[0]
        return ob

    @cmn.coroutine
    def wrapper(self):
        with (yield self.lock.acquire()):
            yield self.get_first_customer()

    @cmn.coroutine
    def noop(self):
        pass

    @cmn.coroutine
    def has_to_wait(self):
        with (yield self.lock.acquire()):
            yield self.noop()
            raise cmn.Return(self.val)


class ChainedRaise(cmn.ExecutorMixin):
    @cmn.run_on_executor
    def raiser(self):
        # pylint: disable=no-self-use
        raise Exception("chained")

    @cmn.coroutine
    def second(self):
        yield self.raiser()

    @cmn.coroutine
    def first(self):
        yield self.second()


class ExecutorTest(unittest.TestCase):
    """The full tornado monty"""
    def test_basic(self):
        helper = DjangoTornadoHelper()

        @gen.coroutine
        def customer():
            val = yield helper.get_first_customer()
            raise cmn.Return(val)

        val = helper.io_loop.run_sync(customer)
        self.assertEqual(val.name, 'All customers')

    def test_chained(self):
        helper = ChainedRaise()

        @cmn.coroutine
        def coro():
            yield helper.first()
        with self.assertRaises(Exception):
            cmn.run_sync(coro)

    def test_via_loop(self):
        helper = DjangoTornadoHelper()

        vals = []

        @gen.coroutine
        def customer():
            val = yield helper.get_first_customer()
            vals.append(val)
            helper.io_loop.stop()

        cmn.add_timeout(dt.timedelta(
            seconds=0.1
        ), customer)
        helper.io_loop.start()
        self.assertEqual(vals[0].name, 'All customers')

    def test_lock(self):
        helper = DjangoTornadoBlockHelper()

        vals = []

        @gen.coroutine
        def waitfor():
            val = yield helper.has_to_wait()
            vals.append(val)
            helper.io_loop.stop()

        @gen.coroutine
        def lockit():
            vals.append(1)
            yield helper.wrapper()

        cmn.add_timeout(dt.timedelta(
            seconds=0
        ), lockit)
        cmn.add_timeout(dt.timedelta(
            seconds=0.1
        ), waitfor)
        helper.io_loop.start()
        self.assertEqual(vals[0], 1)
        self.assertEqual(vals[1], 2)

    def test_via_loop_block(self):
        helper = DjangoTornadoBlockHelper()

        vals = []

        @gen.coroutine
        def add_vals():
            vals.append(1)

        @gen.coroutine
        def customer():
            vals.append(0)
            val = yield helper.get_first_customer()
            vals.append(val)
            helper.io_loop.stop()

        cmn.add_timeout(dt.timedelta(
            seconds=0.0
        ), customer)
        cmn.add_timeout(dt.timedelta(
            seconds=0.1 * cmn.s.TIMING_TEST_SCALE
        ), add_vals)
        cmn.add_timeout(dt.timedelta(
            seconds=0.2 * cmn.s.TIMING_TEST_SCALE
        ), add_vals)
        cmn.add_timeout(dt.timedelta(
            seconds=0.3 * cmn.s.TIMING_TEST_SCALE
        ), add_vals)
        helper.io_loop.start()
        # Customer is executed first
        self.assertEqual(vals[0], 0)
        # All callbacks are executed
        self.assertEqual(len(vals), 5)
        # Custormer blocks and ends last
        self.assertEqual(vals[-1].name, 'All customers')
        # Meanwhile the others were executed
        self.assertEqual(vals[1], 1)
