# -*- coding: UTF-8 -*-
# pylint: disable=too-many-arguments

"""
Handle alerts
"""


# System imports
import django.db.transaction                    as transaction
import django.db.utils                          as dbutils

# Project imports
import symonitoring_alert.common                as cmn
import symonitoring_model.models                as mod
import symonitoring_alert.constants             as const
import symonitoring_alert.communication.alert   as calert
import symonitoring_alert.dummy                 as dummy
import symonitoring_alert.alert.state           as als

s = cmn.get_settings()
ALERT_DISPLAY_STATES = [
    const.NagiosLog.STATE_NO_SLA,
    const.NagiosLog.STATE_NO_TEAM,
    const.NagiosLog.STATE_WARNING,
]


class TimeoutHandler(object):
    """Class for timeout handler to avoid closures which we had a bug with"""

    def __init__(
            self,
            alert_obj,
            timeframe,
            rule,
            users,
            user
    ):
        self.alert_obj = alert_obj
        self.timeframe = timeframe
        self.rule = rule
        self.users = users
        self.user = user

    @cmn.coroutine
    def handle(self):
        """Timeout handler"""
        # This lock is needed to prevent overlapping timeouts on the same
        # logdata. The lock has to be locked here and in handle of
        # nagioslog.py.
        with (yield self.alert_obj.lock.acquire()):
            logdata = self.alert_obj.logdata
            synced = False

            try:
                yield logdata.sync_ex()
                synced = True
            except (mod.Alert.DoesNotExist, dbutils.IntegrityError):
                logdata.log.processed = (
                    const.NagiosLog.STATE_DOES_NOT_EXIST
                )
                cmn.lg.warning("Alert does not exist anymore "
                               "(it probably was deleted meanwhile)")
                yield self.alert_obj.save_logdata()

            if synced:
                malert = logdata.alert.model
                if not malert.acknowledged:
                    # If alerting is globally disabled, save logdata to
                    # hold and return, _no_ alerting here!
                    if (yield logdata.check_global_disable_ex()):
                        yield logdata.hold.get_or_create_ex()
                        return

                    yield self.alert_obj.schedule_next_timeout(
                        self.timeframe,
                        self.rule,
                        self.users,
                        self.user
                    )
                    yield self.alert_obj.dispatch_action(
                        self.timeframe,
                        self.rule,
                        self.users,
                        self.user
                    )
                    yield logdata.alert_state.save_ex()
                else:
                    cmn.lg.debug(
                        "Alert (ID %d) has been acknowledged in the"
                        " meantime",
                        malert.id
                    )


class AlertInit(cmn.ExecutorMixin, cmn.LogDataLockMixin):
    """Find critical alerts sms/phone/email"""

    def display_alert(self, logdata):
        """Display (save to db) the alert in the GUI"""
        # When we reach this the alert was within a timeframe and we
        # have users
        logdata.alert.create()
        logdata.alert.save()
        logdata.host.hastrouble = True
        cmn.lg.debug(
            "Creating alert for logid %d: %d",
            logdata.log.id,
            logdata.alert.model.id
        )

    @cmn.run_on_executor
    def handle(self):
        """Main handler function"""
        logdata = self.logdata
        cmn.lg.debug("Handling logid %d", logdata.log.id)
        with transaction.atomic():
            if logdata.has_recovered:
                self.handle_has_recovered(logdata)
            else:
                self.handle_alert(logdata)
            if logdata.alert.model:
                if logdata.alert.model.acknowledged:
                    logdata.alert_state.delete()
                else:
                    logdata.alert_state.sync()

    def handle_has_recovered(self, logdata):
        """Handling of recovered logs/alerts"""
        # If alerting is globally disabled, save logdata to
        # hold and return, _no_ alerting here!
        if logdata.check_global_disable():
            logdata.log.processed = (
                const.NagiosLog.STATE_GLOBAL_DISABLE
            )
            logdata.hold.get_or_create()
            return

        # WARNINGs are only the same as a recovery (OK) if there already is an
        # open issue. If not it has to be treated like an alert but without
        # notification.
        if not logdata.alert.model:
            if logdata.is_warning and not logdata.is_soft:
                logdata.log.processed = (
                    const.NagiosLog.STATE_WARNING
                )
                self.display_alert(logdata)
            else:
                logdata.log.processed = (
                    const.NagiosLog.STATE_NO_PREVIOUS_DOWN
                )
            return

        cmn.lg.debug(
            "Handling recovery of logid %d and alertid %d",
            logdata.log.id,
            logdata.alert.model.id
        )
        logdata.alert.sync()
        logdata.host = mod.Host.objects.select_related().get(
            id=logdata.host.id
        )
        malert = logdata.alert.model
        if malert.datetimerecovery is None:
            if not malert.acknowledged:
                malert.acknowledged              = True
                malert.acknowledgedautomatically = True
                malert.datetimeacknowledged      = cmn.now()
            malert.datetimerecovery     = logdata.log.datetime
            malert.comment              = logdata.log.output
            logdata.log.processed       = const.NagiosLog.STATE_ALERT_CLOSED
            logdata.is_from_acknowledge = True
            logdata.host.hastrouble     = False
        else:
            logdata.log.processed = (
                const.NagiosLog.STATE_NO_PREVIOUS_DOWN
            )
        if logdata.service_instance:
            logdata.service_instance = (
                mod.Hostserviceinstance.objects.select_related(
                ).get(
                    id=logdata.service_instance.id
                )
            )
            logdata.service_instance.hastrouble = False
            logdata.service_instance.save()
        logdata.alert.save()
        logdata.host.save()
        cmn.lg.debug(
            "Alert (ID {0}) has been resolved.".format(
                logdata.alert.model.id
            )
        )

    def handle_alert(self, logdata):
        """Handling of an alert itself"""
        cmn.lg.debug("Checking if new alert logid %d", logdata.log.id)

        logdata.log.processed = logdata.check_current_state()

        if logdata.log.processed:
            # State is another than before, so no alerting
            # needed
            if logdata.log.processed in ALERT_DISPLAY_STATES:
                self.display_alert(logdata)
            return

        # Create (and display) alert for logdata
        self.display_alert(logdata)

        # Set state to alerted only when we are sure that
        # an alert was created and saved
        logdata.log.processed = const.NagiosLog.STATE_ALERTED

        if logdata.service_instance:
            # Re-getting service instance object as
            # it may have been changed meanwhile
            logdata.service_instance = (
                mod.Hostserviceinstance.objects.select_related().get(
                    id=logdata.service_instance.id
                )
            )
            logdata.service_instance.hastrouble = True
            logdata.service_instance.save()


class AlertExec(cmn.ExecutorMixin, cmn.LogDataLockMixin):
    """Exection of alerting"""

    def __init__(self):
        super(AlertExec, self).__init__()
        self.user_rule_alerted = {}
        self.dummy = dummy.Dummy()

    @cmn.run_on_executor
    def get_template_ex(self, user, notification_type):
        """Memoized version of get_template bound to this object instance"""
        return self.get_template(
            user,
            notification_type
        )

    @cmn.memoize
    def get_template(self, user, notification_type):
        """Memoized version of get_template bound to this object instance"""
        return calert.UserCommunicationAlert.get_template(
            user,
            notification_type
        )

    @cmn.run_on_executor
    def handle_EMAIL(self, timeframe, user):
        """Send alert via email

        :type timeframe: symonitoring_model.models.Timeframe
        :type user: symonitoring_model.models.User
        """
        res = False
        if user.email:
            template = self.get_template(
                user,
                const.ActionType.EMAIL
            )
            comm = calert.UserCommunicationAlert(
                logdata   = self.logdata,
                user      = user,
                timeframe = timeframe,
                template  = template,
            )
            res = comm.send_email()
        if not res:
            cmn.lg.debug(
                "Could not send mail to: %s (%s)", user.email, user.fullname
            )
        return res

    @cmn.coroutine
    def handle_XMPP(self, timeframe, user, orig_action):
        """Send alert via XMPP"""
        res = False
        if user.jabber:
            template = yield self.get_template_ex(
                user,
                const.ActionType.XMPP
            )
            comm = calert.UserCommunicationAlert(
                logdata   = self.logdata,
                user      = user,
                timeframe = timeframe,
                template  = template,
            )
            res = yield comm.send_xmpp(orig_action=orig_action)

        if not res:
            cmn.lg.debug(
                "Could not send xmpp to: %s (%s)", user.jabber, user.fullname
            )
        raise cmn.Return(res)

    @cmn.coroutine
    def handle_MATTERMOST(self, timeframe, user):
        res = False
        if user.mattermost:
            template = yield self.get_template_ex(
                user,
                const.ActionType.MATTERMOST
            )
            comm = calert.UserCommunicationAlert(
                logdata   = self.logdata,
                user      = user,
                timeframe = timeframe,
                template  = template,
            )
            res = yield comm.send_mattermost()
        if not res:
            cmn.lg.debug(
                "Could not send mattermost message to: %s (%s)",
                user.mattermost, user.fullname
            )
        raise cmn.Return(res)

    @cmn.coroutine
    def handle_PHONE(self, timeframe, user):
        """Send alert via phone

        :type user: symonitoring_model.models.User
        """
        res = False
        if user.phone:
            comm = calert.UserCommunicationAlert(
                logdata   = self.logdata,
                user      = user,
                timeframe = timeframe,
                template  = None,
            )
            res = yield comm.call_phone()
        if not res:
            cmn.lg.debug("Could not call: %s (%s)", user.phone, user.fullname)
        raise cmn.Return(res)

    @cmn.coroutine
    def handle_SMS(self, timeframe, user):
        """Send alert via SMS"""
        res = False
        if user.sms:
            template = yield self.get_template_ex(
                user,
                const.ActionType.SMS
            )
            comm = calert.UserCommunicationAlert(
                logdata   = self.logdata,
                user      = user,
                timeframe = timeframe,
                template  = template,
            )
            res = yield comm.send_sms()
        if not res:
            cmn.lg.debug(
                "Could not send sms to: %s (%s)", user.sms, user.fullname
            )
        raise cmn.Return(res)

    @cmn.coroutine
    def dispatch_action(
            self,
            timeframe,
            rule,
            _,
            user
    ):
        """Executes the action

        :type timeframe: symonitoring_model.models.Timeframe
        :type rule: symonitoring_model.models.Timeframerule
        :type user: symonitoring_model.models.User
        """
        res = False
        act = rule.actionid
        # This means for this media restart is disabled
        # typically one email is enough, but there
        # should be multiple calls
        if not act.multiple:
            key = (user.id, rule.id)
            if key in self.user_rule_alerted:
                return
            else:
                self.user_rule_alerted[key] = True
        action_id = self.dummy.force_action(rule.actionid_id)
        if action_id == const.ActionType.EMAIL:
            res = yield self.handle_EMAIL(timeframe, user)
        elif action_id == const.ActionType.SMS:
            res = yield self.handle_SMS(timeframe, user)
        elif action_id == const.ActionType.PHONE:
            res = yield self.handle_PHONE(timeframe, user)
        elif action_id == const.ActionType.MATTERMOST:
            res = yield self.handle_MATTERMOST(timeframe, user)
        elif action_id == const.ActionType.XMPP:
            if action_id != rule.actionid_id:
                res = yield self.handle_XMPP(
                    timeframe,
                    user,
                    orig_action=rule.actionid_id,
                )
            else:
                res = yield self.handle_XMPP(
                    timeframe,
                    user,
                    orig_action=action_id,
                )
        else:
            cmn.lg.debug("Unknown action: %s", action_id)

        yield self.logdata.alert.create_update_notification_datetime()
        yield self.save_alert_and_log(res, user, action_id)

    @cmn.run_on_executor
    def save_alert_and_log(
            self,
            res,
            user,
            action_id
    ):
        """Saving alert log"""
        with transaction.atomic():
            alert_log = mod.Alertlog(
                typeid_id = const.action_to_notification[action_id],
                userid    = user,
                alertid   = self.logdata.alert.model,
                hasfailed = not res
            )
            alert_log.save()

    @cmn.run_on_executor
    def save_logdata(self):
        """Save the logdata in an atomic transaction"""
        with transaction.atomic():
            self.logdata.log.save()

    @cmn.coroutine
    def schedule_next_timeout(
            self,
            timeframe,
            rule,
            users,
            user
    ):
        """Plans the next timeout

        :type timeframe: syalert.gen_models.Timeframe
        :type rule: syalert.gen_models.Timeframerule
        :type user: syalert.gen_models.User
        """
        logdata      = self.logdata
        alert_state  = logdata.alert_state
        alert_state.increment_rule_timeframe(timeframe)
        try:
            rule, user, timeout = alert_state.get_timeout(timeframe)

            cmn.lg.debug(
                "Creating handler for: %s/%s (%s delay, %s timeout)",
                user.fullname,
                logdata.sla.name,
                rule.delay,
                timeout
            )
            handler  = self.mktimeout(
                timeframe,
                rule,
                users,
                user
            )
            cmn.add_timeout(
                timeout,
                handler
            )
        except als.AlertTimeoutReached:
            cmn.lg.warning(
                "Timeframe (ID %d) for alert (ID %d) timeout reached, "
                "skipping..",
                timeframe.id,
                logdata.alert.model.id

            )

    def mktimeout(
            self,
            timeframe,
            rule,
            users,
            user
    ):
        """Make a timeout

        :type timeframe: symonitoring_model.models.Timeframe
        :type rule: symonitoring_model.models.Timeframerule
        :type user: symonitoring_model.models.User
        """
        obj = TimeoutHandler(
            self,
            timeframe,
            rule,
            users,
            user
        )
        return obj.handle

    @cmn.coroutine
    def handle(self):
        """Main handler function"""
        logdata     = self.logdata
        if not int(logdata.log.processed) == const.NagiosLog.STATE_ALERTED:
            return
        alert_logic = logdata.alert
        if alert_logic.model and alert_logic.model.acknowledged:
            yield logdata.alert_state.delete_ex()
        elif alert_logic.model:
            alert = alert_logic.model
            yield alert_logic.create_update_notification_datetime()
            cmn.lg.debug(
                "Handling alert (%d) with state: %s",
                alert.id,
                alert.state.lower(),
            )
            for timeframe in logdata.timeframes:
                cmn.lg.debug(
                    "Planning timeframe (ID %d)",
                    timeframe.id
                )
                assert isinstance(timeframe, mod.Timeframe)
                if len(timeframe.ext_rules) < 1:  # pragma: no cover
                    cmn.lg.critical(
                        "Timeframe (ID %d) for alert (ID %d) has no rules, "
                        "this should not happen!",
                        timeframe.id,
                        alert.id
                    )
                    continue

                try:
                    rule, user, timeout = logdata.alert_state.get_timeout(
                        timeframe
                    )
                    users               = logdata.users[timeframe]
                    handler             = self.mktimeout(
                        timeframe,
                        rule,
                        users,
                        user
                    )
                    cmn.add_timeout(
                        timeout,
                        handler
                    )
                except als.AlertTimeoutReached:  # pragma: no cover
                    cmn.lg.critical(
                        "Timeframe (ID %d) for alert (ID %d) timeout reached, "
                        "the first alert is already skipped!",
                        timeframe.id,
                        alert.id
                    )
