# -*- coding: UTF-8 -*-
# pylint: disable=too-many-arguments

"""
Handle hold (alert) entries
"""

# System imports
import datetime
import threading
import django.db.transaction         as transaction

# Project imports
import symonitoring_alert.common     as cmn
import symonitoring_alert.constants  as const
import symonitoring_alert.hold.alert as hal
import symonitoring_alert.logdata    as ld
import symonitoring_model.models     as mod

_logdata_in_timeout = {}
_lock = threading.RLock()
_REMOVAL_STATES = set([
    const.NagiosLog.STATE_NO_PREVIOUS_DOWN,
    const.NagiosLog.STATE_ALERT_CLOSED,
    const.NagiosLog.STATE_OPEN,
    const.NagiosLog.STATE_ALERTED,
    const.NagiosLog.STATE_WARNING,
    const.NagiosLog.STATE_ALERT_REPETITION,
])
_HOLD_STATES = set([
    const.NagiosLog.STATE_DEPENDENCY_DOWN,
    const.NagiosLog.STATE_GLOBAL_DISABLE,
    const.NagiosLog.STATE_HOST_DISABLED,
    const.NagiosLog.STATE_HOST_DOWN,
    const.NagiosLog.STATE_IN_MAINTENANCE,
    const.NagiosLog.STATE_NO_SLA,
    const.NagiosLog.STATE_NO_TEAM,
])


class HoldTimeoutHandler(object):
    """Handles timeouts of holds, needed e.g. for host dependencies."""

    def __init__(self, logdata):
        self._logdata = logdata

    @cmn.coroutine
    def handle(self):
        """Main handler for handling timed out hold entries."""
        # pylint: disable=global-variable-not-assigned
        global _logdata_in_timeout
        global _lock

        with (yield self._logdata.lock.acquire()):
            logdata = self._logdata
            key = logdata.log.id
            cmn.lg.debug("Handling hold timeout for logid %d", key)

            hold_entry = yield logdata.hold.get_ex()

            if hold_entry:
                logdata.log.processed = const.NagiosLog.STATE_OPEN
                yield logdata.hold.delete_ex()
                yield logdata.save_log_ex()
                cmn.lg.debug(
                    "State is open, deleted holds for logid %d",
                    key
                )
            else:
                cmn.lg.debug(
                    "State is %s, NOT deleted holds for logid %d",
                    const.nagios_log_dict[int(logdata.log.processed)],
                    key
                )

            with _lock:
                del _logdata_in_timeout[key]


class HoldExec(cmn.ExecutorMixin, cmn.LogDataLockMixin):
    """Handles holding of logdata objects based upon their current state."""

    @classmethod
    def handle_host_dependencies(cls, logdata):
        """Handles alerts for the dependant objects (hosts) for the host
        of the given logdata object."""
        for dhost in logdata.host.dependency_set.all():
            dhostalerts = dhost.alert_set.filter(
                serviceinstanceid__isnull=True,
                datetimerecovery__isnull=True,
            )
            for dhalert in dhostalerts:
                dhalert.datetimerecovery = cmn.now()
                dhalert.comment = (
                    "Automatically closed because dependency host is down."
                )
                if not dhalert.acknowledged:
                    dhalert.acknowledged = True
                    dhalert.acknowledgedautomatically = True
                dhalert.save()

                # Keep last NagiosLog entry for alert of current logdata
                # in hold
                latest_nl = dhalert.nagioslog_set.latest('id')
                if latest_nl.state in [
                        const.AlertState.STATE_DOWN,
                        const.AlertState.STATE_CRITICAL,
                ]:
                    hal.HoldAlert.set(latest_nl.id)
                    cmn.lg.debug(
                        "Got/created hold for logid %d in"
                        "handle_host_dependencies",
                        latest_nl.id
                    )
                latest_nl.processed = const.NagiosLog.STATE_DEPENDENCY_DOWN
                latest_nl.save()

    @classmethod
    def handle_duplicated_alerts(cls, logdata):
        """Handles previous alerts for the host of the given logdata
        object.
        """
        if logdata.duplicated_alerts:
            for dalert in logdata.duplicated_alerts:
                dalert = mod.Alert.objects.get(
                    id=dalert.id
                )
                dalert.datetimerecovery = cmn.now()
                dalert.comment = (
                    "Automatically closed because of new alert."
                )
                if not dalert.acknowledged:
                    dalert.acknowledged = True
                    dalert.acknowledgedautomatically = True
                dalert.save()

    @classmethod
    def handle_host_service_dependencies(cls, logdata):
        """Handles alerts for services configured for the host of the
        given logdata object."""
        if (
                not logdata.alert.model.serviceinstanceid and
                logdata.host_service_alerts
        ):
            for halert in logdata.host_service_alerts:
                halert = mod.Alert.objects.get(
                    id=halert.id
                )
                halert.acknowledged = True
                halert.acknowledgedautomatically = True
                halert.comment = (
                    "Automatically acknowledged because host is down."
                )
                halert.save()

                # Keep last NagiosLog entry for alert of current logdata
                # in hold
                latest_nl = halert.nagioslog_set.latest('id')
                if latest_nl.state in [
                        const.AlertState.STATE_DOWN,
                        const.AlertState.STATE_CRITICAL,
                ]:
                    hal.HoldAlert.set(latest_nl.id)
                    cmn.lg.debug(
                        "Got/created hold for logid %d "
                        "in handle_host_service_dependencies",
                        latest_nl.id
                    )
                latest_nl.processed = const.NagiosLog.STATE_HOST_DOWN
                latest_nl.save()

    @cmn.run_on_executor
    def handle(self):
        """Main handler for handling holds."""
        logdata = self.logdata
        cmn.lg.debug("Bookkeeping hold for logid %d", logdata.log.id)
        current_state     = int(logdata.log.processed)

        with transaction.atomic():
            if current_state == const.NagiosLog.STATE_ALERTED:
                self.handle_host_dependencies(logdata)
                self.handle_duplicated_alerts(logdata)
                self.handle_host_service_dependencies(logdata)

            if (
                    current_state in _REMOVAL_STATES or
                    logdata.has_recovered
            ):
                for hold in logdata.hold.previous_related_active_entries():
                    entry = hold.nagioslogid
                    entry.hold_set.all().delete()
                    cmn.lg.debug(
                        "Removed all holds for logid %d by removal state",
                        entry.id
                    )

            else:
                if current_state == const.NagiosLog.STATE_GLOBAL_DISABLE:
                    for hold in logdata.hold.previous_related_active_entries():
                        entry = hold.nagioslogid
                        entry.hold_set.all().delete()
                        cmn.lg.debug(
                            "Removed all holds for logid %d by global disable",
                            entry.id
                        )
                if current_state in _HOLD_STATES:
                    logdata.hold.get_or_create()
                    cmn.lg.debug(
                        "Got/created hold for logid %d",
                        logdata.log.id
                    )


class HoldHandler(cmn.ExecutorMixin):
    """Handles hold (alert) entries."""
    def handle_host_dependcy_down(self, logdata):
        """Wrapper for passing data to callback method."""
        obj = HoldTimeoutHandler(logdata)
        return obj.handle

    @cmn.run_on_executor
    def handle(self, entry):
        """Main handler function."""
        # pylint: disable=global-variable-not-assigned
        global _logdata_in_timeout
        global _lock

        assert isinstance(entry, mod.Hold)
        logdata = ld.LogData(entry.nagioslogid)
        cmn.lg.debug("Handling hold for logid %d", logdata.log.id)
        old_state = int(logdata.log.processed)
        logdata.load()
        current_state = logdata.check_current_state()

        if current_state in _HOLD_STATES and not logdata.has_recovered:
            # State is set, so no alerting needed
            cmn.lg.debug(
                "Not handling %d as state is %s",
                logdata.log.id,
                const.nagios_log_dict[current_state]
            )
            if old_state != current_state:
                logdata.log.processed = current_state
                logdata.log.save()
            return

        if old_state in [
                const.NagiosLog.STATE_HOST_DOWN,
                const.NagiosLog.STATE_DEPENDENCY_DOWN,
        ]:
            delay = datetime.timedelta(
                milliseconds=cmn.s.DEPENDENCY_TIMEOUT
            )
            key = logdata.log.id
            with _lock:
                if key not in _logdata_in_timeout:
                    handler = self.handle_host_dependcy_down(logdata)
                    cmn.add_timeout(
                        delay,
                        handler
                    )
                    _logdata_in_timeout[key] = True
                    cmn.lg.debug(
                        "Setting hold timeout for logid %d, %f s",
                        key,
                        delay.total_seconds()
                    )
                else:
                    cmn.lg.debug(
                        "Already set hold timeout for logid %d, %f s",
                        key,
                        delay.total_seconds()
                    )
                cmn.lg.debug(
                    "current hold timeouts: %s",
                    str(_logdata_in_timeout)
                )

        else:
            logdata.log.processed = const.NagiosLog.STATE_OPEN
            logdata.log.save()
            logdata.hold.delete()
            cmn.lg.debug(
                "Old state is not from dependency, set logid %d to open",
                logdata.log.id,
            )
