# -*- coding: UTF-8 -*-
# pylint: disable=missing-docstring

"""
Helper class for testing of handling alert holds
"""

# System imports
import datetime
import mock
import tornado.gen                              as gen

# Project imports
import symonitoring_alert.common                as cmn
import symonitoring_alert.constants             as const
import symonitoring_model.models                as mod
import symonitoring_alert.handle.nagioslog      as nl
import symonitoring_alert.query                 as q


NOTIFICATION_HOST_DOWN = {
    'address': u'192.24.12.3',
    'attempt': 1,
    'checkcommand': u'check-host-alive',
    'datetime': cmn.now(),
    'duration': u'0d 0h 0m 0s',
    'hostname': u'localhost',
    'lastproblemid': 0,
    'name': u'localhost',
    'objecttype': u'HOST',
    'output': u'PING CRITICAL - Packet loss = 100%',
    'perfdata': u'rta=5000.000000ms;3000.000000;5000.000000;0.000000 pl=100%;'
                '80;100;0',
    'problemid': 0,
    'processed': u'0',
    'state': u'DOWN',
    'statetype': u'HARD',
    'timestamp': cmn.now()
}
NOTIFICATION_SERVICE_DOWN = {
    'address': u'192.168.12.3',
    'attempt': 5,
    'checkcommand': 'tcp!221',
    'datetime': cmn.now(),
    'duration': '0d 0h 4m 2s',
    'hostname': 'localhost',
    'lastproblemid': 12,
    'name': 'ssh__default',
    'objecttype': 'SERVICE',
    'output': 'Connection refused',
    'perfdata': '',
    'problemid': 0,
    'processed': '0',
    'state': 'DOWN',
    'statetype': 'HARD',
    'timestamp': cmn.local_datetime(datetime.datetime(
        2014, 5, 6, 15, 19, 7
    ))
}
NOW = datetime.datetime.now()
NOW_MOCKED_IN_SLA = cmn.local_datetime(
    datetime.datetime(
        NOW.year,
        NOW.month,
        NOW.day,
        23,
        15
    )
)
NOW_MOCKED_OUT_OF_SLA = cmn.local_datetime(
    datetime.datetime(
        NOW.year,
        NOW.month,
        NOW.day,
        6,
        0
    )
)


class HoldTestBase(cmn.TestCase):
    """ Base class for commonly used methods.
    """
    def run_init(self,
                 notification,
                 hostname,
                 now=cmn.now(),
                 state=const.AlertState.STATE_DOWN,
                 is_deactivated=False):
        """ Initialization before running tests.
        """
        # pylint: disable=too-many-arguments

        handler = nl.NagiosLog()
        log = mod.Nagioslog(
            **notification
        )
        log.hostname = hostname
        self.save_log(log)
        if state != const.AlertState.STATE_DOWN:
            log.state = state

        @gen.coroutine
        def coro():
            """ Coroutine for handling nagioslog entries.
            """
            yield handler.handle(log, no_exec=True)

        with mock.patch("symonitoring_alert.common.now") as now_mocked, \
                mock.patch(
                    "symonitoring_alert.logdata.LogData.check_deactivated") \
                as check_deactivated_mocked:
            now_mocked.return_value = now
            check_deactivated_mocked.return_value = is_deactivated
            cmn.run_sync(coro)
            return handler.logdata

    @cmn.coroutine
    def coro_get_alert_holds(self):
        res = yield self.get_alert_holds_ex()
        raise cmn.Return(res)

    @cmn.run_on_executor
    def get_alert_holds_ex(self):
        return self.get_alert_holds()

    def get_alert_holds(self):
        return list(mod.Hold.objects.all().select_related())

    @cmn.run_on_executor
    def get_alert_by_id_ex(self, alert_id):
        return self.get_alert_by_id(alert_id)

    def get_alert_by_id(self, alert_id):
        try:
            alert = mod.Alert.objects.get(
                id=alert_id
            )
        except mod.Alert.DoesNotExist:
            alert = None

        return alert

    @cmn.coroutine
    def handle_nagioslog(self):
        yield q.NagiosLog().handle()

    @cmn.coroutine
    def coro_get_inform_holds(self):
        res = yield self.get_inform_holds_ex()
        raise cmn.Return(res)

    @cmn.run_on_executor
    def get_inform_holds_ex(self):
        return self.get_inform_holds()

    def get_inform_holds(self):
        return list(mod.HoldInform.objects.all().select_related())


class HoldHandlerRunner(object):
    def __init__(self, handler):
        self._logdata = None
        self._handler = handler()

    @property
    def logdata(self):
        return self._logdata

    @logdata.setter
    def logdata(self, logdata):
        self._logdata = logdata

    @cmn.coroutine
    def run(self):
        self._handler.logdata = self._logdata
        with cmn._allow_add_handler_test_only():
            yield self._handler.handle()

    @cmn.coroutine
    def sync_logdata(self):
        yield self._logdata.sync_ex()

    @cmn.coroutine
    def handle_holds(self):
        yield q.Hold().handle()
