# -*- coding: UTF-8 -*-
# pylint: disable=too-many-arguments

"""
Handle hold (inform) entries
"""

# System imports
import django.db.transaction                    as transaction

# Project imports
import symonitoring_alert.common                as cmn
import symonitoring_alert.constants             as const
import symonitoring_alert.handle.inform         as hi
import symonitoring_alert.logdata               as ld
import symonitoring_model.models                as mod


class HoldInformExec(cmn.ExecutorMixin, cmn.LogDataLockMixin):
    """Handles holding of informing of logdata objects based upon their
    current state."""

    @cmn.run_on_executor
    def handle(self):
        """ Main handler for handling informative holds."""

        logdata = self.logdata
        current_state = int(logdata.log.processed)

        cmn.lg.debug("Bookkeeping hold-inform for logid %d", logdata.log.id)

        with transaction.atomic():
            if logdata.log.alert:
                logdata.alert.sync(override_id=logdata.log.alert.id)
                alert = logdata.alert.model
                if not (
                        alert.acknowledged or
                        alert.acknowledgedprocessed
                ):
                    logdata.hold_inform.get_or_create()
                    cmn.lg.debug(
                        "State is %s, added hold informs for logid %d",
                        const.nagios_log_dict[current_state],
                        logdata.log.id
                    )

            for hold in logdata.hold_inform.previous_related_active_entries():
                entry = hold.nagioslogid
                assert isinstance(entry, mod.Nagioslog)
                entry.holdinform_set.all().delete()
                cmn.lg.debug(
                    "State is %s, deleted hold informs for logid %d",
                    const.nagios_log_dict[current_state],
                    logdata.log.id
                )


class HoldInformHandler(cmn.ExecutorMixin):
    """ Handles hold (inform) entries."""

    @cmn.run_on_executor
    def get_alert_id(self, logdata):
        """Silly helper for getting alert_id on execu."""
        logdata.alert.sync(override_id=logdata.log.alert.id)

    @cmn.coroutine
    def handle(self, entry):
        """ Main handler function for hold inform entries."""

        assert isinstance(entry, mod.HoldInform)
        logdata = ld.LogData(entry.nagioslogid)
        yield logdata.load_ex()
        yield self.get_alert_id(logdata)
        alert   = logdata.alert.model
        cmn.lg.debug("Handling hold inform for logid %d", logdata.log.id)

        if alert and alert.acknowledged:
            logdata.is_from_acknowledge = True

            if not (
                    alert.acknowledgedautomatically or
                    alert.acknowledgedprocessed
            ):
                for handler in [hi.InformInit, hi.InformExec]:
                    handler = handler()
                    handler.logdata = logdata
                    yield handler.handle()

            yield logdata.hold_inform.delete_ex()
            cmn.lg.debug(
                "Informed hold for logid %d",
                logdata.log.id
            )
        else:
            cmn.lg.debug(
                "Not informing hold for logid %d, no (acknowledged) alert",
                logdata.log.id
            )
