# -*- coding: UTF-8 -*-
# pylint: disable=missing-docstring,protected-access

"""
Testing of handling informative holds
"""

# System imports
import mock
import datetime

# Project imports
import symonitoring_alert.handle.hold_base_test as base
import symonitoring_alert.common                as cmn
import symonitoring_alert.constants             as const
import symonitoring_alert.handle.hold_inform    as hhldi
import symonitoring_alert.handle.inform         as info
import symonitoring_model.models                as mod
import symonitoring_alert.query                 as q


class HandleHoldInformTest(base.HoldTestBase):
    """Tests the handling of holding informative events."""

    def run_dispatch_and_render(
            self,
            notification,
            hostname,
            state,
            func,
            return_value_func,
    ):
        res = self.run_init(
            notification=notification,
            hostname=hostname,
            state=state,
            now=base.NOW_MOCKED_IN_SLA,
        )
        handler = info.InformExec()
        handler.logdata = res

        with mock.patch(func) as op:
            op.return_value = return_value_func()
            timeframe     = res.timeframes[0]
            user          = res.users[timeframe][0]
            with cmn.disable_dummy_mode():
                @cmn.coroutine
                def coro():
                    yield handler.handle()
                    yield cmn.wait(datetime.timedelta(
                        seconds=0.05 * cmn.s.TIMING_TEST_SCALE
                    ))
                with cmn._allow_add_handler_test_only():
                    cmn.run_sync(coro)
                return (timeframe, user, res, op)

    @cmn.coroutine
    def handle_hold(self):
        with cmn._allow_add_handler_test_only():
            yield q.Hold().handle()
            yield cmn.wait(datetime.timedelta(
                seconds=0.05 * cmn.s.TIMING_TEST_SCALE
            ))

    @cmn.coroutine
    def coro_res(self):
        raise cmn.Return(["ok"])

    def run_common(self):
        hhrunner = base.HoldHandlerRunner(hhldi.HoldInformExec)

        # CRITICAL (1/1) --> Alerted
        (_, user, logdata_11, op) = self.run_dispatch_and_render(
            notification=base.NOTIFICATION_HOST_DOWN,
            hostname="host1DependantOnTesthost24",
            state=const.AlertState.STATE_CRITICAL,
            func="symonitoring_alert.communication.xmpp.send_message",
            return_value_func=self.coro_res,
        )
        # Run holdinformexec
        hhrunner.logdata = logdata_11
        cmn.run_sync(hhrunner.run)

        # Test that actually was alerted an informed
        self.assertTrue(op.called)
        self.assertEqual(user.id, 2)
        message_body = op.call_args[1]['mbody']
        self.assertNotEqual(message_body, '')

        # Test that infofeed is correctly set
        self.assertEqual(2, len(logdata_11.infofeeds))
        feed = logdata_11.infofeeds[0]
        self.assertIsInstance(feed, mod.Infofeed)
        self.assertEqual(8, feed.id)
        self.assertTrue(feed.onalert)
        self.assertTrue(feed.onwarning)
        self.assertTrue(feed.onacknowledge)
        self.assertEqual(
            feed.ext_infofeednotificationtype[0].slaactionid_id,
            const.ActionType.XMPP
        )

        # Test that infofeeds are on hold
        inform_holds = cmn.run_sync(self.coro_get_inform_holds)
        self.assertTrue(
            any(h.nagioslogid_id == logdata_11.log.id for h in inform_holds)
        )

        return (inform_holds, logdata_11)

    def test_critical_followed_by_warning(self):
        hhrunner = base.HoldHandlerRunner(hhldi.HoldInformExec)
        (inform_holds, logdata_11) = self.run_common()

        # WARNING (1/2) --> Do nothing
        (_, _, logdata_12, op) = self.run_dispatch_and_render(
            notification=base.NOTIFICATION_HOST_DOWN,
            hostname="host1DependantOnTesthost24",
            state=const.AlertState.STATE_WARNING,
            func="symonitoring_alert.communication.xmpp.send_message",
            return_value_func=self.coro_res,
        )
        # Run holdinformexec
        hhrunner.logdata = logdata_12
        cmn.run_sync(hhrunner.run)

        # --> NO alert, ack'ed automatically
        self.assertFalse(op.called)

        # --> NO hold anymore
        inform_holds = cmn.run_sync(self.coro_get_inform_holds)
        self.assertFalse(
            any(h.nagioslogid_id == logdata_11.log.id for h in inform_holds)
        )

    def test_critical_followed_by_ack(self):
        hhrunner = base.HoldHandlerRunner(hhldi.HoldInformExec)
        (inform_holds, logdata_11) = self.run_common()
        self.assertTrue(logdata_11.alert.model)

        # ACK (1/x) --> Do nothing
        logdata_11.alert.model.acknowledged = True
        logdata_11.alert.model.acknowledgedby_id = 2

        @cmn.coroutine
        def save_logdata():
            yield logdata_11.save_ex()
        cmn.run_sync(save_logdata)

        # Run holdinformexec
        hhrunner.logdata = logdata_11
        cmn.run_sync(hhrunner.run)
        inform_holds = cmn.run_sync(self.coro_get_inform_holds)
        # --> Entry still in hold table
        self.assertTrue(
            any(h.nagioslogid_id == logdata_11.log.id for h in inform_holds)
        )

        # Run inform handler
        # --> inform ack
        # --> remove from hold
        with mock.patch(
            "symonitoring_alert.communication.xmpp.send_message",
        ) as op:
            op.return_value = self.coro_res()
            cmn.run_sync(self.handle_hold)

        # --> Alert
        # Test that actually was alerted an informed
        self.assertTrue(op.called)
        message_body = op.call_args[1]['mbody']
        self.assertNotEqual(message_body, '')
        # Test that infofeed is correctly set
        self.assertEqual(2, len(logdata_11.infofeeds))
        feed = logdata_11.infofeeds[0]
        self.assertIsInstance(feed, mod.Infofeed)
        self.assertEqual(8, feed.id)
        self.assertTrue(feed.onalert)
        self.assertTrue(feed.onwarning)
        self.assertTrue(feed.onacknowledge)
        self.assertEqual(
            feed.ext_infofeednotificationtype[0].slaactionid_id,
            const.ActionType.XMPP
        )

        # --> Entry _not_ in hold table anymore
        inform_holds = cmn.run_sync(self.coro_get_inform_holds)
        self.assertFalse(
            any(h.nagioslogid_id == logdata_11.log.id for h in inform_holds)
        )
