# -*- coding: UTF-8 -*-
# pylint: disable=missing-docstring,too-many-lines

"""
Testing of handling alert holds
"""

# System imports
import datetime
import mock
import tornado.gen                              as gen

# Project imports
import symonitoring_alert.handle.hold_base_test as base
import symonitoring_alert.common                as cmn
import symonitoring_alert.constants             as const
import symonitoring_alert.handle.hold           as hhld
import symonitoring_alert.logdata               as ld
import symonitoring_alert.query                 as q
import symonitoring_model.models                as mod
import symonitoring_alert.handle.nagioslog      as nl


notification_host_critical = {
    'address': u'127.0.0.1',
    'attempt': 1,
    'checkcommand': u'check-host-alive',
    'datetime': cmn.local_datetime(datetime.datetime(
        2014, 5, 6, 15, 19, 7
    )),
    'duration': u'0d 0h 0m 0s',
    'hostname': u'localhost',
    'lastproblemid': 47,
    'name': u'testhost',
    'objecttype': u'HOST',
    'output': u'PING CRITICAL - Packet loss = 100%',
    'perfdata': u'rta=5000.000000ms;3000.000000;5000.000000;0.000000 pl=100%;'
                '80;100;0',
    'problemid': 48,
    'processed': u'0',
    'state': u'DOWN',
    'statetype': u'HARD',
    'timestamp': cmn.local_datetime(datetime.datetime(
        2014, 5, 6, 15, 19, 7
    ))
}

notification_service_critical = {
    'address': '127.0.0.1',
    'attempt': 5,
    'checkcommand': 'tcp!221',
    'datetime': cmn.local_datetime(datetime.datetime(
        2014, 5, 6, 15, 19, 7
    )),
    'duration': '0d 0h 4m 2s',
    'hostname': 'testhost',
    'lastproblemid': 12,
    'name': 'ssh__default',
    'objecttype': 'SERVICE',
    'output': 'Connection refused',
    'perfdata': '',
    'problemid': 46,
    'processed': '0',
    'state': 'CRITICAL',
    'statetype': 'HARD',
    'timestamp': cmn.local_datetime(datetime.datetime(
        2014, 5, 6, 15, 19, 7
    ))
}


class HandleHoldTest(base.HoldTestBase):
    """ Tests the handling of hold events.
    """

    def test_out_of_sla(self):
        hhrunner = base.HoldHandlerRunner(hhld.HoldExec)

        # CRITICAL (1/1) --> Alerted
        logdata_11 = self.run_init(
            notification=base.NOTIFICATION_HOST_DOWN,
            hostname="testhost25",
            now=base.NOW_MOCKED_IN_SLA,
        )
        self.assertIs(logdata_11.__class__, ld.LogData)
        self.assertEqual(
            logdata_11.log.processed,
            const.NagiosLog.STATE_ALERTED
        )
        self.assertTrue(logdata_11.alert.model)
        self.assertEqual(
            logdata_11.alert.model.state,
            const.AlertState.STATE_DOWN
        )

        # CRITICAL (2/2) --> NO_SLA
        logdata_22 = self.run_init(
            notification=base.NOTIFICATION_HOST_DOWN,
            hostname="testhost24",
            now=base.NOW_MOCKED_OUT_OF_SLA,
        )
        self.assertIs(logdata_22.__class__, ld.LogData)
        self.assertEqual(
            logdata_22.log.processed,
            const.NagiosLog.STATE_NO_SLA
        )
        hhrunner.logdata = logdata_22
        cmn.run_sync(hhrunner.run)
        alert_holds = cmn.run_sync(self.coro_get_alert_holds)
        self.assertTrue(
            any(h.nagioslogid_id == logdata_22.log.id for h in alert_holds)
        )
        self.assertTrue(logdata_22.alert.model)
        self.assertEqual(
            logdata_22.alert.model.state,
            const.AlertState.STATE_DOWN
        )

        # CRITICAL (3/3) --> NO_SLA
        logdata_33 = self.run_init(
            notification=base.NOTIFICATION_HOST_DOWN,
            hostname="testhost23",
            now=base.NOW_MOCKED_OUT_OF_SLA,
        )
        self.assertIs(logdata_33.__class__, ld.LogData)
        self.assertEqual(
            logdata_33.log.processed,
            const.NagiosLog.STATE_NO_SLA
        )
        hhrunner.logdata = logdata_33
        cmn.run_sync(hhrunner.run)
        alert_holds = cmn.run_sync(self.coro_get_alert_holds)
        self.assertTrue(
            any(h.nagioslogid_id == logdata_33.log.id for h in alert_holds)
        )
        self.assertTrue(logdata_33.alert.model)
        self.assertEqual(
            logdata_33.alert.model.state,
            const.AlertState.STATE_DOWN
        )

        # OK (2/4) --> OK
        logdata_24 = self.run_init(
            notification=base.NOTIFICATION_HOST_DOWN,
            hostname="testhost24",
            now=base.NOW_MOCKED_IN_SLA,
            is_deactivated=True,
            state=const.AlertState.STATE_OK
        )
        self.assertIs(logdata_24.__class__, ld.LogData)
        self.assertEqual(
            logdata_24.log.processed,
            const.NagiosLog.STATE_ALERT_CLOSED
        )
        hhrunner.logdata = logdata_24
        cmn.run_sync(hhrunner.run)
        alert_holds = cmn.run_sync(self.coro_get_alert_holds)
        self.assertFalse(
            any(h.nagioslogid_id == logdata_24.log.id for h in alert_holds)
        )

        # HOLD: CRITICAL (3/3) --> ALERT
        @gen.coroutine
        def handle_hold():
            """ Coroutine for handling hold entries.
            """
            yield q.Hold().handle()
            yield q.NagiosLog().handle()

        with mock.patch(
            'symonitoring_alert.common.now'
        ) as now_mocked, mock.patch(
            'symonitoring_alert.common.add_timeout'
        ) as add_timeout_mocked, mock.patch(
            'symonitoring_alert.common.add_callback'
        ):
            now_mocked.return_value = base.NOW_MOCKED_IN_SLA
            cmn.run_sync(handle_hold)

        self.assertTrue(add_timeout_mocked.called)
        alert_holds = cmn.run_sync(self.coro_get_alert_holds)
        self.assertEqual(len(alert_holds), 0)

    def test_host_service_disable(self):
        hhrunner = base.HoldHandlerRunner(hhld.HoldExec)

        # CRITICAL (1/1) --> Alerted
        logdata_11 = self.run_init(
            notification=base.NOTIFICATION_HOST_DOWN,
            hostname="testhost25",
            now=base.NOW_MOCKED_IN_SLA,
        )
        self.assertIs(logdata_11.__class__, ld.LogData)
        self.assertEqual(
            logdata_11.log.processed,
            const.NagiosLog.STATE_ALERTED
        )
        self.assertTrue(logdata_11.alert.model)
        self.assertEqual(
            logdata_11.alert.model.state,
            const.AlertState.STATE_DOWN
        )

        # CRITICAL (2/2) --> DISABLED
        logdata_22 = self.run_init(
            notification=base.NOTIFICATION_HOST_DOWN,
            hostname="testhost24",
            now=base.NOW_MOCKED_IN_SLA,
            is_deactivated=True
        )
        self.assertIs(logdata_22.__class__, ld.LogData)
        self.assertEqual(
            logdata_22.log.processed,
            const.NagiosLog.STATE_HOST_DISABLED
        )
        self.assertFalse(logdata_22.alert.model)
        hhrunner.logdata = logdata_22
        cmn.run_sync(hhrunner.run)
        alert_holds = cmn.run_sync(self.coro_get_alert_holds)
        self.assertTrue(
            any(h.nagioslogid_id == logdata_22.log.id for h in alert_holds)
        )

        # CRITICAL (3/3) --> DISABLED
        logdata_33 = self.run_init(
            notification=base.NOTIFICATION_HOST_DOWN,
            hostname="testhost23",
            now=base.NOW_MOCKED_IN_SLA,
            is_deactivated=True
        )
        self.assertIs(logdata_33.__class__, ld.LogData)
        self.assertEqual(
            logdata_33.log.processed,
            const.NagiosLog.STATE_HOST_DISABLED
        )
        self.assertFalse(logdata_33.alert.model)
        hhrunner.logdata = logdata_33
        cmn.run_sync(hhrunner.run)
        alert_holds = cmn.run_sync(self.coro_get_alert_holds)
        self.assertTrue(
            any(h.nagioslogid_id == logdata_33.log.id for h in alert_holds)
        )

        # CRITICAL (2/4) --> OK
        logdata_24 = self.run_init(
            notification=base.NOTIFICATION_HOST_DOWN,
            hostname="testhost24",
            now=base.NOW_MOCKED_IN_SLA,
            is_deactivated=True,
            state=const.AlertState.STATE_OK
        )
        self.assertIs(logdata_24.__class__, ld.LogData)
        self.assertEqual(
            logdata_24.log.processed,
            const.NagiosLog.STATE_NO_PREVIOUS_DOWN
        )
        self.assertFalse(logdata_24.alert.model)
        hhrunner.logdata = logdata_24
        cmn.run_sync(hhrunner.run)
        alert_holds = cmn.run_sync(self.coro_get_alert_holds)
        self.assertFalse(
            any(h.nagioslogid_id == logdata_24.log.id for h in alert_holds)
        )
        self.assertFalse(
            any(h.nagioslogid_id == logdata_22.log.id for h in alert_holds)
        )

    def test_downtime(self):
        hhrunner = base.HoldHandlerRunner(hhld.HoldExec)

        # CRITICAL (1/1) --> Alerted
        logdata_11 = self.run_init(
            notification=base.NOTIFICATION_HOST_DOWN,
            hostname="testhost25",
            now=base.NOW_MOCKED_IN_SLA,
        )
        self.assertIs(logdata_11.__class__, ld.LogData)
        self.assertEqual(
            logdata_11.log.processed,
            const.NagiosLog.STATE_ALERTED
        )
        self.assertTrue(logdata_11.alert.model)
        self.assertEqual(
            logdata_11.alert.model.state,
            const.AlertState.STATE_DOWN
        )

        # CRITICAL (2/2) --> NO alert / IN_MAINTENANCE
        with mock.patch(
            'symonitoring_alert.logdata.LogData.check_downtime'
        ) as check_downtime_mocked:
            check_downtime_mocked.return_value = True
            logdata_22 = self.run_init(
                notification=base.NOTIFICATION_HOST_DOWN,
                hostname="testhost24",
                now=base.NOW_MOCKED_IN_SLA,
            )
        self.assertIs(logdata_22.__class__, ld.LogData)
        self.assertEqual(
            logdata_22.log.processed,
            const.NagiosLog.STATE_IN_MAINTENANCE
        )
        hhrunner.logdata = logdata_22
        cmn.run_sync(hhrunner.run)
        alert_holds = cmn.run_sync(self.coro_get_alert_holds)
        self.assertTrue(
            any(h.nagioslogid_id == logdata_22.log.id for h in alert_holds)
        )
        self.assertFalse(logdata_22.alert.model)

        # CRITICAL (3/3) --> NO alert / IN_MAINTENANCE
        with mock.patch(
            'symonitoring_alert.logdata.LogData.check_downtime'
        ) as check_downtime_mocked:
            check_downtime_mocked.return_value = True
            logdata_33 = self.run_init(
                notification=base.NOTIFICATION_HOST_DOWN,
                hostname="testhost23",
                now=base.NOW_MOCKED_IN_SLA,
            )
        self.assertIs(logdata_33.__class__, ld.LogData)
        self.assertEqual(
            logdata_33.log.processed,
            const.NagiosLog.STATE_IN_MAINTENANCE
        )
        hhrunner.logdata = logdata_33
        cmn.run_sync(hhrunner.run)
        alert_holds = cmn.run_sync(self.coro_get_alert_holds)
        self.assertTrue(
            any(h.nagioslogid_id == logdata_33.log.id for h in alert_holds)
        )
        self.assertFalse(logdata_33.alert.model)

        # OK (2/4) --> NO alert / OK
        with mock.patch(
            'symonitoring_alert.logdata.LogData.check_downtime'
        ) as check_downtime_mocked:
            logdata_24 = self.run_init(
                notification=base.NOTIFICATION_HOST_DOWN,
                hostname="testhost24",
                now=base.NOW_MOCKED_IN_SLA,
                state=const.AlertState.STATE_OK
            )
        self.assertIs(logdata_24.__class__, ld.LogData)
        self.assertEqual(
            logdata_24.log.processed,
            const.NagiosLog.STATE_NO_PREVIOUS_DOWN
        )
        hhrunner.logdata = logdata_24
        cmn.run_sync(hhrunner.run)
        alert_holds = cmn.run_sync(self.coro_get_alert_holds)
        self.assertFalse(
            any(h.nagioslogid_id == logdata_24.log.id for h in alert_holds)
        )
        self.assertFalse(logdata_24.alert.model)

        # HOLD: CRITICAL (3/3) --> ALERT
        @gen.coroutine
        def handle_hold():
            """ Coroutine for handling hold entries.
            """
            yield q.Hold().handle()
            yield q.NagiosLog().handle()
            yield logdata_33.load_ex()

        with mock.patch(
            'symonitoring_alert.common.now'
        ) as now_mocked, mock.patch(
            'symonitoring_alert.common.add_timeout'
        ) as add_timeout_mocked, mock.patch(
            'symonitoring_alert.common.add_callback'
        ):
            now_mocked.return_value = base.NOW_MOCKED_IN_SLA
            cmn.run_sync(handle_hold)

        self.assertTrue(add_timeout_mocked.called)
        self.assertTrue(logdata_33.alert.model)
        alert_holds = cmn.run_sync(self.coro_get_alert_holds)
        self.assertEqual(len(alert_holds), 0)

    def test_host_service_dependency_service_nok(self):
        hhrunner = base.HoldHandlerRunner(hhld.HoldExec)

        # CRITICAL (1/1) --> Alerted
        logdata_11 = self.run_init(
            notification=base.NOTIFICATION_SERVICE_DOWN,
            hostname="testhost24",
            now=base.NOW_MOCKED_IN_SLA,
        )
        hhrunner.logdata = logdata_11
        cmn.run_sync(hhrunner.run)
        self.assertIs(logdata_11.__class__, ld.LogData)
        self.assertEqual(
            logdata_11.log.processed,
            const.NagiosLog.STATE_ALERTED
        )
        self.assertTrue(logdata_11.alert.model)
        self.assertEqual(
            logdata_11.alert.model.state,
            const.AlertState.STATE_DOWN
        )

        # CRITICAL (2/2)
        # --> Close (1/1) / HOST_DOWN
        # --> Alert (2/2) / ALERTED
        logdata_22 = self.run_init(
            notification=base.NOTIFICATION_HOST_DOWN,
            hostname="testhost24",
            now=base.NOW_MOCKED_IN_SLA,
        )
        hhrunner.logdata = logdata_22
        cmn.run_sync(hhrunner.run)
        self.assertIs(logdata_22.__class__, ld.LogData)
        self.assertEqual(
            logdata_22.log.processed,
            const.NagiosLog.STATE_ALERTED
        )
        self.assertTrue(logdata_22.alert.model)
        self.assertFalse(logdata_22.alert.model.acknowledged)
        self.assertFalse(logdata_22.alert.model.acknowledgedautomatically)

        hhrunner.logdata = logdata_11
        cmn.run_sync(hhrunner.sync_logdata)
        self.assertEqual(
            int(logdata_11.log.processed),
            const.NagiosLog.STATE_HOST_DOWN
        )
        self.assertTrue(logdata_11.alert.model.acknowledged)
        self.assertTrue(logdata_11.alert.model.acknowledgedautomatically)

        hhrunner.logdata = logdata_11
        cmn.run_sync(hhrunner.run)
        alert_holds = cmn.run_sync(self.coro_get_alert_holds)
        self.assertTrue(
            any(h.nagioslogid_id == logdata_11.log.id for h in alert_holds)
        )

        # OK (2/3)
        # --> Close (2/2) / ALERT_CLOSED
        logdata_23 = self.run_init(
            notification=base.NOTIFICATION_HOST_DOWN,
            hostname="testhost24",
            now=base.NOW_MOCKED_IN_SLA,
            state=const.AlertState.STATE_OK
        )
        hhrunner.logdata = logdata_23
        cmn.run_sync(hhrunner.run)
        self.assertIs(logdata_23.__class__, ld.LogData)
        self.assertEqual(
            logdata_23.log.processed,
            const.NagiosLog.STATE_ALERT_CLOSED
        )
        self.assertTrue(logdata_23.alert.model.acknowledged)
        self.assertTrue(logdata_23.alert.model.acknowledgedautomatically)

        # HOLD: CRITICAL (1/1) --> Alert / ALERTED
        # pylint: disable=protected-access
        @gen.coroutine
        def handle_hold():
            with mock.patch(
                'symonitoring_alert.common.now'
            ) as now_mocked:
                now_mocked.return_value = base.NOW_MOCKED_IN_SLA

                @gen.coroutine
                def test_after_10ms():
                    alert_holds = yield self.get_alert_holds_ex()
                    yield self.get_alert_id(logdata_11)
                    self.assertTrue(
                        any(
                            h.nagioslogid_id == logdata_11.log.id
                            for h in alert_holds
                        )
                    )
                    self.assertTrue(logdata_11.alert.model.acknowledged)
                    self.assertTrue(
                        logdata_11.alert.model.acknowledgedautomatically
                    )
                    self.assertEqual(
                        logdata_11.alert.model.comment,
                        "Automatically acknowledged because host is down."
                    )

                @gen.coroutine
                def test_after_150ms():
                    alert_holds = yield self.get_alert_holds_ex()
                    self.assertFalse(
                        any(
                            h.nagioslogid_id == logdata_11.log.id
                            for h in alert_holds
                        )
                    )
                    self.assertFalse(logdata_11.alert.model.acknowledged)
                    self.assertFalse(
                        logdata_11.alert.model.acknowledgedautomatically
                    )

                try:
                    orig_delay = cmn.s.DEPENDENCY_TIMEOUT
                    # in milliseconds
                    cmn.s.DEPENDENCY_TIMEOUT = 100 * cmn.s.TIMING_TEST_SCALE

                    yield q.Hold().handle()
                    yield q.NagiosLog().handle()
                    yield logdata_11.load_ex()
                    yield cmn.wait(datetime.timedelta(
                        seconds=0.01 * cmn.s.TIMING_TEST_SCALE
                    ))
                    yield q.NagiosLog().handle()
                    yield logdata_11.load_ex()
                    yield test_after_10ms()
                    yield cmn.wait(datetime.timedelta(
                        seconds=0.1 * cmn.s.TIMING_TEST_SCALE
                    ))
                    with mock.patch(
                        'symonitoring_alert.common.add_timeout'
                    ) as timeout_mocked:
                        yield q.NagiosLog().handle()
                        self.assertTrue(timeout_mocked.called)
                    yield logdata_11.load_ex()
                    yield test_after_150ms()

                finally:
                    cmn.s.DEPENDENCY_TIMEOUT = orig_delay

        # pylint: disable=protected-access
        with cmn._allow_add_handler_test_only():
            cmn.run_sync(handle_hold)

    def test_host_service_dependency_service_ok(self):
        hhrunner = base.HoldHandlerRunner(hhld.HoldExec)

        # CRITICAL (1/1) --> Alerted
        logdata_11 = self.run_init(
            notification=base.NOTIFICATION_SERVICE_DOWN,
            hostname="testhost24",
            now=base.NOW_MOCKED_IN_SLA,
        )
        hhrunner.logdata = logdata_11
        cmn.run_sync(hhrunner.run)
        self.assertIs(logdata_11.__class__, ld.LogData)
        self.assertEqual(
            logdata_11.log.processed,
            const.NagiosLog.STATE_ALERTED
        )
        self.assertTrue(logdata_11.alert.model)
        self.assertEqual(
            logdata_11.alert.model.state,
            const.AlertState.STATE_DOWN
        )

        # CRITICAL (2/2)
        # --> Close (1/1) / HOST_DOWN
        # --> Alert (2/2) / ALERTED
        logdata_22 = self.run_init(
            notification=base.NOTIFICATION_HOST_DOWN,
            hostname="testhost24",
            now=base.NOW_MOCKED_IN_SLA,
        )
        hhrunner.logdata = logdata_22
        cmn.run_sync(hhrunner.run)
        self.assertIs(logdata_22.__class__, ld.LogData)
        self.assertEqual(
            logdata_22.log.processed,
            const.NagiosLog.STATE_ALERTED
        )
        self.assertTrue(logdata_22.alert.model)
        self.assertFalse(logdata_22.alert.model.acknowledged)
        self.assertFalse(logdata_22.alert.model.acknowledgedautomatically)

        hhrunner.logdata = logdata_11
        cmn.run_sync(hhrunner.sync_logdata)
        self.assertEqual(
            int(logdata_11.log.processed),
            const.NagiosLog.STATE_HOST_DOWN
        )
        self.assertTrue(logdata_11.alert.model.acknowledged)
        self.assertTrue(logdata_11.alert.model.acknowledgedautomatically)

        hhrunner.logdata = logdata_11
        cmn.run_sync(hhrunner.run)
        alert_holds = cmn.run_sync(self.coro_get_alert_holds)
        self.assertTrue(
            any(h.nagioslogid_id == logdata_11.log.id for h in alert_holds)
        )

        # OK (2/3)
        # --> Close (2/2) / ALERT_CLOSED
        logdata_23 = self.run_init(
            notification=base.NOTIFICATION_HOST_DOWN,
            hostname="testhost24",
            now=base.NOW_MOCKED_IN_SLA,
            state=const.AlertState.STATE_OK
        )
        hhrunner.logdata = logdata_23
        cmn.run_sync(hhrunner.run)
        self.assertIs(logdata_23.__class__, ld.LogData)
        self.assertEqual(
            logdata_23.log.processed,
            const.NagiosLog.STATE_ALERT_CLOSED
        )
        self.assertTrue(logdata_23.alert.model.acknowledged)
        self.assertTrue(logdata_23.alert.model.acknowledgedautomatically)

        # HOLD: CRITICAL (1/1) --> Alert / ALERTED
        # pylint: disable=protected-access
        @gen.coroutine
        def handle_hold():
            with mock.patch(
                'symonitoring_alert.common.now'
            ) as now_mocked:
                now_mocked.return_value = base.NOW_MOCKED_IN_SLA

                @gen.coroutine
                def test_after_10ms():
                    alert_holds = yield self.get_alert_holds_ex()
                    yield self.get_alert_id(logdata_11)
                    self.assertTrue(
                        any(
                            h.nagioslogid_id == logdata_11.log.id
                            for h in alert_holds
                        )
                    )
                    self.assertTrue(logdata_11.alert.model.acknowledged)
                    self.assertTrue(
                        logdata_11.alert.model.acknowledgedautomatically
                    )
                    self.assertEqual(
                        logdata_11.alert.model.comment,
                        "Automatically acknowledged because host is down."
                    )

                @gen.coroutine
                def test_after_150ms():
                    alert_holds = yield self.get_alert_holds_ex()
                    yield self.get_alert_id(logdata_11)
                    self.assertFalse(
                        any(
                            h.nagioslogid_id == logdata_11.log.id
                            for h in alert_holds
                        )
                    )
                    self.assertTrue(logdata_11.alert.model)
                    yield logdata_11.alert.sync_ex()
                    alert = logdata_11.alert.model
                    self.assertTrue(alert.acknowledged)
                    self.assertTrue(alert.acknowledgedautomatically)

                @gen.coroutine
                def recover_service():
                    # OK (1/4) --> NO alert / ALERT_CLOSED
                    handler = nl.NagiosLog()
                    log = mod.Nagioslog(
                        **base.NOTIFICATION_SERVICE_DOWN
                    )
                    log.hostname = "testhost24"
                    yield self.save_log_ex(log)
                    log.state = const.AlertState.STATE_OK

                    with mock.patch(
                        "symonitoring_alert.common.now"
                    ) as now_mocked:
                        now_mocked.return_value = base.NOW_MOCKED_IN_SLA
                        yield handler.handle(log, no_exec=True)
                        exec_handler = hhld.HoldExec()
                        exec_handler.logdata = handler.logdata
                        yield exec_handler.handle()
                        yield q.Hold().handle()

                try:
                    orig_delay = cmn.s.DEPENDENCY_TIMEOUT
                    # in milliseconds
                    cmn.s.DEPENDENCY_TIMEOUT = 100 * cmn.s.TIMING_TEST_SCALE

                    yield q.Hold().handle()
                    yield q.NagiosLog().handle()
                    yield logdata_11.load_ex()
                    yield cmn.wait(datetime.timedelta(
                        seconds=0.01 * cmn.s.TIMING_TEST_SCALE
                    ))
                    yield q.NagiosLog().handle()
                    yield logdata_11.load_ex()
                    yield test_after_10ms()
                    yield recover_service()
                    yield cmn.wait(datetime.timedelta(
                        seconds=0.1 * cmn.s.TIMING_TEST_SCALE
                    ))
                    with mock.patch(
                        'symonitoring_alert.common.add_timeout'
                    ) as timeout_mocked:
                        yield q.NagiosLog().handle()
                        self.assertFalse(timeout_mocked.called)
                    yield logdata_11.load_ex()
                    yield test_after_150ms()

                finally:
                    cmn.s.DEPENDENCY_TIMEOUT = orig_delay

        # pylint: disable=protected-access
        with cmn._allow_add_handler_test_only():
            cmn.run_sync(handle_hold)

    def test_host_host_dependency_hostb_nok(self):
        # Test host - host dependencies
        # Host B is dependant on host A
        #
        # First, host B goes down, then host
        # A goes down. Later on, host A receives
        # an OK message, but host B not.
        #
        # When handling holds later on, host B
        # must alert.

        hhrunner = base.HoldHandlerRunner(hhld.HoldExec)

        # Host B goes down
        # CRITICAL (1/1) --> Alerted (1/1)
        logdata_11 = self.run_init(
            notification=base.NOTIFICATION_HOST_DOWN,
            hostname="host1DependantOnTesthost24",
            now=base.NOW_MOCKED_IN_SLA,
        )
        hhrunner.logdata = logdata_11
        cmn.run_sync(hhrunner.run)
        self.assertIs(logdata_11.__class__, ld.LogData)
        self.assertEqual(
            logdata_11.log.processed,
            const.NagiosLog.STATE_ALERTED
        )
        self.assertTrue(logdata_11.alert.model)
        self.assertEqual(
            logdata_11.alert.model.state,
            const.AlertState.STATE_DOWN
        )

        # Host A goes down
        # CRITICAL (2/2) --> Alerted (2/2)
        logdata_22 = self.run_init(
            notification=base.NOTIFICATION_HOST_DOWN,
            hostname="testhost24",
            now=base.NOW_MOCKED_IN_SLA,
        )
        hhrunner.logdata = logdata_22
        cmn.run_sync(hhrunner.run)
        self.assertIs(logdata_22.__class__, ld.LogData)
        self.assertEqual(
            logdata_22.log.processed,
            const.NagiosLog.STATE_ALERTED
        )
        self.assertTrue(logdata_22.alert.model)
        self.assertEqual(
            logdata_22.alert.model.state,
            const.AlertState.STATE_DOWN
        )

        # Assure host B is set correspondingly
        hhrunner.logdata = logdata_11
        cmn.run_sync(hhrunner.sync_logdata)
        # --> Closed alert for (1/1)
        self.assertEqual(
            int(logdata_11.log.processed),
            const.NagiosLog.STATE_DEPENDENCY_DOWN
        )
        self.assertTrue(logdata_11.alert.model)
        self.assertTrue(logdata_11.alert.model.acknowledged)
        self.assertTrue(logdata_11.alert.model.acknowledgedautomatically)
        # --> Dependency down (1/1)
        self.assertEqual(
            logdata_11.alert.model.comment,
            "Automatically closed because dependency host is down."
        )
        # --> Hold (1/1)
        alert_holds = cmn.run_sync(self.coro_get_alert_holds)
        self.assertTrue(
            any(h.nagioslogid_id == logdata_11.log.id for h in alert_holds)
        )

        # Host A comes up again
        # OK (2/3)
        # --> Close alert for (2/2)
        logdata_23 = self.run_init(
            notification=base.NOTIFICATION_HOST_DOWN,
            hostname="testhost24",
            now=base.NOW_MOCKED_IN_SLA,
            state=const.AlertState.STATE_OK
        )
        self.assertIs(logdata_23.__class__, ld.LogData)
        self.assertEqual(
            logdata_23.log.processed,
            const.NagiosLog.STATE_ALERT_CLOSED
        )

        # Hold handler is (periodically) running
        # Host B is still down
        # --> Create alert for (1/1)
        # --> Remove (1/1) from hold
        @gen.coroutine
        def handle_hold():
            with mock.patch("symonitoring_alert.common.now") as now_mocked:
                now_mocked.return_value = base.NOW_MOCKED_IN_SLA

                @gen.coroutine
                def test_after_10ms():
                    alert_holds = yield self.get_alert_holds_ex()
                    yield self.get_alert_id(logdata_11)
                    self.assertTrue(
                        any(
                            h.nagioslogid_id == logdata_11.log.id
                            for h in alert_holds
                        )
                    )
                    self.assertTrue(logdata_11.alert.model.acknowledged)
                    self.assertTrue(
                        logdata_11.alert.model.acknowledgedautomatically
                    )
                    self.assertEqual(
                        logdata_11.alert.model.comment,
                        "Automatically closed because dependency host is down."
                    )

                @gen.coroutine
                def test_after_150ms():
                    alert_holds = yield self.get_alert_holds_ex()
                    self.assertFalse(
                        any(
                            h.nagioslogid_id == logdata_11.log.id
                            for h in alert_holds
                        )
                    )
                    self.assertFalse(logdata_11.alert.model.acknowledged)
                    self.assertFalse(
                        logdata_11.alert.model.acknowledgedautomatically
                    )

                try:
                    orig_delay = cmn.s.DEPENDENCY_TIMEOUT
                    # in milliseconds
                    cmn.s.DEPENDENCY_TIMEOUT = 100 * cmn.s.TIMING_TEST_SCALE

                    yield q.Hold().handle()
                    yield q.NagiosLog().handle()
                    yield logdata_11.load_ex()
                    yield cmn.wait(datetime.timedelta(
                        seconds=0.01 * cmn.s.TIMING_TEST_SCALE
                    ))
                    yield q.NagiosLog().handle()
                    yield logdata_11.load_ex()
                    yield test_after_10ms()
                    yield cmn.wait(datetime.timedelta(
                        seconds=0.1 * cmn.s.TIMING_TEST_SCALE
                    ))
                    with mock.patch(
                        'symonitoring_alert.common.add_timeout'
                    ) as timeout_mocked:
                        yield q.NagiosLog().handle()
                        self.assertTrue(timeout_mocked.called)
                    yield logdata_11.load_ex()
                    yield test_after_150ms()

                finally:
                    cmn.s.DEPENDENCY_TIMEOUT = orig_delay

        @cmn.coroutine
        def coro_res():
            raise cmn.Return(["ok"])

        # pylint: disable=protected-access
        with cmn._allow_add_handler_test_only(), mock.patch(
            "symonitoring_alert.communication.xmpp.send_message",
        ) as op:
            op.return_value = coro_res()
            cmn.run_sync(handle_hold)

    @cmn.run_on_executor
    def get_alert_id(self, logdata):
        """Silly helper for getting alert_id on execu."""
        logdata.alert.sync(override_id=logdata.log.alert.id)

    def test_host_host_dependency_hostb_ok(self):
        # Test host - host dependencies
        # Host B is dependant on host A
        #
        # First, host B goes down, then host
        # A goes down. Later on, host A receives
        # an OK message, but host B not.
        #
        # When handling holds later on,
        # host B receives an OK message.
        # so _no_ alert for host B.

        hhrunner = base.HoldHandlerRunner(hhld.HoldExec)

        # Host B goes down
        # CRITICAL (1/1) --> Alerted (1/1)
        logdata_11 = self.run_init(
            notification=base.NOTIFICATION_HOST_DOWN,
            hostname="host1DependantOnTesthost24",
            now=base.NOW_MOCKED_IN_SLA,
        )
        hhrunner.logdata = logdata_11
        cmn.run_sync(hhrunner.run)
        self.assertIs(logdata_11.__class__, ld.LogData)
        self.assertEqual(
            logdata_11.log.processed,
            const.NagiosLog.STATE_ALERTED
        )
        self.assertTrue(logdata_11.alert.model)
        self.assertEqual(
            logdata_11.alert.model.state,
            const.AlertState.STATE_DOWN
        )

        # Host A goes down
        # CRITICAL (2/2) --> Alerted (2/2)
        logdata_22 = self.run_init(
            notification=base.NOTIFICATION_HOST_DOWN,
            hostname="testhost24",
            now=base.NOW_MOCKED_IN_SLA,
        )
        hhrunner.logdata = logdata_22
        cmn.run_sync(hhrunner.run)
        self.assertIs(logdata_22.__class__, ld.LogData)
        self.assertEqual(
            logdata_22.log.processed,
            const.NagiosLog.STATE_ALERTED
        )
        self.assertTrue(logdata_22.alert.model)
        self.assertEqual(
            logdata_22.alert.model.state,
            const.AlertState.STATE_DOWN
        )

        # Assure host B is set correspondingly
        hhrunner.logdata = logdata_11
        cmn.run_sync(hhrunner.sync_logdata)
        # --> Closed alert for (1/1)
        self.assertEqual(
            int(logdata_11.log.processed),
            const.NagiosLog.STATE_DEPENDENCY_DOWN
        )
        self.assertTrue(logdata_11.alert.model)
        self.assertTrue(logdata_11.alert.model.acknowledged)
        self.assertTrue(logdata_11.alert.model.acknowledgedautomatically)
        # --> Dependency down (1/1)
        self.assertEqual(
            logdata_11.alert.model.comment,
            "Automatically closed because dependency host is down."
        )
        # --> Hold (1/1)
        alert_holds = cmn.run_sync(self.coro_get_alert_holds)
        self.assertTrue(
            any(h.nagioslogid_id == logdata_11.log.id for h in alert_holds)
        )

        # Host A comes up again
        # OK (2/3)
        # --> Close alert for (2/2)
        logdata_23 = self.run_init(
            notification=base.NOTIFICATION_HOST_DOWN,
            hostname="testhost24",
            now=base.NOW_MOCKED_IN_SLA,
            state=const.AlertState.STATE_OK
        )
        self.assertIs(logdata_23.__class__, ld.LogData)
        self.assertEqual(
            logdata_23.log.processed,
            const.NagiosLog.STATE_ALERT_CLOSED
        )

        # Hold handler is (periodically) running
        # Host B is still down
        # --> Create alert for (1/1)
        # --> Remove (1/1) from hold
        @gen.coroutine
        def handle_hold():
            with mock.patch("symonitoring_alert.common.now") as now_mocked:
                now_mocked.return_value = base.NOW_MOCKED_IN_SLA

                @gen.coroutine
                def test_after_10ms():
                    alert_holds = yield self.get_alert_holds_ex()
                    self.assertTrue(
                        any(
                            h.nagioslogid_id == logdata_11.log.id
                            for h in alert_holds
                        )
                    )
                    yield self.get_alert_id(logdata_11)
                    self.assertTrue(logdata_11.alert.model.acknowledged)
                    self.assertTrue(
                        logdata_11.alert.model.acknowledgedautomatically
                    )
                    self.assertEqual(
                        logdata_11.alert.model.comment,
                        "Automatically closed because dependency host is down."
                    )

                @gen.coroutine
                def test_after_150ms():
                    alert_holds = yield self.get_alert_holds_ex()
                    yield self.get_alert_id(logdata_11)
                    # --> NO alert for (1/1) as OK was received
                    self.assertTrue(logdata_11.alert.model.acknowledged)
                    self.assertTrue(
                        logdata_11.alert.model.acknowledgedautomatically
                    )
                    # --> Hold for (1/1) must have been removed
                    self.assertFalse(
                        any(
                            h.nagioslogid_id == logdata_11.log.id
                            for h in alert_holds
                        )
                    )

                @gen.coroutine
                def recover_host_b():
                    # OK (1/4) --> NO alert / ALERT_CLOSED
                    handler = nl.NagiosLog()
                    log = mod.Nagioslog(
                        **base.NOTIFICATION_HOST_DOWN
                    )
                    log.hostname = "host1DependantOnTesthost24"
                    yield self.save_log_ex(log)
                    log.state = const.AlertState.STATE_OK

                    with mock.patch(
                        "symonitoring_alert.common.now"
                    ) as now_mocked:
                        now_mocked.return_value = base.NOW_MOCKED_IN_SLA
                        yield handler.handle(log, no_exec=True)
                        exec_handler = hhld.HoldExec()
                        exec_handler.logdata = handler.logdata
                        yield exec_handler.handle()
                        yield q.Hold().handle()

                try:
                    orig_delay = cmn.s.DEPENDENCY_TIMEOUT
                    # in milliseconds
                    cmn.s.DEPENDENCY_TIMEOUT = 100 * cmn.s.TIMING_TEST_SCALE

                    yield q.Hold().handle()
                    yield q.NagiosLog().handle()
                    yield logdata_11.load_ex()
                    yield cmn.wait(datetime.timedelta(
                        seconds=0.01 * cmn.s.TIMING_TEST_SCALE
                    ))
                    yield q.NagiosLog().handle()
                    yield logdata_11.load_ex()
                    yield test_after_10ms()
                    yield recover_host_b()
                    yield cmn.wait(datetime.timedelta(
                        seconds=0.1 * cmn.s.TIMING_TEST_SCALE
                    ))
                    with mock.patch(
                        'symonitoring_alert.common.add_timeout'
                    ) as timeout_mocked:
                        yield q.NagiosLog().handle()
                        self.assertFalse(timeout_mocked.called)
                    yield logdata_11.load_ex()
                    yield test_after_150ms()

                finally:
                    cmn.s.DEPENDENCY_TIMEOUT = orig_delay

        # pylint: disable=protected-access
        with cmn._allow_add_handler_test_only():
            cmn.run_sync(handle_hold)

    def run_common_global_disable(self):
        hhrunner = base.HoldHandlerRunner(hhld.HoldExec)

        # CRITICAL (1/1) --> Alerted
        logdata_11 = self.run_init(
            notification=base.NOTIFICATION_HOST_DOWN,
            hostname="testhost25",
            now=base.NOW_MOCKED_IN_SLA,
        )
        self.assertIs(logdata_11.__class__, ld.LogData)
        self.assertEqual(
            logdata_11.log.processed,
            const.NagiosLog.STATE_ALERTED
        )
        self.assertTrue(logdata_11.alert.model)
        self.assertEqual(
            logdata_11.alert.model.state,
            const.AlertState.STATE_DOWN
        )

        with cmn._allow_block_for_test_only():
            q.QueryBase._global_disable_cache = None
            avoid = mod.Setting(
                key="avoidalert_date"
            )
            avoid.value = base.NOW_MOCKED_IN_SLA + datetime.timedelta(
                minutes=1
            )
            avoid.save()

        # CRITICAL (2/2) --> DISABLED
        logdata_22 = self.run_init(
            notification=base.NOTIFICATION_HOST_DOWN,
            hostname="testhost24",
            now=base.NOW_MOCKED_IN_SLA,
        )
        self.assertIs(logdata_22.__class__, ld.LogData)
        self.assertEqual(
            logdata_22.log.processed,
            const.NagiosLog.STATE_GLOBAL_DISABLE
        )
        self.assertFalse(logdata_22.alert.model)
        hhrunner.logdata = logdata_22
        cmn.run_sync(hhrunner.run)
        alert_holds = cmn.run_sync(self.coro_get_alert_holds)
        self.assertTrue(
            any(h.nagioslogid_id == logdata_22.log.id for h in alert_holds)
        )

        # CRITICAL (3/3) --> DISABLED
        logdata_33 = self.run_init(
            notification=base.NOTIFICATION_HOST_DOWN,
            hostname="testhost23",
            now=base.NOW_MOCKED_IN_SLA,
        )
        self.assertIs(logdata_33.__class__, ld.LogData)
        self.assertEqual(
            logdata_33.log.processed,
            const.NagiosLog.STATE_GLOBAL_DISABLE
        )
        self.assertFalse(logdata_33.alert.model)
        hhrunner.logdata = logdata_33
        cmn.run_sync(hhrunner.run)
        alert_holds = cmn.run_sync(self.coro_get_alert_holds)
        self.assertTrue(
            any(h.nagioslogid_id == logdata_33.log.id for h in alert_holds)
        )

        # CRITICAL (2/4) --> OK
        logdata_24 = self.run_init(
            notification=base.NOTIFICATION_HOST_DOWN,
            hostname="testhost24",
            now=base.NOW_MOCKED_IN_SLA,
            state=const.AlertState.STATE_OK
        )
        self.assertIs(logdata_24.__class__, ld.LogData)
        self.assertEqual(
            logdata_24.log.processed,
            const.NagiosLog.STATE_GLOBAL_DISABLE
        )
        self.assertFalse(logdata_24.alert.model)
        hhrunner.logdata = logdata_24
        cmn.run_sync(hhrunner.run)
        alert_holds = cmn.run_sync(self.coro_get_alert_holds)
        self.assertTrue(
            any(h.nagioslogid_id == logdata_24.log.id for h in alert_holds)
        )
        self.assertFalse(
            any(h.nagioslogid_id == logdata_22.log.id for h in alert_holds)
        )
        self.assertTrue(
            any(h.nagioslogid_id == logdata_33.log.id for h in alert_holds)
        )
        with cmn._allow_block_for_test_only():
            q.QueryBase._global_disable_cache = None
            mod.Setting.objects.filter(
                key="avoidalert_date"
            ).all().delete()
        return (alert_holds, logdata_11, logdata_22, logdata_33, logdata_24)

    def test_host_global_disable(self):
        try:
            (
                alert_holds,
                _,
                _,
                logdata_33,
                logdata_24,
            ) = self.run_common_global_disable()
            hold_handler = hhld.HoldHandler()

            @cmn.coroutine
            def coro():
                for hold in alert_holds:
                    yield hold_handler.handle(hold)
                yield nl.NagiosLog().handle(logdata_33.log, no_exec=True)
                yield nl.NagiosLog().handle(logdata_24.log, no_exec=True)
                yield logdata_33.sync_log_ex()
                yield logdata_24.sync_log_ex()

            with mock.patch(
                'symonitoring_alert.common.now'
            ) as now_mocked:
                now_mocked.return_value = base.NOW_MOCKED_IN_SLA
                cmn.run_sync(coro)
            cmn.lg.debug("logdata_33.id: %s", logdata_33.log.id)
            cmn.lg.debug("logdata_24.id: %s", logdata_24.log.id)
            self.assertEqual(
                int(logdata_33.log.processed),
                const.NagiosLog.STATE_ALERTED
            )
            self.assertEqual(
                int(logdata_24.log.processed),
                const.NagiosLog.STATE_NO_PREVIOUS_DOWN
            )
        finally:
            with cmn._allow_block_for_test_only():
                q.QueryBase._global_disable_cache = None
                mod.Setting.objects.filter(
                    key="avoidalert_date"
                ).all().delete()

    def test_host_global_disable_plus_no_sla(self):
        try:
            (
                alert_holds,
                _,
                _,
                logdata_33,
                logdata_24,
            ) = self.run_common_global_disable()
            hold_handler = hhld.HoldHandler()

            @cmn.coroutine
            def coro():
                for hold in alert_holds:
                    yield hold_handler.handle(hold)
                yield nl.NagiosLog().handle(logdata_33.log, no_exec=True)
                yield nl.NagiosLog().handle(logdata_24.log, no_exec=True)
                yield logdata_33.sync_log_ex()
                yield logdata_24.sync_log_ex()

            with mock.patch(
                'symonitoring_alert.common.now'
            ) as now_mocked:
                now_mocked.return_value = base.NOW_MOCKED_OUT_OF_SLA
                cmn.run_sync(coro)
            self.assertEqual(
                int(logdata_33.log.processed),
                const.NagiosLog.STATE_NO_SLA
            )
            self.assertEqual(
                int(logdata_24.log.processed),
                const.NagiosLog.STATE_NO_PREVIOUS_DOWN
            )
            with mock.patch(
                'symonitoring_alert.common.now'
            ) as now_mocked:
                now_mocked.return_value = base.NOW_MOCKED_IN_SLA
                cmn.run_sync(coro)
            self.assertEqual(
                int(logdata_33.log.processed),
                const.NagiosLog.STATE_ALERTED
            )
            self.assertEqual(
                int(logdata_24.log.processed),
                const.NagiosLog.STATE_NO_PREVIOUS_DOWN
            )
        finally:
            with cmn._allow_block_for_test_only():
                q.QueryBase._global_disable_cache = None
                mod.Setting.objects.filter(
                    key="avoidalert_date"
                ).all().delete()


class HoldExecTest(cmn.TestCase):
    "Test the class hold.HoldExec"

    def handle_log(self, log, hostname=None, name=None):
        """Handle a log entry"""
        handler = nl.NagiosLog()
        if hostname is not None:
            log['hostname'] = hostname
        if name is not None:
            log['name'] = name
        log = mod.Nagioslog(
            **log
        )
        self.save_log(log)

        @gen.coroutine
        def coro():
            yield handler.handle(log, no_exec=True)
            hold_exec_handler = hhld.HoldExec()
            hold_exec_handler.logdata = handler.logdata
            yield hold_exec_handler.handle()

        cmn.run_sync(coro)
        return log

    def test_host_dependencies_host_down(self):
        host1_log = self.handle_log(
            notification_host_critical,
            'host1DependantOnTesthost24'
        )
        host2_log = self.handle_log(
            notification_host_critical,
            'host2DependantOnTesthost24'
        )
        host3_log = self.handle_log(
            notification_host_critical,
            'host3DependantOnTesthost24'
        )

        with cmn._allow_block_for_test_only():
            host2_log.alert.acknowledged     = True
            host2_log.alert.datetimerecovery = cmn.now()
            host2_log.alert.save()
            host3_log.alert.acknowledged     = True
            host3_log.alert.save()
        with mock.patch(
            'symonitoring_alert.logdata.LogData.check_timeframe'
        ) as check_tf_mocked, mock.patch(
            'symonitoring_alert.logdata.LogData.check_team'
        ) as check_team_mocked:
            check_tf_mocked.return_value = True
            check_team_mocked.return_value = True
            self.handle_log(
                notification_host_critical,
                'testhost24'
            )
        with cmn._allow_block_for_test_only():
            alert_host1 = mod.Alert.objects.get(
                pk=host1_log.alert.pk
            )
            alert_host2 = mod.Alert.objects.get(
                pk=host2_log.alert.pk
            )
            alert_host3 = mod.Alert.objects.get(
                pk=host3_log.alert.pk
            )
        self.assertTrue(alert_host1.acknowledged)
        self.assertTrue(alert_host1.acknowledgedautomatically)
        self.assertIsNotNone(alert_host1.datetimerecovery)
        self.assertTrue(alert_host2.acknowledged)
        self.assertFalse(alert_host2.acknowledgedautomatically)
        self.assertIsNotNone(alert_host2.datetimerecovery)
        self.assertTrue(alert_host3.acknowledged)
        self.assertFalse(alert_host3.acknowledgedautomatically)
        self.assertIsNotNone(alert_host3.datetimerecovery)

    # This would be the correct behaviour, see ticket #11621
    @cmn.unittest.skipIf(not cmn.is_flag_set("B11621"), "in dev")
    def test_host_dependencies_services_down(self):
        service1_log = self.handle_log(
            notification_service_critical,
            'host1DependantOnTesthost24',
            'ssh__instance1'
        )
        service2_log = self.handle_log(
            notification_service_critical,
            'host1DependantOnTesthost24',
            'ssh__instance2'
        )
        service3_log = self.handle_log(
            notification_service_critical,
            'host1DependantOnTesthost24',
            'ssh2'
        )
        with cmn._allow_block_for_test_only():
            service2_log.alert.acknowledged     = True
            service2_log.alert.datetimerecovery = cmn.now()
            service2_log.alert.save()
            service3_log.alert.acknowledged     = True
            service3_log.alert.save()
        with mock.patch(
            'symonitoring_alert.logdata.LogData.check_timeframe'
        ) as check_tf_mocked, mock.patch(
            'symonitoring_alert.logdata.LogData.check_team'
        ) as check_team_mocked:
            check_tf_mocked.return_value = True
            check_team_mocked.return_value = True
            self.handle_log(
                notification_host_critical,
                'testhost24'
            )
        with cmn._allow_block_for_test_only():
            alert_service1 = mod.Alert.objects.get(
                pk=service1_log.alert.pk
            )
            alert_service2 = mod.Alert.objects.get(
                pk=service2_log.alert.pk
            )
            alert_service3 = mod.Alert.objects.get(
                pk=service3_log.alert.pk
            )
        self.assertTrue(alert_service1.acknowledged)
        self.assertTrue(alert_service1.acknowledgedautomatically)
        self.assertIsNotNone(alert_service1.datetimerecovery)
        self.assertTrue(alert_service2.acknowledged)
        self.assertFalse(alert_service2.acknowledgedautomatically)
        self.assertIsNotNone(alert_service2.datetimerecovery)
        self.assertTrue(alert_service3.acknowledged)
        self.assertFalse(alert_service3.acknowledgedautomatically)
        self.assertIsNotNone(alert_service3.datetimerecovery)
