# -*- coding: UTF-8 -*-

"""
Handle informative events
"""

# System imports
import itertools
import datetime

# Project imports
import symonitoring_alert.common                as cmn
import symonitoring_alert.constants             as const
import symonitoring_model.models                as mod
import symonitoring_alert.communication.inform  as sc
import symonitoring_alert.dummy                 as dummy

s = cmn.get_settings()


class InformInit(cmn.ExecutorMixin, cmn.LogDataLockMixin):
    """Find info-stream events"""
    @cmn.coroutine
    def handle(self):
        """Main handler function"""
        yield self.query_collect_infofeed_data()

    @cmn.run_on_executor
    def query_collect_infofeed_data(self):
        """Retrieve infofeeds which match the current alert"""
        # pylint: disable=line-too-long
        logdata = self.logdata
        infofeeds_host = mod.Infofeed.objects.select_related().filter(
            infofeedhost__hostid=logdata.host
        )
        infofeeds_hostgroup = mod.Infofeed.objects.select_related().filter(
            infofeedhostgroup__hostgroupid__hostgrouphost__hostid=logdata.host
        ).exclude(
            infofeedhostgroup__infofeedhostgroupexcludedhosts__hostid=
            logdata.host
        )
        customers = list(itertools.chain(
            [logdata.host.customerid],
            self._get_super_customers(logdata.host.customerid)
        ))
        infofeeds_customer = mod.Infofeed.objects.select_related().filter(
            infofeedcustomer__customerid__in=customers
        ).exclude(
            infofeedcustomer__infofeedcustomerexcludedhosts__hostid=
            logdata.host
        )
        logdata.infofeeds = list(
            set(
                itertools.chain(
                    infofeeds_host,
                    infofeeds_hostgroup,
                    infofeeds_customer
                )
            )
        )
        for feed in logdata.infofeeds:
            feed.ext_infofeednotificationtype = list(
                feed.infofeednotificationtype_set.select_related()
            )

    def _get_super_customers(self, customer):
        """Recursively collects all super customers of a given customer"""
        if customer.parentid is None:
            return
        yield customer.parentid
        # The yield from-syntax only exists in python 3.3 or later
        for parent in self._get_super_customers(customer.parentid):
            yield parent


class InformUsersHandler(object):
    """Handle a request/callback to inform users"""

    def __init__(self, feed, state, action, dummy, logdata):
        self.feed    = feed
        self.state   = state
        self.action  = action
        self.dummy   = dummy
        self.logdata = logdata

    @cmn.coroutine
    def handle(self):
        """Send the notification to all users of the feed"""
        force_inform_users = yield self.dummy.force_inform_users(
            [self.feed.userid]
        )
        users_exec = InformUsersExec(self.logdata)
        for user in force_inform_users:
            yield users_exec.dispatch_action(
                self.dummy.force_action(self.action.slaactionid_id),
                user,
                self.state
            )


class InformUsersExec(cmn.ExecutorMixin):
    """Inform users"""

    def __init__(self, logdata, *args, **kwargs):
        super(InformUsersExec, self).__init__(*args, **kwargs)
        self.logdata = logdata
        self.dummy   = dummy.Dummy()

    @cmn.run_on_executor
    def get_template_ex(self, user, notification_type):
        """Memoized version of get_template bound to this object instance"""

        return self.get_template(
            user,
            notification_type
        )

    @cmn.memoize
    def get_template(self, user, notification_type):
        """Memoized version of get_template bound to this object instance"""

        return sc.UserCommunicationInform.get_template(
            user,
            notification_type
        )

    @cmn.run_on_executor
    def handle_EMAIL(self, user):
        """Send alert via email

        :type timeframe: symonitoring_model.models.Timeframe
        :type user: symonitoring_model.models.User"""

        res = False
        if user.email:
            template = self.get_template(
                user,
                const.ActionType.EMAIL
            )
            comm = sc.UserCommunicationInform(
                logdata   = self.logdata,
                user      = user,
                template  = template,
            )
            res = comm.send_email()
        if not res:
            cmn.lg.debug(
                "Could not send mail to: %s (%s)", user.email, user.fullname
            )
        return res

    @cmn.coroutine
    def handle_XMPP(self, user, orig_action):
        """Send alert via XMPP"""

        res = False
        if user.jabber:
            template = yield self.get_template_ex(
                user,
                const.ActionType.XMPP
            )
            comm = sc.UserCommunicationInform(
                logdata   = self.logdata,
                user      = user,
                template  = template,
            )
            res = yield comm.send_xmpp(orig_action=orig_action)

        if not res:
            cmn.lg.debug(
                "Could not send xmpp to: %s (%s)", user.jabber, user.fullname
            )
        raise cmn.Return(res)

    @cmn.coroutine
    def handle_MATTERMOST(self, user):
        res = False
        if user.mattermost:
            template = yield self.get_template_ex(
                user,
                const.ActionType.MATTERMOST
            )
            comm = sc.UserCommunicationInform(
                logdata   = self.logdata,
                user      = user,
                template  = template,
            )
            res = yield comm.send_mattermost()

        if not res:
            cmn.lg.debug(
                "Could not send mattermost message to: %s (%s)",
                user.mattermost, user.fullname
            )
        raise cmn.Return(res)

    @cmn.coroutine
    def handle_PHONE(self, user):
        """Send alert via phone

        :type user: symonitoring_model.models.User"""
        res = False
        if user.phone:
            comm = sc.UserCommunicationInform(
                logdata   = self.logdata,
                user      = user,
                template  = None,
            )
            res = yield comm.call_phone()
        if not res:
            cmn.lg.debug("Could not call: %s (%s)", user.phone, user.fullname)
        raise cmn.Return(res)

    @cmn.coroutine
    def handle_SMS(self, user):
        """Send alert via SMS"""

        res = False
        if user.sms:
            template = yield self.get_template_ex(
                user,
                const.ActionType.SMS
            )
            comm = sc.UserCommunicationInform(
                logdata   = self.logdata,
                user      = user,
                template  = template,
            )
            res = yield comm.send_sms()
        if not res:
            cmn.lg.debug(
                "Could not send sms to: %s (%s)", user.sms, user.fullname
            )
        raise cmn.Return(res)

    @cmn.coroutine
    def dispatch_action(
            self,
            actionid,
            user,
            type_ = None
    ):
        """Executes the action

        :type user: symonitoring_model.models.User
        """
        res = False
        action_id = self.dummy.force_action(actionid)
        if action_id == const.ActionType.EMAIL:
            res = yield self.handle_EMAIL(user)
        elif action_id == const.ActionType.SMS:
            res = yield self.handle_SMS(user)
        elif action_id == const.ActionType.PHONE:
            res = yield self.handle_PHONE(user)
        elif action_id == const.ActionType.MATTERMOST:
            res = yield self.handle_MATTERMOST(user)
        elif action_id == const.ActionType.XMPP:
            if action_id != actionid:
                res = yield self.handle_XMPP(
                    user,
                    orig_action=actionid,
                )
            else:
                res = yield self.handle_XMPP(
                    user,
                    orig_action=action_id,
                )
        else:
            cmn.lg.debug("Unknown action: %s", action_id)
        action = action_id
        try:
            action = const.action_type_dict[action]
        except IndexError:
            pass
        if res:
            cmn.lg.debug(
                "Successfully informed %s via %s for type %s",
                user.fullname,
                action,
                type_
            )
        else:
            cmn.lg.warning(
                "Failed to inform %s via %s for type %s",
                user.fullname,
                action,
                type_
            )


class InformExec(cmn.ExecutorMixin, cmn.LogDataLockMixin):
    """Handle info-stream events"""

    QUIET_STATES = set([
        const.NagiosLog.STATE_NO_PREVIOUS_DOWN,
        const.NagiosLog.STATE_MALFORMED_ROW,
        const.NagiosLog.STATE_HOST_DISABLED,
        const.NagiosLog.STATE_ALERT_REPETITION,
        const.NagiosLog.STATE_IN_MAINTENANCE,
        const.NagiosLog.STATE_SOFT,
        const.NagiosLog.STATE_HOST_DOWN,
        const.NagiosLog.STATE_DEPENDENCY_DOWN,
        const.NagiosLog.STATE_DOES_NOT_EXIST,
        const.NagiosLog.STATE_GLOBAL_DISABLE,
    ])

    def __init__(self, *args, **kwargs):
        super(InformExec, self).__init__(*args, **kwargs)
        self.debug_last = ""
        self.dummy      = dummy.Dummy()

    @cmn.coroutine
    def dispatch_feed(self, feed, state):
        """Dispatch the given feed for the given type."""
        for action in feed.ext_infofeednotificationtype:
            inform_users = InformUsersHandler(
                feed,
                state,
                action,
                self.dummy,
                self.logdata
            )
            if action.slaactionid.delay == 0:
                inform_users.handle()
            else:
                delay = datetime.timedelta(minutes=action.slaactionid.delay)
                cmn.add_timeout(delay, inform_users.handle)

    @cmn.coroutine
    def handle(self):
        """Schedules a handler for inform exec"""
        cmn.add_callback(self.handle_async)

    @cmn.coroutine
    def handle_async(self):
        """Main handler function"""
        # pylint: disable=too-many-branches
        with (yield self.lock.acquire()):
            logdata = self.logdata
            alert   = logdata.alert.model

            if logdata.log.processed in InformExec.QUIET_STATES:
                cmn.lg.warn(
                    "Not informing for %d as state %s is in QUIET_STATES" %
                    (
                        logdata.log.id,
                        const.nagios_log_dict[logdata.log.processed]
                    )
                )
                return

            if not logdata.is_soft and not logdata.is_from_acknowledge:
                if logdata.is_warning or (
                        alert and alert.state.lower() == "warning"
                ):
                    for feed in logdata.infofeeds:
                        if feed.onwarning:
                            yield self.dispatch_feed(feed, "warning")
                else:
                    for feed in logdata.infofeeds:
                        if feed.onalert:
                            yield self.dispatch_feed(feed, "alert")
            if alert and alert.acknowledged:
                if not logdata.acknowledged_processed and not (
                        logdata.is_warning or (
                            alert and alert.state.lower() == "warning"
                        )
                ):
                    for feed in logdata.infofeeds:
                        if feed.onacknowledge:
                            yield self.dispatch_feed(feed, "acknowledge")
                elif logdata.is_ok:
                    for feed in logdata.infofeeds:
                        if feed.onalert and not alert.state.lower(
                        ) == "warning":
                            yield self.dispatch_feed(feed, "up")
                        else:
                            if feed.onwarning and alert.state.lower(
                            ) == "warning":
                                yield self.dispatch_feed(feed, "up")
                            elif (
                                    feed.onacknowledge and
                                    alert.acknowledgedby_id
                            ):
                                yield self.dispatch_feed(feed, "up")

                alert.acknowledgedprocessed = True
                yield self.save_alert()

    @cmn.run_on_executor
    def save_alert(self):
        """Saves the alert"""
        self.logdata.alert.save()
