# -*- coding: UTF-8 -*-

"""
Testing handling
"""

# pylint: disable=missing-docstring,import-error,protected-access

# System imports
import mock
import datetime
import itertools
import tornado.gen                          as gen
import django.template                      as dtempl

# Project imports
import symonitoring_alert.common            as cmn
import symonitoring_model.models            as mod
import symonitoring_alert.handle.nagioslog  as nl
import symonitoring_alert.constants         as const
import symonitoring_alert.handle.inform     as info
import symonitoring_alert.default_templates as deftempl

notification_host_critical = {
    'address': u'192.24.12.3',
    'attempt': 1,
    'checkcommand': u'check-host-alive',
    'datetime': cmn.local_datetime(datetime.datetime(
        2014, 5, 6, 15, 19, 7
    )),
    'duration': u'0d 0h 0m 0s',
    'hostname': u'localhost',
    'lastproblemid': 47,
    'name': u'localhost',
    'objecttype': u'HOST',
    'output': u'PING CRITICAL - Packet loss = 100%',
    'perfdata': u'rta=5000.000000ms;3000.000000;5000.000000;0.000000 pl=100%;'
                '80;100;0',
    'problemid': 48,
    'processed': u'0',
    'state': u'DOWN',
    'statetype': u'HARD',
    'timestamp': cmn.local_datetime(datetime.datetime(
        2014, 5, 6, 15, 19, 7
    ))
}

notification_host_service_critical = {
    'address': u'192.24.12.3',
    'attempt': 1,
    'checkcommand': u'check-host-alive',
    'datetime': cmn.local_datetime(datetime.datetime(
        2014, 5, 6, 15, 19, 7
    )),
    'duration': u'0d 0h 0m 0s',
    'hostname': u'localhost',
    'lastproblemid': 47,
    'name': u'ssh__default',
    'objecttype': u'SERVICE',
    'output': u'PING CRITICAL - Packet loss = 100%',
    'perfdata': u'rta=5000.000000ms;3000.000000;5000.000000;0.000000 pl=100%;'
                '80;100;0',
    'problemid': 48,
    'processed': u'0',
    'state': u'DOWN',
    'statetype': u'HARD',
    'timestamp': cmn.local_datetime(datetime.datetime(
        2014, 5, 6, 15, 19, 7
    ))
}

notification_host_warning = {
    'address': u'192.24.12.3',
    'attempt': 1,
    'checkcommand': u'check-host-alive',
    'datetime': cmn.local_datetime(datetime.datetime(
        2014, 5, 6, 15, 19, 7
    )),
    'duration': u'0d 0h 0m 0s',
    'hostname': u'localhost',
    'lastproblemid': 47,
    'name': u'localhost',
    'objecttype': u'HOST',
    'output': u'PING CRITICAL - Packet loss = 100%',
    'perfdata': u'rta=5000.000000ms;3000.000000;5000.000000;0.000000 pl=100%;'
                '80;100;0',
    'problemid': 48,
    'processed': u'0',
    'state': u'warning',
    'statetype': u'HARD',
    'timestamp': cmn.local_datetime(datetime.datetime(
        2014, 5, 6, 15, 19, 7
    ))
}

notification_host_ok = {
    'address': u'192.24.12.3',
    'attempt': 1,
    'checkcommand': u'check-host-alive',
    'datetime': cmn.local_datetime(datetime.datetime(
        2014, 5, 6, 15, 19, 7
    )),
    'duration': u'0d 0h 0m 0s',
    'hostname': u'localhost',
    'lastproblemid': 48,
    'name': u'localhost',
    'objecttype': u'HOST',
    'output': u'PING OK - Packet loss = 0%, RTA = 1.87 ms',
    'perfdata': u'rta=5000.000000ms;3000.000000;5000.000000;0.000000 pl=100%;'
                '80;100;0',
    'problemid': 49,
    'processed': u'0',
    'state': u'UP',
    'statetype': u'HARD',
    'timestamp': cmn.local_datetime(datetime.datetime(
        2014, 5, 6, 15, 19, 7
    ))
}


class InformTest(cmn.TestCase):

    def run_init(self, notification, hostname):
        """Runs a notification"""
        handler = nl.NagiosLog()
        log = mod.Nagioslog(
            **notification
        )
        log.hostname = hostname
        self.save_log(log)

        @gen.coroutine
        def coro():
            yield handler.handle(log, no_exec=True)

        cmn.run_sync(coro)
        return handler.logdata

    def test_query_localhost(self):
        res = self.run_init(
            notification_host_critical,
            "localhost"
        )
        self.assertEqual(1, len(res.infofeeds))
        feed = res.infofeeds[0]
        self.assertIsInstance(feed, mod.Infofeed)
        self.assertEqual(1, feed.id)
        self.assertTrue(feed.onalert)
        self.assertFalse(feed.onwarning)
        self.assertFalse(feed.onacknowledge)
        self.assertEqual(
            feed.ext_infofeednotificationtype[0].slaactionid_id,
            const.ActionType.EMAIL
        )

    def test_query_testhost1(self):
        res = self.run_init(
            notification_host_critical,
            "testhost1"
        )
        self.assertEqual(1, len(res.infofeeds))
        feed = res.infofeeds[0]
        self.assertIsInstance(feed, mod.Infofeed)
        self.assertEqual(2, feed.id)
        self.assertFalse(feed.onalert)
        self.assertTrue(feed.onwarning)
        self.assertFalse(feed.onacknowledge)
        self.assertEqual(
            feed.ext_infofeednotificationtype[0].slaactionid_id,
            const.ActionType.XMPP
        )

    def test_query_testhost22(self):
        res = self.run_init(
            notification_host_critical,
            "testhost22"
        )
        self.assertEqual(1, len(res.infofeeds))
        feed = res.infofeeds[0]
        self.assertIsInstance(feed, mod.Infofeed)
        self.assertEqual(3, feed.id)
        self.assertFalse(feed.onalert)
        self.assertFalse(feed.onwarning)
        self.assertTrue(feed.onacknowledge)
        self.assertEqual(
            feed.ext_infofeednotificationtype[0].slaactionid_id,
            const.ActionType.PHONE
        )

    def test_query_testhost2(self):
        res = self.run_init(
            notification_host_critical,
            "testhost2"
        )
        self.assertEqual(0, len(res.infofeeds))

    def test_query_testhost3(self):
        res = self.run_init(
            notification_host_critical,
            "testhost3"
        )
        self.assertEqual(1, len(res.infofeeds))
        feed = res.infofeeds[0]
        self.assertIsInstance(feed, mod.Infofeed)
        self.assertEqual(4, feed.id)
        self.assertTrue(feed.onalert)
        self.assertFalse(feed.onwarning)
        self.assertFalse(feed.onacknowledge)
        self.assertEqual(
            feed.ext_infofeednotificationtype[0].slaactionid_id,
            const.ActionType.SMS
        )

    def test_query_testhost23(self):
        res = self.run_init(
            notification_host_critical,
            "testhost23"
        )
        self.assertEqual(0, len(res.infofeeds))

    def test_query_testhost24(self):
        res = self.run_init(
            notification_host_critical,
            "testhost24"
        )
        self.assertEqual(1, len(res.infofeeds))
        feed = res.infofeeds[0]
        self.assertIsInstance(feed, mod.Infofeed)
        self.assertEqual(5, feed.id)
        self.assertFalse(feed.onalert)
        self.assertTrue(feed.onwarning)
        self.assertFalse(feed.onacknowledge)
        self.assertEqual(
            feed.ext_infofeednotificationtype[0].slaactionid_id,
            const.ActionType.XMPP
        )

    def dispatch_mocked(self, notification, hostname=None):
        """Runs a notification"""
        handler = nl.NagiosLog()
        log = mod.Nagioslog(
            **notification
        )
        if hostname is not None:
            log.hostname = hostname
        self.save_log(log)

        @gen.coroutine
        def coro():
            yield handler.handle(log)
            yield cmn.wait(datetime.timedelta(
                seconds=0.05 * cmn.s.TIMING_TEST_SCALE
            ))

        with cmn._allow_add_handler_test_only(), mock.patch(
            "symonitoring_alert.handle.inform.InformUsersExec.dispatch_action"
        ) as op, mock.patch(
            "symonitoring_alert.common.add_timeout"
        ):
            op.return_value = cmn.wait(0)
            cmn.run_sync(coro)
        return (handler, op)

    def test_dispatch_warning_testhost1(self):
        _, op = self.dispatch_mocked(
            notification_host_warning,
            "testhost1"
        )
        self.assertTrue(op.called)

    def test_dispatch_alert_testhost1(self):
        _, op = self.dispatch_mocked(
            notification_host_critical,
            "testhost1"
        )
        self.assertFalse(op.called)

    def test_dispatch_alert_hostDummy_with_dummy(self):
        old_inform = cmn.s.DUMMY_INFORM
        old_force  = cmn.s.DUMMY_FORCE
        try:
            cmn.s.DUMMY_INFORM = ('admin',)
            cmn.s.DUMMY_FORCE  = 3
            _, op = self.dispatch_mocked(
                notification_host_critical,
                "hostDummy"
            )
            self.assertTrue(op.called)
            self.assertEqual(len(op.call_args_list)  , 1)
            self.assertEqual(op.call_args[0][0]      , const.ActionType.EMAIL)
            self.assertEqual(op.call_args[0][1].name , 'admin')
            self.assertEqual(op.call_args[0][2]      , 'alert')
        finally:
            cmn.s.DUMMY_INFORM = old_inform
            cmn.s.DUMMY_FORCE  = old_force

    def test_dispatch_alert_testhost3(self):
        with cmn.disable_dummy_mode():
            handler, op = self.dispatch_mocked(
                notification_host_critical,
                "testhost3"
            )
        self.assertTrue(op.called)
        self.assertEqual(len(op.call_args_list)  , 1)
        self.assertEqual(op.call_args[0][0]      , const.ActionType.SMS)
        self.assertEqual(op.call_args[0][1].name , 'admin')
        self.assertEqual(op.call_args[0][2]      , 'alert')
        logdata = handler.logdata
        logdata.is_from_acknowledge = True
        logdata.alert.model.acknowledged = True
        inf_exec = info.InformExec()
        inf_exec.logdata = logdata

        @gen.coroutine
        def coro():
            with mock.patch("symonitoring_alert.common.add_timeout"):
                yield inf_exec.handle_async()
        with mock.patch(
            "symonitoring_alert.handle.inform.InformUsersExec.dispatch_action"
        ) as op:
            op.return_value = cmn.wait(0)
            cmn.run_sync(coro)
        self.assertFalse(op.called)

    def test_dispatch_alert_testhost2(self):
        _, op = self.dispatch_mocked(
            notification_host_critical,
            "testhost2"
        )
        self.assertFalse(op.called)

    def test_dispatch_warning_testhost2(self):
        _, op = self.dispatch_mocked(
            notification_host_warning,
            "testhost2"
        )
        self.assertFalse(op.called)

    def test_dispatch_alert_testhost3p(self):
        with cmn.disable_dummy_mode():
            _, op = self.dispatch_mocked(
                notification_host_critical,
                "testhost3"
            )
        self.assertTrue(op.called)
        self.assertEqual(len(op.call_args_list)  ,  1)
        self.assertEqual(op.call_args[0][0]      ,  const.ActionType.SMS)
        self.assertEqual(op.call_args[0][1].name ,  'admin')
        self.assertEqual(op.call_args[0][2]      ,  'alert')

    def test_dispatch_warning_testhost22(self):
        _, op = self.dispatch_mocked(
            notification_host_warning,
            "testhost22"
        )
        self.assertFalse(op.called)

    def test_dispatch_alert_testhost22(self):
        _, op = self.dispatch_mocked(
            notification_host_critical,
            "testhost22"
        )
        self.assertFalse(op.called)

    def test_dispatch_warning_testhost23(self):
        _, op = self.dispatch_mocked(
            notification_host_critical,
            "testhost23"
        )
        self.assertFalse(op.called)

    def test_dispatch_warning_testhost24(self):
        _, op = self.dispatch_mocked(
            notification_host_warning,
            "testhost24"
        )
        self.assertTrue(op.called)

    def test_dispatch_alert_testhost24(self):
        _, op = self.dispatch_mocked(
            notification_host_critical,
            "testhost24"
        )
        self.assertFalse(op.called)

    @cmn.coroutine
    def nop_true(self):
        raise cmn.Return(True)

    def run_dispatch_and_handle(self, func, action_type):
        res = self.run_init(
            notification_host_critical,
            "testhost22"
        )
        handler = info.InformUsersExec(res)
        with mock.patch(
            "symonitoring_alert.communication.inform.UserCommunicationInform"
        ) as op:
            getattr(op.return_value, func).return_value  = self.nop_true()
            timeframe     = res.timeframes[0]
            user          = res.users[timeframe][0]
            with cmn.disable_dummy_mode():
                @cmn.coroutine
                def coro():
                    yield handler.dispatch_action(
                        action_type,
                        user,
                    )
                cmn.run_sync(coro)
                return op

    def test_dispatch_and_handle_phone(self):
        op = self.run_dispatch_and_handle('call_phone', const.ActionType.PHONE)
        self.assertTrue(op.return_value.call_phone.called)

    def test_dispatch_and_handle_sms(self):
        op = self.run_dispatch_and_handle('send_sms', const.ActionType.SMS)
        self.assertTrue(op.return_value.send_sms.called)

    def test_dispatch_and_handle_email(self):
        op = self.run_dispatch_and_handle('send_email', const.ActionType.EMAIL)
        self.assertTrue(op.return_value.send_email.called)

    def test_dispatch_and_handle_xmpp(self):
        op = self.run_dispatch_and_handle('send_xmpp', const.ActionType.XMPP)
        self.assertTrue(op.return_value.send_xmpp.called)

    def test_dispatch_and_handle_mattermost(self):
        op = self.run_dispatch_and_handle('send_mattermost',
                                          const.ActionType.MATTERMOST)
        self.assertTrue(op.return_value.send_mattermost.called)

    def run_dispatch_and_render(self, func, return_value_func, action_type):
        res = self.run_init(
            notification_host_critical,
            "testhost22"
        )
        handler = info.InformUsersExec(res)

        with mock.patch(
            func
        ) as op:
            op.return_value = return_value_func()
            timeframe     = res.timeframes[0]
            user          = res.users[timeframe][0]
            with cmn.disable_dummy_mode():
                @cmn.coroutine
                def coro():
                    yield handler.dispatch_action(
                        action_type,
                        user,
                    )
                cmn.run_sync(coro)
                return (timeframe, user, res, op)

    def run_dispatch_and_render_email(self):
        import email.mime.text as et

        res = self.run_init(
            notification_host_critical,
            "testhost22"
        )
        handler = info.InformUsersExec(res)
        orig = et.MIMEText

        def pass_through(*args, **kwargs):
            return orig(*args, **kwargs)

        with mock.patch(
            "email.mime.text.MIMEText",
        ) as op, mock.patch(
            "smtplib.SMTP.sendmail"
        ) as op2:
            op.side_effect = pass_through
            op2.return_value = {}
            timeframe     = res.timeframes[0]
            user          = res.users[timeframe][0]
            with cmn.disable_dummy_mode():
                @cmn.coroutine
                def coro():
                    yield handler.dispatch_action(
                        const.ActionType.EMAIL,
                        user,
                    )
                cmn.run_sync(coro)
                return (user, res, op, op2)

    def test_dispatch_and_render_xmpp(self):
        @cmn.coroutine
        def coro_res():
            raise cmn.Return(["ok"])

        _, user, logdata, op = self.run_dispatch_and_render(
            "symonitoring_alert.communication.xmpp.send_message",
            coro_res,
            const.ActionType.XMPP
        )
        self.assertTrue(op.called)
        message_body = op.call_args[1]['mbody']
        self.assertNotEqual(message_body, '')

        default_xmpp_template = deftempl.templates[const.ActionType.XMPP]
        tmpl = dtempl.Template(default_xmpp_template)
        tplc = dtempl.Context({
            'part'      : const.TemplatePart.PART_PLAIN,
            'user'      : user,
            'logdata'   : logdata,
            'alert'     : logdata.alert.model,
            'type'      : const.template_type_dict[
                const.TemplateType.INFOFEED
            ].lower(),
            'is_action' : "%s: " % const.action_type_dict[
                const.ActionType.XMPP
            ],
        })
        rendered_tmpl = tmpl.render(tplc).strip()
        self.assertEqual(message_body, rendered_tmpl)

    def test_dispatch_and_render_email(self):
        user, logdata, op, op2 = self.run_dispatch_and_render_email()
        self.assertTrue(op.called)
        message_body = op.call_args[0][0]
        self.assertNotEqual(message_body, '')

        self.assertTrue(op2.called)

        default_template = deftempl.templates[const.ActionType.EMAIL]
        tmpl = dtempl.Template(default_template)
        plain_context = dtempl.Context({
            'part'      : const.TemplatePart.PART_PLAIN,
            'user'      : user,
            'logdata'   : logdata,
            'alert'     : logdata.alert.model,
            'type'      : const.template_type_dict[
                const.TemplateType.INFOFEED
            ].lower(),
            'is_action' : "%s: " % const.action_type_dict[
                const.ActionType.EMAIL
            ],
        })
        plain = tmpl.render(plain_context).strip()

        self.assertEqual(
            message_body,
            plain.encode('utf-8')
        )

    def test_get_super_customers(self):
        # We test the method logic here, not the method's use for infofeed
        with cmn._allow_block_for_test_only():
            rootcustomer = mod.Customer.objects.get(pk=1)
            supercustomer01 = mod.Customer.objects.get(pk=7)
            supercustomer02 = mod.Customer.objects.get(pk=8)
            supercustomer03 = mod.Customer.objects.get(pk=9)
            inform_init = info.InformInit()
            expected_customers = [
                supercustomer02, supercustomer01, rootcustomer
            ]
            customers = list(itertools.chain(
                inform_init._get_super_customers(supercustomer03)
            ))
            self.assertEqual(customers, expected_customers)

    def test_get_super_customers_no_parent(self):
        # We test the method logic here, not the method's use for infofeed
        with cmn._allow_block_for_test_only():
            rootcustomer = mod.Customer.objects.get(pk=1)
            inform_init = info.InformInit()
            customers = list(itertools.chain(
                inform_init._get_super_customers(rootcustomer)
            ))
            self.assertEqual(customers, [])

    def test_dispatch_alert_super_customer(self):
        with cmn._allow_block_for_test_only():
            res = self.run_init(
                notification_host_critical,
                "supercustomerHost01"
            )
            infofeed1 = mod.Infofeed.objects.get(pk=6)
            infofeed2 = mod.Infofeed.objects.get(pk=7)
            self.assertEqual(res.infofeeds, [infofeed1, infofeed2])

    def test_global_disable(self):
        # pylint: disable=protected-access
        with mock.patch(
            'symonitoring_alert.query.QueryBase.global_disable'
        ) as global_disable_mocked:
            global_disable_mocked.return_value = True
            _, op = self.dispatch_mocked(
                notification_host_critical,
                "testhost1"
            )
        self.assertFalse(op.called)

    def test_up_global_disable_after_alert(self):
        # --> Alert
        # --> Global disable
        # --> OK for alert
        #     --> NO alerting and informing!

        # pylint: disable=protected-access
        with mock.patch(
            'symonitoring_alert.logdata.LogData.check_timeframe'
        ) as check_tf_mocked, mock.patch(
            'symonitoring_alert.logdata.LogData.check_team'
        ) as check_team_mocked:
            check_tf_mocked.return_value = True
            check_team_mocked.return_value = True
            _, op_down = self.dispatch_mocked(
                notification_host_critical,
                "host1DependantOnTesthost24"
            )
        self.assertTrue(op_down.called)
        with mock.patch(
            'symonitoring_alert.query.QueryBase.global_disable'
        ) as global_disable_mocked:
            global_disable_mocked.return_value = True
            _, op_up = self.dispatch_mocked(
                notification_host_ok,
                "host1DependantOnTesthost24"
            )
        self.assertFalse(op_up.called)

    def test_up_after_alert_global_disable(self):
        # --> Alert
        # --> OK for alert
        # --> Global disable
        #     --> NO alerting and informing!

        # pylint: disable=protected-access
        with mock.patch(
            'symonitoring_alert.logdata.LogData.check_timeframe'
        ) as check_tf_mocked, mock.patch(
            'symonitoring_alert.logdata.LogData.check_team'
        ) as check_team_mocked, mock.patch(
            'symonitoring_alert.query.QueryBase.global_disable'
        ) as global_disable_mocked:
            check_tf_mocked.return_value = True
            check_team_mocked.return_value = True
            _, op_down = self.dispatch_mocked(
                notification_host_critical,
                "host1DependantOnTesthost24"
            )
            global_disable_mocked.return_value = True
            _, op_up = self.dispatch_mocked(
                notification_host_ok,
                "host1DependantOnTesthost24"
            )
        self.assertFalse(op_down.called)
        self.assertFalse(op_up.called)

    def test_service_down_after_host_down(self):
        patch_tf = 'symonitoring_alert.logdata.LogData.check_timeframe'
        # First localhost goes down, ...
        with mock.patch(patch_tf) as check_tf_mocked:
            check_tf_mocked.return_value = False
            _, op_down = self.dispatch_mocked(
                notification_host_critical
            )
        self.assertTrue(op_down.called)
        # ... then the assigned service ssh.
        with mock.patch(patch_tf) as check_tf_mocked:
            check_tf_mocked.return_value = False
            _, op_down = self.dispatch_mocked(
                notification_host_service_critical
            )
        self.assertFalse(op_down.called)

    def test_dependency_down(self):
        with mock.patch(
            'symonitoring_alert.logdata.LogData.check_timeframe'
        ) as check_tf_mocked, mock.patch(
            'symonitoring_alert.logdata.LogData.check_team'
        ) as check_team_mocked:
            check_tf_mocked.return_value = True
            check_team_mocked.return_value = True

            with cmn.disable_dummy_mode():
                handler, op = self.dispatch_mocked(
                    notification_host_critical,
                    "testhost24"
                )
            self.assertEqual(
                int(handler.logdata.log.processed),
                const.NagiosLog.STATE_ALERTED
            )
            # No infofeeds, no informing
            self.assertFalse(op.called)

            handler, op = self.dispatch_mocked(
                notification_host_critical,
                "host1DependantOnTesthost24"
            )
            self.assertEqual(
                int(handler.logdata.log.processed),
                const.NagiosLog.STATE_DEPENDENCY_DOWN
            )
            # Assure we do have infofeeds here
            self.assertGreaterEqual(len(handler.logdata.infofeeds), 1)
            # ..but no informing due to status!
            self.assertFalse(op.called)

    def test_dispatch_feed_delayed(self):
        with cmn._allow_block_for_test_only():
            slaaction = mod.Slaaction.objects.get(pk=4)
            test_delay = 5
            orig_delay = slaaction.delay
            slaaction.delay = test_delay
            slaaction.save()
        try:
            with cmn.disable_dummy_mode():
                (handler, op) = self.dispatch_mocked(
                    notification_host_critical,
                    "host1DependantOnTesthost24"
                )
            logdata = handler.logdata
            logdata.is_from_acknowledge = True
            logdata.alert.model.acknowledged = True
            inf_exec = info.InformExec()
            inf_exec.logdata = logdata

            @gen.coroutine
            def coro():
                yield inf_exec.handle_async()
            with mock.patch(
                'symonitoring_alert.handle.inform'
                '.InformUsersExec.dispatch_action'
            ) as op, mock.patch(
                'symonitoring_alert.common.add_timeout'
            ) as add:
                op.return_value = cmn.wait(0)
                cmn.run_sync(coro)
                self.assertTrue(add.called)
                self.assertEqual(
                    add.call_args[0][0],
                    datetime.timedelta(minutes=test_delay)
                )
        finally:
            with cmn._allow_block_for_test_only():
                slaaction.delay = orig_delay
                slaaction.save()
