# -*- coding: UTF-8 -*-

"""
Handle nagioslog
"""

# System imports

# Project imports
import symonitoring_alert.common                as cmn
import symonitoring_alert.handle.alert          as al
import symonitoring_alert.handle.hold           as hld
import symonitoring_alert.handle.hold_inform    as hldi
import symonitoring_alert.handle.inform         as ir
import symonitoring_alert.logdata               as ld


class NagiosLog(cmn.ExecutorMixin, cmn.LogDataLockMixin):
    """Handle notification log, these will be converted to
    alerts/infostream"""

    def __init__(self):
        super(NagiosLog, self).__init__()
        # Alerts can be handled in different ways
        self._logdata = None
        self.exec_handlers = [
            al.AlertExec,
            ir.InformExec,
            hld.HoldExec,
            hldi.HoldInformExec,
        ]
        self.init_handlers = [
            al.AlertInit,
            ir.InformInit,
        ]

    @cmn.run_on_executor
    def save_ex(self, log):
        """ Run save() on executor """
        return self.save(log)

    def save(self, log):
        """Save the log entry"""
        if self.logdata:
            return self.logdata.save()
        else:
            return log.save()

    @cmn.coroutine
    def handle(self, log, no_init=False, no_exec=False):
        """Main handler function"""
        # Mark log entry as processed
        logdata = ld.LogData(log)
        try:
            yield logdata.load_ex()
        except ld.MalformedRow:
            logdata = None
        # WE HAVE TO SET THIS FOR THE LOCKING!
        self.logdata = logdata
        # This lock is needed to prevent overlapping timeouts on the same
        # logdata. The lock has to be locked here and on all entry points
        # from add_timeout. (see mktimeout in alert.py)
        if no_init and no_exec:
            return
        with (yield self.lock.acquire()):
            if logdata:
                # Send notification to all handlers
                if not no_init:
                    for handler in self.init_handlers:
                        try:
                            handler_obj = handler()
                            handler_obj.logdata = logdata
                            yield handler_obj.handle()
                        except Exception as e:  # pylint: disable=broad-except
                            # Isolate errors between handlers
                            cmn.lg.exception("Handler raised an exception")
                            if cmn.raise_handler_needed(e):
                                raise
                if not no_exec:
                    for handler in self.exec_handlers:  # pragma: no cover
                        try:
                            handler_obj = handler()
                            handler_obj.logdata = logdata
                            yield handler_obj.handle()
                        except Exception as e:  # pylint: disable=broad-except
                            # Isolate errors between handlers
                            cmn.lg.exception("Handler raised an exception")
                            if cmn.raise_handler_needed(e):
                                raise
            # Even if we get no logdata, processed will be changed
            # by error handlers. -> save
            yield self.save_ex(log)
