# -*- coding: UTF-8 -*-

"""
Testing handling of nagios log entries
"""

# pylint: disable=missing-docstring,unused-argument,protected-access

# System imports
import mock
import datetime
import tornado.gen              as gen

# Project imports
import symonitoring_alert.common           as cmn
import symonitoring_model.models           as mod
import symonitoring_alert.constants        as const
import symonitoring_alert.handle.nagioslog as nl
import symonitoring_alert.logdata          as ld

s = cmn.get_settings()

timeframe_in = {
    'friday': 1,
    'id': 123,
    'monday': 1,
    'saturday': 1,
    'slaid_id': 2,
    'sunday': 1,
    'thursday': 1,
    'timefrom': datetime.time(10, 0),
    'timeto': datetime.time(9, 30),
    'tuesday': 1,
    'wednesday': 1
}

notification_host_critical_out_of_sla = {
    'address': u'192.24.12.3',
    'attempt': 1,
    'checkcommand': u'check-host-alive',
    'datetime': cmn.local_datetime(datetime.datetime(
        2014, 5, 6, 15, 19, 7
    )),
    'duration': u'0d 0h 0m 0s',
    'hostname': u'testhost24',
    'id': 123,
    'lastproblemid': 47,
    'name': u'testhost24',
    'objecttype': u'HOST',
    'output': u'PING CRITICAL - Packet loss = 100%',
    'perfdata': u'rta=5000.000000ms;3000.000000;5000.000000;0.000000 pl=100%;'
                '80;100;0',
    'problemid': 48,
    'processed': u'0',
    'state': u'DOWN',
    'statetype': u'HARD',
    'timestamp': cmn.local_datetime(datetime.datetime(
        2014, 5, 6, 15, 19, 7
    ))
}

notification_service_critical_out_of_sla = {
    'address': u'192.24.12.3',
    'attempt': 1,
    'checkcommand': u'tcp!21',
    'datetime': cmn.local_datetime(datetime.datetime(
        2014, 5, 6, 15, 19, 7
    )),
    'duration': u'0d 0h 0m 0s',
    'hostname': u'testhost24',
    'id': 123,
    'lastproblemid': 47,
    'name': u'ssh__default',
    'objecttype': u'SERVICE',
    'output': u'CONNECTION REFUSED',
    'perfdata': u'rta=5000.000000ms;3000.000000;5000.000000;0.000000 pl=100%;'
                '80;100;0',
    'problemid': 48,
    'processed': u'0',
    'state': u'CRITICAL',
    'statetype': u'HARD',
    'timestamp': cmn.local_datetime(datetime.datetime(
        2014, 5, 6, 15, 19, 7
    ))
}

notification_service_ok_out_of_sla = {
    'address': u'192.24.12.3',
    'attempt': 1,
    'checkcommand': u'tcp!21',
    'datetime': cmn.local_datetime(datetime.datetime(
        2014, 5, 6, 15, 28, 7
    )),
    'duration': u'0d 0h 0m 0s',
    'hostname': u'testhost24',
    'id': 124,
    'lastproblemid': 47,
    'name': u'ssh__default',
    'objecttype': u'SERVICE',
    'output': u'TCP OK - 0.028 second response time on port 21',
    'perfdata': u'rta=5000.000000ms;3000.000000;5000.000000;0.000000 pl=100%;'
                '80;100;0',
    'problemid': 48,
    'processed': u'0',
    'state': u'OK',
    'statetype': u'HARD',
    'timestamp': cmn.local_datetime(datetime.datetime(
        2014, 5, 6, 15, 28, 7
    ))
}


class NagiosLogTest(cmn.TestCase):
    @cmn.run_on_executor
    def get_host_ex(self, hostname):
        return mod.Host.objects.filter(
            hostname=hostname
        ).select_related()[0]

    def get_host(self, hostname):
        return self.get_host_ex(hostname=hostname)

    @cmn.run_on_executor
    def get_timeframes_ex(self, log, host, timeframe):
        return ld.LogData.filter_timeframe_users(host, [timeframe])

    def get_timeframes(self, log, host, timeframe):
        return self.get_timeframes_ex(log, host, timeframe)

    def timeframe_mocked(self, timeframe, hostname, now, timefrom, timeto):
        """Mocks the given timeframe
        under respect of the given value for the
        current point in time (`now`)."""
        log = nl.NagiosLog()
        tf = mod.Timeframe(
            **timeframe_in
        )
        tf.timefrom = timefrom
        tf.timeto = timeto

        @gen.coroutine
        def coro():
            host = yield self.get_host(hostname)
            (tfs, users) = yield self.get_timeframes(log, host, tf)
            raise cmn.Return((tfs, users))

        with mock.patch(
            "symonitoring_alert.common.now"
        ) as op:
            op.return_value = now
            (tfs, users) = cmn.run_sync(coro)

        return (tfs, users, op)

    def test_filter_timeframes_from_is_bigger_to_is_same(self):
        # Tests the filtering of timeframes when
        # the `from` attribute of a timeframe is bigger than
        # the current point in time (now) and the `to` attribute
        # of a timeframe is the same as the current point in time.

        # from > now = to

        # from:   1000
        # to:     0930
        # now:    0930

        now = datetime.datetime.now()
        now_mocked = cmn.local_datetime(
            datetime.datetime(now.year, now.month, now.day, 9, 30)
        )

        (tfs, _, _) = self.timeframe_mocked(
            timeframe_in,
            "testhost22",
            now_mocked,
            datetime.time(10, 0),
            datetime.time(9, 30)
        )
        self.assertEqual(1, len(tfs))

    def test_filter_timeframes_from_is_same_to_is_bigger_than(self):
        # Tests the filtering of timeframes when
        # the `from` attribute of a timeframe is the same as
        # the current point in time (now) and the `to` attribute
        # of a timeframe is bigger than the current point in time.

        # from = now < to

        # from:   1000
        # to:     0930
        # now:    1000

        now = datetime.datetime.now()
        now_mocked = cmn.local_datetime(
            datetime.datetime(now.year, now.month, now.day, 10, 0)
        )

        (tfs, _, _) = self.timeframe_mocked(
            timeframe_in,
            "testhost22",
            now_mocked,
            datetime.time(10, 0),
            datetime.time(9, 30)
        )
        self.assertEqual(0, len(tfs))

    def test_filter_timeframes_from_is_smaller_to_is_same(self):
        # Tests the filtering of timeframes when
        # the `from` attribute of a timeframe is smaller as
        # the current point in time (now) and the `to` attribute
        # of a timeframe is the same as the current point in time.

        # from < now = to

        # from:   1000
        # to:     1030
        # now:    1030

        now = datetime.datetime.now()
        now_mocked = cmn.local_datetime(
            datetime.datetime(now.year, now.month, now.day, 10, 30)
        )

        (tfs, _, _) = self.timeframe_mocked(
            timeframe_in,
            "testhost22",
            now_mocked,
            datetime.time(10, 0),
            datetime.time(10, 30)
        )
        self.assertEqual(1, len(tfs))

    def test_filter_timeframes_from_is_same_to_is_bigger(self):
        # Tests the filtering of timeframes when
        # the `from` attribute of a timeframe is the same as
        # the current point in time (now) and the `to` attribute
        # of a timeframe is bigger as the current point in time.

        # from = now < to

        # from:   1000
        # to:     1030
        # now:    1000

        now = datetime.datetime.now()
        now_mocked = cmn.local_datetime(
            datetime.datetime(now.year, now.month, now.day, 10, 0)
        )

        (tfs, _, _) = self.timeframe_mocked(
            timeframe_in,
            "testhost22",
            now_mocked,
            datetime.time(10, 0),
            datetime.time(10, 30)
        )
        self.assertEqual(1, len(tfs))

    def test_filter_timeframes_from_is_smaller_to_is_bigger(self):
        # Tests the filtering of timeframes when
        # the `from` attribute of a timeframe is smaller as
        # the current point in time (now) and the `to` attribute
        # of a timeframe is bigger as the current point in time.

        # from < now < to

        # from:   1000
        # to:     1030
        # now:    1015

        now = datetime.datetime.now()
        now_mocked = cmn.local_datetime(
            datetime.datetime(now.year, now.month, now.day, 10, 15)
        )

        (tfs, _, _) = self.timeframe_mocked(
            timeframe_in,
            "testhost22",
            now_mocked,
            datetime.time(10, 0),
            datetime.time(10, 30)
        )
        self.assertEqual(1, len(tfs))

    def test_filter_timeframes_all_the_same(self):
        # Tests the filtering of timeframes when
        # the `from` attribute of a timeframe, the `to` attribute
        # and the current point in time (now) all have the same
        # value.

        # from = now = to

        # from:   1000
        # to:     1000
        # now:    1000

        now = datetime.datetime.now()
        now_mocked = cmn.local_datetime(
            datetime.datetime(now.year, now.month, now.day, 10, 0)
        )

        (tfs, _, _) = self.timeframe_mocked(
            timeframe_in,
            "testhost22",
            now_mocked,
            datetime.time(10, 0),
            datetime.time(10, 0)
        )
        self.assertEqual(1, len(tfs))

    def test_filter_timeframes_outside(self):
        # Tests the filtering of timeframes when
        # the `from` attribute of a timeframe is bigger than,
        # and the `to` attribute is smaller than the
        # current point in time (now).

        # from > now < to

        # from:   1000
        # to:     0930
        # now:    0945

        now = datetime.datetime.now()
        now_mocked = cmn.local_datetime(
            datetime.datetime(now.year, now.month, now.day, 9, 45)
        )

        (tfs, _, _) = self.timeframe_mocked(
            timeframe_in,
            "testhost22",
            now_mocked,
            datetime.time(10, 0),
            datetime.time(9, 30)
        )
        self.assertEqual(0, len(tfs))

    def trigger_no_sla_query(self, notification, hostname, now=cmn.now()):
        """Runs a notification concerning a host
        who is not within an active SLA.
        """
        handler = nl.NagiosLog()
        log = mod.Nagioslog(
            **notification
        )
        log.hostname = hostname
        self.save_log(log)

        @gen.coroutine
        def coro():
            yield handler.handle(log, no_init=False, no_exec=True)
            raise cmn.Return(handler.logdata)

        with mock.patch(
            "symonitoring_alert.common.now"
        ) as now_mocked:
            now_mocked.return_value = now
            return cmn.run_sync(coro)

    @cmn.run_on_executor
    def save_log(self, log):
        log.save()

    @cmn.run_on_executor
    def delete_log_by_id(self, log_id):
        log_entries = mod.Nagioslog.objects.select_related().filter(
            id=log_id
        )
        for log_entry in log_entries:
            try:
                log_entry.delete()
            except Exception:  # pylint: disable=broad-except
                pass

    @cmn.run_on_executor
    def get_alert_holds_ex(self):
        return list(mod.Hold.objects.all().select_related())

    @gen.coroutine
    def get_alert_holds(self):
        alert_holds = yield self.get_alert_holds_ex()
        raise cmn.Return(alert_holds)

    def trigger_query_now_mocked(self, now, logdata):
        """Runs a query for handling all log entries
        which are currently not within a SLA with a
        mocked current point in time (`now`).
        Returns the given logdata (if present) after
        handling all before mentioned log entries.
        """
        handler = nl.NagiosLog()

        @gen.coroutine
        def coro():
            yield handler.handle(logdata.log, no_init=False, no_exec=True)

        with mock.patch(
            "symonitoring_alert.common.now"
        ) as op:
            op.return_value = now
            with cmn._allow_block_for_test_only():
                cmn.run_sync(coro)
                logdata_changed = mod.Nagioslog.objects.get(
                    id=logdata.log.id
                )
                return (logdata_changed, op)

    def test_no_sla_alerts(self):
        # Tests the behaviour of symonitoring_alert when a log
        # entry is first not within a valid SLA (or particularly
        # not within a timeframe for the given host) but
        # then slips inside a valid SLA (timeframe) as the
        # time proceeds.

        now = datetime.datetime.now()
        now_mocked_out_of_sla = cmn.local_datetime(
            datetime.datetime(now.year, now.month, now.day, 6, 0)
        )
        now_mocked_in_sla = cmn.local_datetime(
            datetime.datetime(now.year, now.month, now.day, 23, 15)
        )

        logdata = self.trigger_no_sla_query(
            notification_host_critical_out_of_sla,
            "testhost24",
            now_mocked_out_of_sla
        )

        self.assertEqual(
            logdata.log.processed,
            const.NagiosLog.STATE_NO_SLA
        )

        (logdata_changed, op) = self.trigger_query_now_mocked(
            now_mocked_in_sla,
            logdata
        )
        self.assertTrue(op.called)
        self.assertEqual(
            logdata_changed.processed,
            str(const.NagiosLog.STATE_ALERTED)
        )

    def trigger_no_sla_query_to_db(
            self,
            notification,
            hostname,
            now=cmn.now()
    ):
        """Runs a notification concerning a host
        who is not within an active SLA.
        """
        handler = nl.NagiosLog()
        log = mod.Nagioslog(
            **notification
        )
        log.hostname = hostname
        self.save_log(log)

        @gen.coroutine
        def coro():
            yield handler.handle(log, no_init=False, no_exec=True)

        with mock.patch(
            "symonitoring_alert.common.now"
        ) as now_mocked, mock.patch(
            "symonitoring_alert.logdata.LogData.save"
        ):
            now_mocked.return_value = now
            cmn.run_sync(coro)

        return handler.logdata

    def trigger_handle_no_sla(self, logdata, now=cmn.now()):
        log_handler = nl.NagiosLog()

        with mock.patch(
            'symonitoring_alert.handle.alert.AlertExec.mktimeout'
        ) as mktimeout, mock.patch(
            'symonitoring_alert.common.now'
        ) as now_mocked, mock.patch(
            'symonitoring_alert.common.add_callback'
        ):

            now_mocked.return_value = now

            @gen.coroutine
            def coro():
                yield log_handler.handle(logdata.log)
            cmn.run_sync(coro)
            return mktimeout

    def test_following_up_no_sla_alerts(self):
        # Tests the behaviour of symonitoring_alert when a log
        # entry is not within a valid SLA (or particularly
        # not within a timeframe for the given host) with
        # critical and this log entry gets followed by a similar
        # log entry but with state ok.

        # ID 123
        notification    = notification_service_critical_out_of_sla
        # ID 124
        ok_notification = notification_service_ok_out_of_sla

        now = datetime.datetime.now()
        now_out_of_sla = cmn.local_datetime(
            datetime.datetime(now.year, now.month, now.day, 3, 0)
        )

        try:
            # Handle log data to be out of SLA
            logdata_crit = self.trigger_no_sla_query_to_db(
                notification,
                "testhost24",
                now=now_out_of_sla
            )
            self.assertEqual(
                logdata_crit.log.processed,
                const.NagiosLog.STATE_NO_SLA
            )

            # Handle following up log data for same event
            # but with status OK
            logdata_ok = self.trigger_no_sla_query_to_db(
                ok_notification,
                "testhost24"
            )
            self.assertEqual(
                logdata_ok.log.processed,
                const.NagiosLog.STATE_NO_PREVIOUS_DOWN
            )

            # Re-handle out of SLA log data again
            # which already received an OK message
            # so no alerting must happen
            mktimeout_handler = self.trigger_handle_no_sla(
                logdata_crit,
                now=now_out_of_sla
            )
            self.assertFalse(mktimeout_handler.called)

        finally:
            self.delete_log_by_id(notification['id'])
            self.delete_log_by_id(ok_notification['id'])
