# -*- coding: UTF-8 -*-


# pylint: disable=missing-docstring,too-many-lines,no-self-use
# pylint: disable=protected-access,line-too-long
# pylint: disable=trailing-whitespace

"""
Testing timing of alert scheduling


                                       .---------------------------------------.
    symonitoring_alert Timing Tests    |                Syntax                 |
    --------------------               |---------------------------------------|
                                       | -> User / Rule / Seconds              |
 .-----------------------------------. |                                       |
 |              Method               | | Where -> Represents now() and         |
 |-----------------------------------| | Seconds the timeout the next          |
 | * Mock add_timeout() and now()    | | handler is scheduled.                 |
 | * Transpart mock mktimeout        | |                                       |
 | * Call handler (->) inspect calls | | If we call (->) the handler to early  |
 '-----------------------------------' |   we get a timeout > 0                |
                                       | If we call (->) the handler to late   |
                                       |   we get a timeout == 0               |
      min       rule                   '---------------------------------------'
           ____
       0  |      <- handler at min 0          --.                    
          |         is alway called             | 5 min restart delay
          |___                                  |                    
       1  |      0                            --'                    
          |    
          |___ 
       2  |      1
          |    
          |___     --.                   
       3  |      2   | 10 min cycle delay
          |          | between users     
          |____    --'                   
      13

"""  # noqa

# System imports
import datetime
import tornado.gen                  as gen
import mock

# Project imports
import symonitoring_alert.common               as cmn
import symonitoring_alert.handle.nagioslog     as nl
import symonitoring_alert.handle.alert         as al
import symonitoring_model.models               as mod
import symonitoring_alert.constants            as const
import symonitoring_alert.query                as query

notification_host_critical = {
    'address': u'192.24.12.3',
    'attempt': 1,
    'checkcommand': u'check-host-alive',
    'datetime': cmn.local_datetime(datetime.datetime(
        2014, 5, 6, 15, 19, 7
    )),
    'duration': u'0d 0h 0m 0s',
    'hostname': u'localhost',
    'id': 123,
    'lastproblemid': 47,
    'name': u'localhost',
    'objecttype': u'HOST',
    'output': u'PING CRITICAL - Packet loss = 100%',
    'perfdata': u'rta=5000.000000ms;3000.000000;5000.000000;0.000000 pl=100%;'
                '80;100;0',
    'problemid': 48,
    'processed': u'0',
    'state': u'DOWN',
    'statetype': u'HARD',
    'timestamp': cmn.local_datetime(datetime.datetime(
        2014, 5, 6, 15, 19, 7
    ))
}


class TimingTest(cmn.TestCase):
    def __init__(self, *args, **kwargs):
        super(TimingTest, self).__init__(*args, **kwargs)
        self._called_rule = None
        self._called_user = None

    def prepare_alert(self, hostname, now):
        """Creates alert and handles first callback"""
        handler = nl.NagiosLog()
        log = mod.Nagioslog(
            **notification_host_critical
        )
        log.hostname = hostname
        self.save_log(log)

        @gen.coroutine
        def coro():
            yield handler.handle(log, no_exec=True)
            alert = al.AlertExec()
            alert.logdata = handler.logdata
            yield alert.handle()
            raise cmn.Return(alert)

        with mock.patch("symonitoring_alert.common.add_timeout") as add, \
                mock.patch("symonitoring_alert.common.now") as n:
            n.return_value = now
            res = cmn.run_sync(coro)
            self.assertTrue(n.called)
            self.assertTrue(add.called)
            self.assertEqual(len(add.call_args_list), 1)
            self.assertEqual(
                cmn.round_to_two_seconds(add.call_args[0][0]),
                datetime.timedelta(seconds=60)
            )
            return (res, add.call_args[0][1])

    def handle_at(self, alert, handler, now):
        orig = al.AlertExec.mktimeout

        @gen.coroutine
        def coro():
            yield handler()

        @gen.coroutine
        def coro_true():
            raise cmn.Return(True)

        def pass_through(*args, **kwargs):
            self._called_rule = args[1]
            self._called_user = args[3]
            return orig(alert, *args, **kwargs)

        with mock.patch("symonitoring_alert.common.add_timeout") as add, \
                mock.patch("symonitoring_alert.common.now") as n, \
                mock.patch(
                    "symonitoring_alert.communication.UserCommunicationBase.send_xmpp"  # noqa
                ) as xmpp, \
                mock.patch(
                    "symonitoring_alert.handle.alert.AlertExec.mktimeout",
                ) as alexec:  # noqa
            alexec.side_effect = pass_through
            # pylint: disable=assignment-from-no-return
            xmpp.return_value = coro_true()
            n.return_value = now
            cmn.run_sync(coro)
            return (add, add.call_args[0][1])

    def run_iteration(
            self,
            alert,
            handler,
            now,
            delay_seconds,
            expect_seconds,
            user,
            actionid
    ):
        # pylint: disable=too-many-arguments
        later = now + datetime.timedelta(seconds=delay_seconds)
        (add, handler) = self.handle_at(alert, handler, later)
        self.assertTrue(add.called)
        self.assertEqual(len(add.call_args_list), 1)
        self.assertEqual(
            cmn.round_to_two_seconds(add.call_args[0][0]),
            datetime.timedelta(seconds=expect_seconds)
        )
        self.assertEqual(
            self._called_rule.actionid_id,
            actionid
        )
        self.assertEqual(
            self._called_user.name,
            user,
        )
        return handler

#  Timeout Test one cycle I
#  ------------------------
#
#   ____
#  |    
#  |     <- 0/1/90 
#  |___     0/2/150
#  |        1/0/810
#  |    
#  |___ 
#  |    
#  |    
#  |___ 
#  |    
#  |    
#  |____
#
    def test_timing_1cycle1(self):
        now = cmn.now()
        (alert, handler) = self.prepare_alert("timingHost00", now)
        handler = self.run_iteration(
            alert, handler, now,
            30,
            90,
            "TimingUser00",
            const.ActionType.XMPP
        )
        handler = self.run_iteration(
            alert, handler, now,
            30,
            150,
            "TimingUser00",
            const.ActionType.PHONE
        )
        handler = self.run_iteration(
            alert, handler, now,
            30,
            810,
            "TimingUser01",
            const.ActionType.EMAIL
        )

#
#  Timeout Test one cycle II
#  -------------------------
#
#   ____
#  |    
#  |    
#  |___  _> 0/1/60 
#  |        0/2/120
#  |        1/0/780
#  |___ 
#  |    
#  |    
#  |___ 
#  |    
#  |    
#  |____
#
    def test_timing_1cycle2(self):
        now = cmn.now()
        (alert, handler) = self.prepare_alert("timingHost00", now)
        handler = self.run_iteration(
            alert, handler, now,
            60,
            60,
            "TimingUser00",
            const.ActionType.XMPP
        )
        handler = self.run_iteration(
            alert, handler, now,
            60,
            120,
            "TimingUser00",
            const.ActionType.PHONE
        )
        handler = self.run_iteration(
            alert, handler, now,
            60,
            780,
            "TimingUser01",
            const.ActionType.EMAIL
        )

#
#  Timeout Test one cycle III
#  --------------------------
#
#   ____
#  |    
#  |    
#  |___ 
#  |    
#  |     -> 0/1/30 
#  |___     0/2/90 
#  |        1/0/720
#  |    
#  |___ 
#  |    
#  |    
#  |____
#
    def test_timing_1cycle3(self):
        now = cmn.now()
        (alert, handler) = self.prepare_alert("timingHost00", now)
        handler = self.run_iteration(
            alert, handler, now,
            90,
            30,
            "TimingUser00",
            const.ActionType.XMPP
        )
        handler = self.run_iteration(
            alert, handler, now,
            90,
            90,
            "TimingUser00",
            const.ActionType.PHONE
        )
        handler = self.run_iteration(
            alert, handler, now,
            90,
            750,
            "TimingUser01",
            const.ActionType.EMAIL
        )

    def test_timing_1cycle3_alert_timeout(self):
        now = cmn.now()
        (alert, handler) = self.prepare_alert("timingHost00", now)
        alert.logdata.sla.timeout = 5
        handler = self.run_iteration(
            alert, handler, now,
            90,
            30,
            "TimingUser00",
            const.ActionType.XMPP
        )
        handler = self.run_iteration(
            alert, handler, now,
            90,
            90,
            "TimingUser00",
            const.ActionType.PHONE
        )
        with self.assertRaises(TypeError):
            handler = self.run_iteration(
                alert, handler, now,
                90,
                750,
                "TimingUser01",
                const.ActionType.EMAIL
            )

#
#  Timeout Test one cycle II
#  -------------------------
#
#   ____
#  |    
#  |    
#  |___ 
#  |
#  |    -> 0/1/30
#  |___ 
#  |    
#  |    
#  |___ 
#  |    
#  |    -> 0/2/0
#  |____   0/2/630
#
    def test_timing_1cycle3_2_times(self):
        now = cmn.now()
        (alert, handler) = self.prepare_alert("timingHost00", now)
        handler = self.run_iteration(
            alert, handler, now,
            90,
            30,
            "TimingUser00",
            const.ActionType.XMPP
        )
        handler = self.run_iteration(
            alert, handler, now,
            210,
            0,
            "TimingUser00",
            const.ActionType.PHONE
        )
        handler = self.run_iteration(
            alert, handler, now,
            210,
            630,
            "TimingUser01",
            const.ActionType.EMAIL
        )

#
#  Timeout Test two cycle I  
#  --------------------------
#   ____
#  |    
#  | ...
#  |____
#   ____
#  |    
#  |     -> 0/1/0  
#  |___     0/2/0  
#  |        1/0/30 
#  |        1/1/90 
#  |___     1/2/150
#  |        2/0/810
#  |               
#  |___ 
#  |    
#  |    
#  |____
#
    def test_timing_2cycle1(self):
        now = cmn.now()
        (alert, handler) = self.prepare_alert("timingHost00", now)
        handler = self.run_iteration(
            alert, handler, now,
            810,
            0,
            "TimingUser00",
            const.ActionType.XMPP
        )
        handler = self.run_iteration(
            alert, handler, now,
            810,
            0,
            "TimingUser00",
            const.ActionType.PHONE
        )
        handler = self.run_iteration(
            alert, handler, now,
            810,
            30,
            "TimingUser01",
            const.ActionType.EMAIL
        )
        handler = self.run_iteration(
            alert, handler, now,
            810,
            90,
            "TimingUser01",
            const.ActionType.XMPP
        )
        handler = self.run_iteration(
            alert, handler, now,
            810,
            150,
            "TimingUser01",
            const.ActionType.PHONE
        )
        handler = self.run_iteration(
            alert, handler, now,
            810,
            810,
            "TimingUser02",
            const.ActionType.EMAIL
        )

#
#  Timeout Test two cycle I  
#  --------------------------
#   ____
#  |    
#  | ...
#  |____
#   ____
#  |    
#  |     -> 0/1/0  
#  |___     0/2/0  
#  |        1/0/30 
#  |        1/1/90 
#  |___     1/2/150
#  |        2/0/810
#  |        2/1/870
#  |___     2/1/930
#  |        0/0/1890
#  |    
#  |____
#
    def test_timing_restart(self):
        now = cmn.now()
        (alert, handler) = self.prepare_alert("timingHost00", now)
        handler = self.run_iteration(
            alert, handler, now,
            810,
            0,
            "TimingUser00",
            const.ActionType.XMPP
        )
        handler = self.run_iteration(
            alert, handler, now,
            810,
            0,
            "TimingUser00",
            const.ActionType.PHONE
        )
        handler = self.run_iteration(
            alert, handler, now,
            810,
            30,
            "TimingUser01",
            const.ActionType.EMAIL
        )
        handler = self.run_iteration(
            alert, handler, now,
            810,
            90,
            "TimingUser01",
            const.ActionType.XMPP
        )
        handler = self.run_iteration(
            alert, handler, now,
            810,
            150,
            "TimingUser01",
            const.ActionType.PHONE
        )
        handler = self.run_iteration(
            alert, handler, now,
            810,
            810,
            "TimingUser02",
            const.ActionType.EMAIL
        )
        handler = self.run_iteration(
            alert, handler, now,
            810,
            870,
            "TimingUser02",
            const.ActionType.XMPP
        )
        handler = self.run_iteration(
            alert, handler, now,
            810,
            930,
            "TimingUser02",
            const.ActionType.PHONE
        )
        handler = self.run_iteration(
            alert, handler, now,
            810,
            1890,
            "TimingUser00",
            const.ActionType.EMAIL
        )

    def test_timing_1cycle3_with_reload(self):
        now = cmn.now()
        (alert, handler) = self.prepare_alert("timingHost00", now)
        handler = self.run_iteration(
            alert, handler, now,
            90,
            30,
            "TimingUser00",
            const.ActionType.XMPP
        )
        handler = self.run_iteration(
            alert, handler, now,
            90,
            90,
            "TimingUser00",
            const.ActionType.PHONE
        )
        log = mod.Nagioslog(
            **notification_host_critical
        )
        log.hostname = "timingHost00"
        log.processed = const.NagiosLog.STATE_ALERTED
        log.alert = alert.logdata.alert.model
        self.save_log(log)
        log_init = query.NagiosLog()

        @cmn.coroutine
        def coro():
            yield log_init.initialize_alerts()

        with mock.patch("symonitoring_alert.common.now") as n, \
                mock.patch("symonitoring_alert.common.add_timeout") as add:
            n.return_value = now + datetime.timedelta(seconds=120)
            cmn.run_sync(coro)
            self.assertTrue(add.called)
            self.assertEqual(len(add.call_args_list), 1)
            self.assertEqual(
                cmn.round_to_two_seconds(add.call_args[0][0]),
                datetime.timedelta(seconds=720)
            )
