# -*- coding: UTF-8 -*-

"""
Handling of hold events
"""

# System imports
import six
import weakref
from abc import ABCMeta

# Project imports
import symonitoring_alert.common           as cmn


@six.add_metaclass(ABCMeta)
class HoldBase(cmn.ExecutorMixin, cmn.LogDataLockMixin):
    """ Base class for handling hold events.
    """

    def __init__(self, logdata):
        """Constructor."""

        super(HoldBase, self).__init__()
        self._logdata = weakref.ref(logdata)
        self._hold    = None

    @property
    def logdata(self):
        """:rtype: symonitoring_alert.logdata.LogData"""
        return self._logdata()

    @property
    def hold(self):
        """:rtype: symonitoring_alert.mod.Hold"""
        return self._hold

    @cmn.run_on_executor
    def get_or_create_ex(self):
        """ Run get_or_create() on executor """
        return self.get_or_create()

    def get_or_create(self):
        """Puts a NagiosLog entry on hold from being processed."""
        raise NotImplementedError

    @cmn.run_on_executor
    def delete_ex(self):  # pragma: no cover
        """ Run delete() on executor """
        return self.delete()

    def delete(self):
        """Removes all NagiosLog entries matching given entry from being on
        hold."""
        raise NotImplementedError

    @cmn.run_on_executor
    def get_ex(self):
        """ Run get() on executor """
        return self.get()

    def get(self):
        """Gets all entries being on hold matching given NagiosLog entry."""
        raise NotImplementedError

    @classmethod
    @cmn.run_on_executor
    def set_ex(cls, log_id):
        """ Run set() on executor """
        return cls.set(log_id)

    @classmethod
    def set(cls, log_id):
        """Puts the given NagiosLog entry on hold from being processed."""
        raise NotImplementedError

    def related_active_entries(self):
        """ Get all related nagios log entries """
        raise NotImplementedError

    def previous_related_active_entries(self):
        """ Get previous related nagios log entries """
        log = self.logdata.log
        if log.id:
            return self.related_active_entries().filter(
                nagioslogid_id__lt = log.id
            )
        else:
            return self.related_active_entries()
