# -*- coding: UTF-8 -*-

"""
Testing of hold informative class
"""

# pylint: disable=missing-docstring,too-many-lines,no-self-use

# System imports
import datetime
import tornado.gen                         as gen

# Project imports
import symonitoring_alert.common           as cmn
import symonitoring_model.models           as mod
import symonitoring_alert.logdata          as ld
import symonitoring_alert.handle.nagioslog as nl
import symonitoring_alert.hold.inform      as hi

NOTIFICATION_HOST_CRITICAL = {
    'address': u'192.24.12.3',
    'attempt': 1,
    'checkcommand': u'check-host-alive',
    'datetime': cmn.local_datetime(datetime.datetime(
        2014, 5, 6, 15, 19, 7
    )),
    'duration': u'0d 0h 0m 0s',
    'hostname': u'localhost',
    'lastproblemid': 47,
    'name': u'localhost',
    'objecttype': u'HOST',
    'output': u'PING CRITICAL - Packet loss = 100%',
    'perfdata': u'rta=5000.000000ms;3000.000000;5000.000000;0.000000 pl=100%;'
                '80;100;0',
    'problemid': 48,
    'processed': u'0',
    'state': u'DOWN',
    'statetype': u'HARD',
    'timestamp': cmn.local_datetime(datetime.datetime(
        2014, 5, 6, 15, 19, 7
    ))
}


class HoldInittTest(cmn.TestCase):
    def run_init(self, notification, hostname=None):
        handler = nl.NagiosLog()
        log = mod.Nagioslog(
            **notification
        )
        if hostname is not None:
            log.hostname = hostname
        self.save_log(log)

        @gen.coroutine
        def coro():
            yield handler.handle(log, no_exec=True)

        cmn.run_sync(coro)
        return handler.logdata

    @cmn.run_on_executor
    def get_inform_holds(self):
        return list(mod.HoldInform.objects.all().select_related())

    @gen.coroutine
    def coro_get(self):
        inform_holds = yield self.get_inform_holds()
        raise cmn.Return(inform_holds)

    def test_sync(self):
        res = self.run_init(
            NOTIFICATION_HOST_CRITICAL,
            "localhost"
        )
        self.assertIs(res.__class__, ld.LogData)

        # Save/Update logdata in hold inform
        @gen.coroutine
        def coro_sync():
            hold_inform = hi.HoldInform(res)
            yield hold_inform.get_or_create_ex()
        cmn.run_sync(coro_sync)

        inform_holds = cmn.run_sync(self.coro_get)
        self.assertEqual(1, len(inform_holds))
        self.assertEqual(
            res.log.id,
            inform_holds[0].nagioslogid_id
        )

    def test_delete(self):
        res = self.run_init(
            NOTIFICATION_HOST_CRITICAL,
            "localhost"
        )
        self.assertIs(res.__class__, ld.LogData)

        @gen.coroutine
        def coro_sync():
            hold_inform = hi.HoldInform(res)
            yield hold_inform.get_or_create_ex()
        cmn.run_sync(coro_sync)

        @gen.coroutine
        def coro_del():
            hold_inform = hi.HoldInform(res)
            yield hold_inform.delete_ex()
        cmn.run_sync(coro_del)

        inform_holds = cmn.run_sync(self.coro_get)
        self.assertLessEqual(0, len(inform_holds))

    def test_related_active_entries(self):
        logdata_active = self.run_init(
            NOTIFICATION_HOST_CRITICAL
        )
        log_active = logdata_active.log
        log_new = mod.Nagioslog(
            **NOTIFICATION_HOST_CRITICAL
        )
        self.save_log(log_new)
        logdata_new = ld.LogData(log_new)

        # Save/Update logdata in hold
        @gen.coroutine
        def coro_sync():
            yield logdata_active.hold_inform.get_or_create_ex()
            yield logdata_new.hold_inform.get_or_create_ex()
        cmn.run_sync(coro_sync)

        @gen.coroutine
        def coro():
            entries = yield logdata_new.hold_inform.related_active_entries_ex()
            raise cmn.Return(entries)
        related_entries = cmn.run_sync(coro)
        with cmn._allow_block_for_test_only():
            related_entries = [entry.nagioslogid for entry in related_entries]
            self.assertEqual(len(related_entries), 2)
            self.assertEqual(
                set(related_entries),
                set([log_active, log_new])
            )
