# -*- coding: UTF-8 -*-

# pylint: disable=too-many-arguments,too-many-return-statements

"""Struct to store collected logging data"""

# System imports
import dateutil.rrule
import datetime
import toro
import django.db.transaction          as transaction
import django.db.models               as dmod

# Project imports
import symonitoring_alert.alert.state as state
import symonitoring_alert.common      as cmn
import symonitoring_alert.constants   as const
import symonitoring_model.models      as mod
import symonitoring_alert.dummy       as dummy
import symonitoring_alert.alert       as alogic
import symonitoring_alert.hold.alert  as hal
import symonitoring_alert.hold.inform as hin


class MalformedRow(Exception):
    """ Exception raised when logdata cannot be loaded """
    pass


class LogData(cmn.ExecutorMixin):
    """Struct to store collected logging data"""

    _ping_command = None

    __slots__ = (
        '_log',
        '_host',
        '_service',
        '_sla',
        '_timeframes',
        '_open_hostalerts',
        '_users',
        '_command',
        '_service_instance',
        '_alert',
        '_duplicated_alerts',
        '_host_service_alerts',
        '_downtimes',
        '_lock',
        '_infofeeds',
        '_is_from_acknowledge',
        '_alert_state',
        '_hold',
        '_hold_inform',
    )

    def __init__(
            self,
            log,
    ):
        super(LogData, self).__init__()
        self._log                 = log
        self._host                = None
        self._sla                 = None
        self._service             = None
        self._timeframes          = None
        self._open_hostalerts     = None
        self._users               = None
        self._command             = None
        self._service_instance    = None
        self._downtimes           = None
        self._duplicated_alerts   = None
        self._host_service_alerts = None
        self._lock                = toro.Lock()
        self._infofeeds           = []
        self._is_from_acknowledge = False
        self._alert_state         = state.AlertState(self)
        self._alert               = alogic.Alert(self)
        self._hold                = hal.HoldAlert(self)
        self._hold_inform         = hin.HoldInform(self)

    @property
    def log(self):
        """:rtype: symonitoring_model.models.Nagioslog"""
        return self._log

    @property
    def host(self):
        """:rtype: symonitoring_model.models.Host"""
        return self._host

    @host.setter
    def host(self, x):
        """:type x: symonitoring_model.models.Host"""
        self._host = x

    @property
    def service(self):
        """:rtype: symonitoring_model.models.Service"""
        return self._service

    @property
    def sla(self):
        """:rtype: symonitoring_model.models.Sla"""
        return self._sla

    @property
    def timeframes(self):
        """:rtype: list(symonitoring_model.models.Timeframe)"""
        return self._timeframes

    @property
    def open_hostalerts(self):
        """:rtype: list(symonitoring_model.models.Alert)"""
        return self._open_hostalerts

    @property
    def users(self):
        """:rtype: list(symonitoring_model.models.User)"""
        return self._users

    @property
    def command(self):
        """:rtype: symonitoring_model.models.Command"""
        return self._command

    @property
    def service_instance(self):
        """:rtype: symonitoring_model.models.Hostserviceinstance"""
        return self._service_instance

    @service_instance.setter
    def service_instance(self, x):
        """:type x: symonitoring_model.models.Hostserviceinstance"""
        self._service_instance = x

    @property
    def alert(self):
        """:rtype: symonitoring_alert.alert.Alert"""
        return self._alert

    @property
    def infofeeds(self):
        """:rtype: list(symonitoring_model.models.Infofeed)"""
        return self._infofeeds

    @infofeeds.setter
    def infofeeds(self, x):
        """:type x: list(symonitoring_model.models.Infofeed)"""
        self._infofeeds = x

    @property
    def is_from_acknowledge(self):
        """:rtype: bool"""
        return self._is_from_acknowledge

    @is_from_acknowledge.setter
    def is_from_acknowledge(self, x):
        """:type x: bool"""
        self._is_from_acknowledge = x

    @property
    def duplicated_alerts(self):
        """:rtype: list(symonitoring_model.models.Alert)"""
        return self._duplicated_alerts

    @property
    def host_service_alerts(self):
        """:rtype: list(symonitoring_model.models.Alert)"""
        return self._host_service_alerts

    @property
    def downtimes(self):
        """:rtype: list(symonitoring_model.models.Downtime)"""
        return self._downtimes

    @property
    def lock(self):
        """:rtype: toro.Lock"""
        return self._lock

    @property
    def is_service(self):
        """:rtype: boolean"""
        return self.log.objecttype.lower() == "service"

    @property
    def is_host(self):
        """:rtype: boolean"""
        return self.log.objecttype.lower() == "host"

    @property
    def is_soft(self):
        """:rtype: boolean"""
        return self.log.statetype.lower() == "soft"

    @property
    def is_warning(self):
        """:rtype: boolean"""
        return self.log.state.lower() == "warning"

    @property
    def is_ok(self):
        """:rtype: boolean"""
        return (
            self.log.state.lower() == "up" or
            self.log.state.lower() == "ok"
        )

    @property
    def has_recovered(self):
        """:rtype: boolean"""
        return (
            self.is_ok or
            self.is_warning
        ) and not self.is_soft

    @property
    def alert_state(self):
        """:rtype: symonitoring_alert.alert.state.AlertState"""
        return self._alert_state

    @property
    def acknowledged_processed(self):
        """:rtype: boolean"""
        return (
            self._alert.model and (
                self._alert.model.acknowledgedautomatically or
                self._alert.model.acknowledgedprocessed
            )
        )

    def previous_related_active_entries(self):
        """Get all related nagios log entries"""
        log = self.log
        if log.id:
            return self.related_active_entries().filter(
                id__lt = log.id
            )
        else:
            return self.related_active_entries()

    @cmn.run_on_executor
    def related_active_entries_ex(self):
        """ Run related_active_entries() on executor """
        return self.related_active_entries()

    def related_active_entries(self):
        """Get all related nagios log entries"""
        log = self.log
        return mod.Nagioslog.objects.select_related().filter(
            ~dmod.Q(processed = const.NagiosLog.STATE_ALERT_CLOSED),
            alert__datetimerecovery__isnull=True,
            hostname=log.hostname,
            name=log.name
        ).order_by('-id')

    @property
    def hold(self):
        """:rtype: symonitoring_alert.hold.HoldAlert"""
        return self._hold

    @property
    def hold_inform(self):
        """:rtype: symonitoring_alert.hold.HoldInform"""
        return self._hold_inform

    @cmn.run_on_executor
    def load_ex(self):
        """ Run load() on executor """
        return self.load()

    def load(self):
        """Collect all data related to log entry"""
        log = self.log
        assert isinstance(log, mod.Nagioslog)

        with transaction.atomic():
            try:
                host = mod.Host.objects.filter(
                    hostname=log.hostname
                ).select_related()[0]
            except IndexError:
                log.processed = const.NagiosLog.STATE_MALFORMED_ROW
                msg = "Log for un-managed host: %s" % log.hostname
                cmn.lg.warning(msg)
                raise MalformedRow(msg)

            assert isinstance(host, mod.Host)

            # sla priority: service sla, host sla, customer sla
            sla                 = None
            open_hostalerts     = None
            command             = None
            service_instance    = None
            service             = None
            alert               = None
            downtimes           = None
            duplicated_alerts   = None
            host_service_alerts = None

            downtimes = list(host.downtime_set.filter(
                dmod.Q(datetimeto__gt=cmn.now()) |
                dmod.Q(datetimeto__isnull=True)
            ).select_related())

            entries = self.related_active_entries().filter(
                alert__isnull = False
            )
            entries_len = len(entries)
            # Sanity check: if our bookkeeping is bad this will trigger
            if entries_len > 5:  # pragma: no cover
                cmn.lg.critical(
                    "Found %d related alerts, this should not happen.",
                    entries_len
                )
            if entries:
                alert = entries[0].alert

            if host.customerid:
                if host.customerid.slaid:
                    sla = host.customerid.slaid

            if host.slaid:
                sla = host.slaid

            if log.objecttype.lower() == "host":
                if not LogData._ping_command:
                    try:
                        self._ping_command = mod.Command.objects.filter(
                            name="pingHost"
                        ).select_related()[0]
                    except IndexError:  # pragma: no cover
                        log.processed = const.NagiosLog.STATE_MALFORMED_ROW
                        msg = (
                            "No command named 'pingHost' (this "
                            "means that no alerts for hosts can "
                            "be created!)"
                        )
                        cmn.lg.critical(msg)
                        raise MalformedRow(msg)
                command = self._ping_command
                host_service_alerts = list(host.alert_set.filter(
                    acknowledged=False,
                    serviceinstanceid__isnull=False,
                ).select_related())

            duplicated_alerts = list(host.alert_set.filter(
                datetimerecovery__isnull=True,
                serviceinstanceid__isnull=True,
            ).select_related())

            if log.objecttype.lower() == "service":
                # The service as to be resolved via a name
                service_instance_name = None
                service_name = None
                tmp_name = log.name
                if "__" in tmp_name:
                    split_name            = tmp_name.split("__")
                    service_name          = split_name[0]
                    service_instance_name = split_name[1]
                else:
                    service_name          = tmp_name

                try:
                    service = mod.Service.objects.filter(
                        name=service_name
                    ).select_related()[0]
                except IndexError:  # pragma: no cover
                    log.processed = const.NagiosLog.STATE_MALFORMED_ROW
                    msg = "No service named '%s'" % service_name
                    cmn.lg.warning(msg)
                    raise MalformedRow(msg)

                assert isinstance(service, mod.Service)

                open_hostalerts = duplicated_alerts

                try:
                    command_name = log.checkcommand.split("!")[0]
                    command = mod.Command.objects.filter(
                        name=command_name
                    ).select_related()[0]
                except IndexError:  # pragma: no cover
                    log.processed = const.NagiosLog.STATE_MALFORMED_ROW
                    msg = "No command named '%s'" % command_name
                    cmn.lg.warning(msg)
                    raise MalformedRow(msg)

                try:
                    service_instance = mod.Hostserviceinstance.objects.filter(
                        hostserviceid__hostid = host,
                        hostserviceid__serviceid = service,
                        name = service_instance_name,
                    ).select_related()[0]
                except IndexError:
                    log.processed = const.NagiosLog.STATE_MALFORMED_ROW
                    msg = (
                        "No service instance named '%s' for host '%s' and "
                        "service '%s'" % (
                            service_instance_name,
                            host.hostname,
                            service.name
                        )
                    )
                    cmn.lg.warning(msg)
                    raise MalformedRow(msg)
                assert isinstance(service_instance, mod.Hostserviceinstance)
                if service_instance.slaid:
                    sla = service_instance.slaid
                duplicated_alerts = list(host.alert_set.filter(
                    datetimerecovery__isnull=True,
                    serviceinstanceid=service_instance
                ))

            timeframes = list(sla.timeframe_set.select_related())
            (
                filter_timeframes,
                filter_users
            ) = self.filter_timeframe_users(host, timeframes)

        self._host                = host
        self._service             = service
        self._sla                 = sla
        self._timeframes          = filter_timeframes
        self._open_hostalerts     = open_hostalerts
        self._users               = filter_users
        self._command             = command
        self._service_instance    = service_instance
        self._alert.model         = alert
        self._duplicated_alerts   = duplicated_alerts
        self._host_service_alerts = host_service_alerts
        self._downtimes           = downtimes

    @classmethod
    def get_customer_users(cls, customerid):
        """Get users of host"""
        users = list(mod.User.objects.filter(
            customerteam__date=datetime.date.today(),
            customerteam__customerid=customerid
        ).order_by('customerteam__priority').select_related())

        if not users:
            # If we didn't find a user get the default team
            users = list(mod.User.objects.filter(
                customerteam__date__isnull=True,
                customerteam__customerid=customerid
            ).order_by('customerteam__priority').select_related())
        d = dummy.Dummy()
        return d.force_team(users)

    @classmethod
    def filter_timeframe_users(cls, host, timeframes):
        """Get all matching timeframes and its users"""
        now = cmn.now()
        day = now.strftime("%A").lower()
        filter_timeframes = []
        filter_users = {}
        for timeframe in timeframes:
            assert isinstance(timeframe, mod.Timeframe)
            today = getattr(timeframe, day) == 1
            if today:
                from_ = cmn.make_datetime_from_day_time(
                    now,
                    timeframe.timefrom
                )
                to = cmn.make_datetime_from_day_time(
                    now,
                    timeframe.timeto
                )
                append = False
                if from_ > to:
                    append = (from_ >= now <= to)
                elif from_ == to:
                    append = True
                else:
                    append = (from_ <= now <= to)
                if append:
                    timeframe.ext_rules = sorted(list(
                        timeframe.timeframerule_set.select_related()
                    ), key=lambda x: x.delay)
                    filter_timeframes.append(timeframe)
                    if timeframe.customerid:
                        timeframe.ext_active_customerid = (
                            timeframe.customerid
                        )
                    else:
                        timeframe.ext_active_customerid = host.customerid
                    filter_users[timeframe] = cls.get_customer_users(
                        timeframe.ext_active_customerid,
                    )
        return (filter_timeframes, filter_users)

    def check_deactivated(self):
        """Check if host, service or customer is deactivated

        :type logdata: symonitoring_alert.logdata.LogData
        """
        service_enabled = True
        if self.service_instance:
            service_enabled = self.service_instance.isenabled
        return not (
            self.host.customerid.isenabled and
            self.host.isenabled and
            service_enabled
        )

    def check_downtime(self):
        """Check if we are in a downtime"""

        now = cmn.now()

        for downtime in self.downtimes:

            assert isinstance(downtime, mod.Downtime)

            is_active = False
            is_related = False

            # Check if downtime is active
            if downtime.rrule:
                rule = dateutil.rrule.rrulestr(
                    downtime.rrule.strip(";"),  # pylint: disable=no-member
                    dtstart=downtime.datetimefrom
                )

                from_ = rule.before(now)
                if from_:
                    to = from_ + datetime.timedelta(
                        hours=downtime.durationhours
                    )
                    if from_ <= now <= to:
                        is_active = True
            else:
                from_ = downtime.datetimefrom
                to = downtime.datetimeto
                if from_ > to:
                    found = (from_ >= now >= to)
                else:
                    found = (from_ <= now <= to)
                if found:
                    is_active = True

            is_service_downtime = downtime.serviceinstanceid is not None
            is_host_downtime = not is_service_downtime

            # Check whether downtime is scoped to a certain service or if it is
            # related to the entire host
            if is_active and is_host_downtime:
                return True
            elif is_active and is_service_downtime:
                if downtime.serviceinstanceid.id == self.service_instance.id:
                        return True

        return False

    def check_team(self):
        """Check if we have any users to inform"""
        keys = self.users.keys()
        if keys:
            for key in keys:
                timeframe_users = self.users[key]
                if timeframe_users:
                    return True
            cmn.lg.error(
                "No users found in timeframes, this is a "
                "inconsistancy."
            )
        # Shouldn't happen but is ok
        return False

    def check_hostdown(self):
        """Check if host of service is down"""
        return bool(self.open_hostalerts)

    def check_dependencydown(self):
        """Check if dependency of host is down"""
        return self._dependency_alert_exists(self.host.dependencyid)

    def _dependency_alert_exists(self, host):
        """Check if there are open alerts for host dependencies (recursive)"""
        if not host:
            return False
        assert isinstance(host, mod.Host)
        dependency_alerts = self._dependency_alert_exists(host.dependencyid)
        host_alerts = host.alert_set.filter(
            serviceinstanceid__isnull=True,
            datetimerecovery__isnull=True,
        ).count() > 0
        return host_alerts or dependency_alerts

    @cmn.run_on_executor
    def check_global_disable_ex(self):
        """Run check_global_disable() on executor"""
        return self.check_global_disable()

    def check_global_disable(self):
        """Returns if global disable has been invoked"""
        import symonitoring_alert.query as q
        return q.QueryBase().global_disable()

    def check_current_state(self, ignore_global_disable=False):
        """Returns the current state"""
        if self.is_soft:
            return const.NagiosLog.STATE_SOFT
        if not ignore_global_disable and self.check_global_disable():
            return const.NagiosLog.STATE_GLOBAL_DISABLE
        if self.is_warning:
            return const.NagiosLog.STATE_WARNING
        if self.check_deactivated():
            return const.NagiosLog.STATE_HOST_DISABLED
        if self.check_downtime():
            return const.NagiosLog.STATE_IN_MAINTENANCE
        if self.is_service:
            if self.check_hostdown():
                return const.NagiosLog.STATE_HOST_DOWN
        elif not self.is_host:  # pragma: no cover
            return const.NagiosLog.STATE_MALFORMED_ROW
        if not self.check_timeframe():
            return const.NagiosLog.STATE_NO_SLA
        if not self.check_team():
            return const.NagiosLog.STATE_NO_TEAM
        if self.check_dependencydown():
            return const.NagiosLog.STATE_DEPENDENCY_DOWN

        return None

    def check_timeframe(self):
        """Check if alert was within timeframe"""
        return bool(self.timeframes)

    @cmn.run_on_executor
    def sync_ex(self):
        """Run sync() on executor"""
        return self.sync()

    def sync(self):
        """Sync all related (live) data"""
        self.sync_log()
        self.alert_state.sync()
        self.alert.sync()

    @cmn.run_on_executor
    def sync_log_ex(self):
        """Run sync_log() on executor"""
        return self.sync_log()

    @cmn.run_on_executor
    def save_log_ex(self):
        """Saves log on executor"""
        self.log.save()

    def sync_log(self):
        """Sync NagiosLog related (live) data"""
        if self.log:
            if self.log.id:
                self._log = mod.Nagioslog.objects.get(
                    id=self.log.id
                )

    @cmn.run_on_executor
    def save_ex(self):
        """Run save() on executor"""
        return self.save()

    def save(self):
        """Save all related (live) data"""
        if self.alert:
            self.alert.save()
            if self.alert.model:
                self.log.alert = self.alert.model
        self.alert_state.save()
        self.log.save()
