# -*- coding: UTF-8 -*-

"""
Testing handling
"""

# pylint: disable=missing-docstring,too-many-lines,no-self-use
# pylint: disable=protected-access

# System imports
import datetime
import tornado.gen                         as gen
import mock

# Project imports
import symonitoring_alert.common           as cmn
import symonitoring_alert.handle.nagioslog as nl
import symonitoring_alert.logdata          as ld
import symonitoring_model.models           as mod

DATETIME_IN_DOWNTIME = datetime.datetime(2014, 5, 6, 15, 20, 7)

downtime_active = {
    'hostid': 0,
    'serviceinstanceid': None,
    'datetimefrom': cmn.local_datetime(datetime.datetime(
        2014, 5, 6, 14, 19, 7
    )),
    'datetimeto': cmn.local_datetime(datetime.datetime(
        2014, 5, 6, 16, 19, 7
    )),
    'durationhours': None,
    'rrule': None,
}

downtime_inactive = {
    'hostid': 0,
    'serviceinstanceid': None,
    'datetimefrom': cmn.local_datetime(datetime.datetime(
        2013, 5, 6, 14, 19, 7
    )),
    'datetimeto': cmn.local_datetime(datetime.datetime(
        2013, 5, 6, 16, 19, 7
    )),
    'durationhours': None,
    'rrule': None,
}

notification_host_ok = {
    'address': u'127.0.0.1',
    'attempt': 1,
    'checkcommand': u'check-host-alive',
    'datetime': cmn.local_datetime(datetime.datetime(
        2014, 5, 6, 15, 19, 7
    )),
    'duration': u'0d 0h 0m 0s',
    'hostname': u'localhost',
    'lastproblemid': 0,
    'name': u'localhost',
    'objecttype': u'HOST',
    'output': u'PING OK - Packet loss = 0%, RTA = 140.21 ms',
    'perfdata': u'rta=140.210007ms;3000.000000;5000.000000;0.000000 pl=0%;80;'
                '100;0',
    'problemid': 1,
    'processed': u'0',
    'state': u'UP',
    'statetype': u'HARD',
    'timestamp': cmn.local_datetime(datetime.datetime(
        2014, 5, 6, 15, 19, 7
    ))
}

notification_host_critical = {
    'address': u'192.24.12.3',
    'attempt': 1,
    'checkcommand': u'check-host-alive',
    'datetime': cmn.local_datetime(datetime.datetime(
        2014, 5, 6, 15, 19, 7
    )),
    'duration': u'0d 0h 0m 0s',
    'hostname': u'localhost',
    'lastproblemid': 47,
    'name': u'localhost',
    'objecttype': u'HOST',
    'output': u'PING CRITICAL - Packet loss = 100%',
    'perfdata': u'rta=5000.000000ms;3000.000000;5000.000000;0.000000 pl=100%;'
                '80;100;0',
    'problemid': 48,
    'processed': u'0',
    'state': u'DOWN',
    'statetype': u'HARD',
    'timestamp': cmn.local_datetime(datetime.datetime(
        2014, 5, 6, 15, 19, 7
    ))
}

notification_service_ok = {
    'address': '127.0.0.1',
    'attempt': 5,
    'checkcommand': 'tcp!22',
    'datetime': cmn.local_datetime(datetime.datetime(
        2014, 5, 6, 15, 19, 7
    )),
    'duration': '0d 0h 0m 20',
    'hostname': 'testhost13',
    'lastproblemid': 18,
    'name': 'ssh__default',
    'objecttype': 'SERVICE',
    'output': 'TCP OK - 0.000 second response time on port 22',
    'perfdata': 'time=0.000154s;;;0.000000;10.000000',
    'problemid': 0,
    'processed': '0',
    'state': 'OK',
    'statetype': 'HARD',
    'timestamp': cmn.local_datetime(datetime.datetime(
        2014, 5, 6, 15, 19, 7
    ))
}

notification_service_critical = {
    'address': '127.0.0.1',
    'attempt': 5,
    'checkcommand': 'tcp!221',
    'datetime': cmn.local_datetime(datetime.datetime(
        2014, 5, 6, 15, 19, 7
    )),
    'duration': '0d 0h 4m 2s',
    'hostname': 'testhost7',
    'lastproblemid': 12,
    'name': 'ssh__default',
    'objecttype': 'SERVICE',
    'output': 'Connection refused',
    'perfdata': '',
    'problemid': 46,
    'processed': '0',
    'state': 'CRITICAL',
    'statetype': 'HARD',
    'timestamp': cmn.local_datetime(datetime.datetime(
        2014, 5, 6, 15, 19, 7
    ))
}


class LogdataTest(cmn.TestCase):
    # pylint: disable=too-many-public-methods

    def handle_notification(self, notification, hostname=None):
        if hostname is not None:
            notification['hostname'] = hostname
        handler = nl.NagiosLog()
        log = mod.Nagioslog(
            **notification
        )
        self.save_log(log)

        @gen.coroutine
        def coro():
            yield handler.handle(log, no_exec=True)

        cmn.run_sync(coro)
        return log

    def test_related_active_entries(self):
        log_active = self.handle_notification(
            notification_host_critical
        )
        log_new = mod.Nagioslog(
            **notification_host_critical
        )
        self.save_log(log_new)
        logdata = ld.LogData(log_new)

        @gen.coroutine
        def coro():
            entries = yield logdata.related_active_entries_ex()
            raise cmn.Return(entries)
        related_entries = cmn.run_sync(coro)
        with cmn._allow_block_for_test_only():
            self.assertEqual(len(related_entries), 2)
            self.assertAlmostEqual(
                set(related_entries),
                set([log_active, log_new])
            )


class LogdataDowntimeTest(cmn.TestCase):

    def handle_downtime(self, now, notification, downtime, is_service_downtime=False):

        with mock.patch("symonitoring_alert.common.now") as now_mocked:
            now_mocked.return_value = now

            with cmn._allow_block_for_test_only():
                nagioslog = mod.Nagioslog(
                    **notification
                )
                nagioslog.save()

                logdata = ld.LogData(nagioslog)

                # Load related resources (service, host, ...)
                logdata.load()

                downtime['hostid'] = logdata.host
                if is_service_downtime and downtime.get('serviceinstanceid') is None:
                    downtime['serviceinstanceid'] = logdata.service_instance
                elif downtime.get('serviceinstanceid') is not None:
                    pass
                else:
                    downtime['serviceinstanceid'] = None

                downtime = mod.Downtime(**downtime)
                downtime.save()

                # Load newly created downtime
                logdata.load()

                return logdata.check_downtime()

    def test_logdata_without_downtime(self):

        with cmn._allow_block_for_test_only():
            nagioslog = mod.Nagioslog(
                **notification_host_critical
            )
            nagioslog.save()

            logdata = ld.LogData(nagioslog)
            logdata.load()

            self.assertFalse(logdata.check_downtime())


    def test_logdata_with_unrelated_service_downtime(self):

        notification = notification_service_critical
        downtime = downtime_active

        with cmn._allow_block_for_test_only():

            # Get a random host that is not the one used in the nagioslog
            host = mod.Host.objects.exclude(
                hostname=notification['hostname']
            ).select_related()[0]

            # Get a randrom service that is assigned to that host
            downtime['serviceinstanceid'] = mod.Hostserviceinstance.objects.filter(
                hostserviceid__hostid=host.id
            ).select_related()[0]

        in_downtime = self.handle_downtime(
            DATETIME_IN_DOWNTIME,
            notification,
            downtime,
            is_service_downtime=True
        )

        # Ensure that a log entry with an active service downtime, that does not
        # relate to the log entry at  all, is not in downtime.
        self.assertFalse(in_downtime)


    def test_logdata_with_active_service_downtime(self):
        in_downtime = self.handle_downtime(
            DATETIME_IN_DOWNTIME,
            notification_service_critical,
            downtime_active,
            is_service_downtime=True
        )

        self.assertTrue(in_downtime)

    def test_logdata_with_inactive_service_downtime(self):
        in_downtime = self.handle_downtime(
            DATETIME_IN_DOWNTIME,
            notification_service_critical,
            downtime_inactive,
            is_service_downtime=True
        )

        self.assertFalse(in_downtime)

    def test_logdata_with_active_host_downtime(self):
        in_downtime = self.handle_downtime(
            DATETIME_IN_DOWNTIME,
            notification_host_critical,
            downtime_active
        )

        self.assertTrue(in_downtime)

    def test_logdata_with_inactive_host_downtime(self):
        in_downtime = self.handle_downtime(
            DATETIME_IN_DOWNTIME,
            notification_host_critical,
            downtime_inactive
        )

        self.assertFalse(in_downtime)
