# -*- coding: UTF-8 -*-

"""
Query the db for new alerts
"""

# System imports
import dateutil.parser
import datetime
import django.db                as db

# Project imports
import symonitoring_alert.common                as cmn
import symonitoring_model.models                as mod
import symonitoring_alert.constants             as const
import symonitoring_alert.handle.nagioslog      as nl
import symonitoring_alert.handle.alert          as al
import symonitoring_alert.handle.hold           as hld
import symonitoring_alert.handle.hold_inform    as hldi

s = cmn.get_settings()


class QueryBase(cmn.ExecutorMixin):
    """ Base class for all query handlers """
    _global_disable_cache = None

    @cmn.run_on_executor
    def global_disable_ex(self):
        """Call global_global disable on executor"""
        return self.global_disable()

    def global_disable(self):
        """Get global disable setting"""
        if QueryBase._global_disable_cache:
            elapsed = cmn.now() - QueryBase._global_disable_cache[1]
            if elapsed < datetime.timedelta(
                    seconds=10
            ):
                return QueryBase._global_disable_cache[0]
        try:
            avoid_alert_date = mod.Setting.objects.filter(
                key="avoidalert_date"
            ).all()[0]
        except IndexError:
            return False
        assert isinstance(avoid_alert_date, mod.Setting)
        if not avoid_alert_date.value:
            return False
        value = cmn.local_datetime(
            dateutil.parser.parse(avoid_alert_date.value)
        ) > cmn.now()
        QueryBase._global_disable_cache = (value, cmn.now())
        return value

    @cmn.run_on_executor
    def reconnect_db(self):
        """ Forces django to reconnect the database """
        # Forces django to reconnect to the database
        db.connection.close()
        # Forces pymysql to reconnect to the database
        db.close_old_connections()


class Hold(QueryBase):
    """ Query hold tables """

    @cmn.run_on_executor
    def get_hold_entries(self):
        """Read existing hold entries"""
        hold = mod.Hold.objects.all().select_related().order_by('id')
        inform = mod.HoldInform.objects.all().select_related().order_by('id')
        cmn.lg.info(
            "Existing hold/inform entries: %s/%s",
            len(hold),
            len(inform),
        )
        return (hold, inform)

    @cmn.run_on_executor
    def remove_bad_hold_entry(self, entry):
        """Removes a bad entry from hold table"""
        entry.delete()

    @cmn.coroutine
    def handle(self):
        """Handle hold table entries"""
        res = yield self.global_disable_ex()
        if res:
            return
        try:
            (hold, hold_inform) = yield self.get_hold_entries()

            for entry in hold:
                try:
                    yield hld.HoldHandler().handle(entry)
                except Exception as e:  # pylint: disable=broad-except
                    # Isolate errors between handlers
                    cmn.lg.exception(
                        "Hold handler raised an exception %s (%s)",
                        e,
                        entry.nagioslogid_id
                    )
                    yield self.remove_bad_hold_entry(entry)
                    if cmn.raise_handler_needed(e):
                        raise

            for entry in hold_inform:
                try:
                    yield hldi.HoldInformHandler().handle(entry)
                except Exception as e:  # pylint: disable=broad-except
                    # Isolate errors between handlers
                    cmn.lg.exception(
                        "Hold inform handler raised an exception %s (%s)",
                        e,
                        entry.nagioslogid_id
                    )
                    yield self.remove_bad_hold_entry(entry)
                    if cmn.raise_handler_needed(e):
                        raise

        except db.utils.DatabaseError:
            cmn.lg.exception(
                "There was a problem accessing the database"
            )
            yield self.reconnect_db()


class NagiosLog(QueryBase):
    """Query the nagios log"""

    @cmn.coroutine
    def initialize_alerts(self):
        """Read existing alerts and start notifications"""
        res = yield self.get_init_alerts()
        # Parallel handling
        yield [
            self._initialize_alert(alert, entry) for alert, entry in res
        ]

    @cmn.coroutine
    def _initialize_alert(self, alert, entry):
        """ Start notifcations for alert """
        try:
            handler = nl.NagiosLog()
            yield handler.handle(entry, no_init=True, no_exec=True)

            logdata = handler.logdata
            yield logdata.alert.sync_ex(alert.id)
            yield logdata.alert_state.sync_ex()

            alert = logdata.alert.model

            if not(
                    alert.acknowledgedautomatically and
                    not handler.logdata.alert.model.acknowledged
            ):
                with (yield self.lock.acquire()):
                    logdata.alert_state.increment_rule()
                    alert_handler = al.AlertExec()
                    alert_handler.logdata = handler.logdata
                    yield alert_handler.handle()
                    yield handler.save_ex(logdata.log)
            else:
                yield logdata.alert_state.delete_ex()
                cmn.lg.debug(
                    "Alert (ID %d) was already acknowledged, "
                    "skipping.." % alert.id
                )
        except Exception as e:  # pylint: disable=broad-except
            # Isolate errors between handlers
            cmn.lg.exception(
                "Handler raised an exception %s (%s)",
                e,
                entry.id
            )
            if cmn.raise_handler_needed(e):
                raise

    @cmn.run_on_executor
    def get_acknowledged_alerts(self):
        """Read existing alerts and start notifications"""
        alerts = mod.Alert.objects.filter(
            acknowledged=True,
            acknowledgedprocessed=False
        ).select_related()
        cmn.lg.info("Acknowldged alerts: %s", len(alerts))
        return self.get_alert_entries(alerts)

    def get_alert_entries(self, alerts):
        """Get the nagioslog entry belonging to a alert"""
        alert_list = []
        for alert in alerts:
            assert isinstance(alert, mod.Alert)
            try:
                entry = mod.Nagioslog.objects.filter(
                    alert=alert
                ).select_related().order_by('-id')[0]
            except IndexError:  # pragma: no cover
                cmn.lg.warning(
                    "Couldn't find NagiosLog for alert %s",
                    alert.id
                )
                continue
            alert_list.append((alert, entry))
        return alert_list

    @cmn.run_on_executor
    def get_init_alerts(self):
        """Read existing alerts and start notifications"""
        alerts = mod.Alert.objects.filter(
            acknowledged=False
        ).select_related()
        cmn.lg.info("Existing alerts: %s", len(alerts))
        return self.get_alert_entries(alerts)

    @cmn.coroutine
    def handle(self):
        """Handle nagios log"""
        try:
            log_entries = yield self.query_nagios()
            for entry in log_entries:
                try:
                    yield nl.NagiosLog().handle(entry)
                except Exception as e:  # pylint: disable=broad-except
                    # Isolate errors between handlers
                    cmn.lg.exception(
                        "Handler raised an exception %s (%s)",
                        e,
                        entry.id
                    )
                    if cmn.raise_handler_needed(e):
                        raise
        except db.utils.DatabaseError as e:
            cmn.lg.exception(
                "There was a problem accessing the database: %s" % e
            )
            yield self.reconnect_db()

    @cmn.run_on_executor
    def query_nagios(
            self,
    ):
        """Get all new alerts non-blocking"""
        processed_status = const.NagiosLog.STATE_OPEN
        log = mod.Nagioslog.objects.filter(
            processed=processed_status,
        ).order_by("id").select_related()
        cmn.lg.debug(
            "New log entries with state %s: %s",
            processed_status,
            len(log)
        )
        return log
